/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Mon Dez 11 13:52:07 CET 2000
    copyright            : (C) 2000 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/




#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

//zum Testen jeweils ein Beispiel auskommentieren
//#define BSP_BACKPROP  //Backprop-Netz - Grundlagen
//#define BSP_KOHONEN   //Kohonen-Netz
//#define BSP_MULTI     //Multi-Netz
//#define BSP_SNNS      //Einbindung des SNNS
#define TEST_VERGLEICH_SNNS_NEURONICA
//#define BENCHMARK1      //Benchmark zum Vergleich der Lernschritte  (z.B. beim SNNS)
//#define REGRESSION_TESTS  //zum Testen der korrekten Funktionalitt
//#define DEBUG_BSP     //nur zum Testen

//jeweils eines auskommentieren, um die jeweiligen Programme zu erhalten
//#define PROGRAMM_NND2NET
//#define PROGRAMM_PATTERNSHUFFLE

#include <iostream.h>
#include <stdlib.h>
#include <time.h>   //Zufallszahlen initialisieren
//#include <unistd.h> //for sleep
#include "neuronica/neuronica.h"

int main(int argc, char *argv[])
{
    CNeuronicNet* net;                          //Das Netz
    std::vector<CDoubleVector> inputs;          //die Input-Werte
    std::vector<CDoubleVector> outputs;         //die Sollvorgaben
    CDoubleVector ergebnis;                     //der Output des Netzes
    double data;
    CPatterns patterns;                     
    char i;
    CDoubleVector input_min,input_max,output_min,output_max;   //

#ifdef BSP_SNNS
    //Netz erzeugen
    CNeuronicSnnsNet* snet=new CNeuronicSnnsNet();
    //Laden der Patterns
    CNeuronica::load_patterns("xor_raw.pat",inputs,outputs);
    //-----------------
    cout<<"Test zum Laden eines mit Neuronica trainierten Netzes"<<endl;
    snet->load("xor_trained.nnd","NND");
    for (ulong i=0; i<inputs.size();i++)
    {
        //dem Netz die Inputs prsentieren
        snet->set_input(inputs[i],false);
        //die Werte vorwrts propagieren
        snet->feed_forward();
        //das Netz-Ergebnis auslesen
        snet->get_output(ergebnis,false);
        cout<< i+1<<". Fall ("<<inputs[i]<<") :"<<ergebnis<<endl;
    }
    snet->reset();

    //-----------------
    cout<<"\nUntrainiertes SNNS-Netz laden und trainieren:"<<endl;
    snet->load("xor_snnstest.net");

    //Einstellen der Lernfunktion:
    snet->set_learning_function("Std_Backpropagation");
    //snet->set_learning_function("BackpropMomentum");
    //snet->set_learning_function("BackPercolation");
    //snet->set_learning_function("Quickprop");
    //snet->set_learning_function("Monte-Carlo");
    //snet->set_learning_function("BackpropWeightDecay");
    //snet->set_learning_function("Sim_Ann_WWTA");
    //snet->set_learning_function("Rprop");

    //snet->train(Anzahl_der_Trainingsschritte)
    
    long anz_snns_iterationen=5000;
    for (long jj=0; jj<=anz_snns_iterationen;jj++)
    {
        for (ulong i=0; i<inputs.size();i++)
        {
            //dem Netz die Inputs repsentieren
            snet->set_input(inputs[i],false);
            //die Werte vorwrts propagieren
            snet->feed_forward();
            //das Netz-Ergebnis auslesen
            snet->get_output(ergebnis,false);
            //den Fehler ermitteln und setzen
            snet->set_error(outputs[i]-ergebnis,false);			//das false bedeutet, dass keine Skalierung verwendet werden soll
            //den Fehler zurck propagieren
            snet->feed_back();
            if ((jj % 500) ==0) cout<< i+1<<". Fall ("<<inputs[i]<<") :"<<ergebnis<<endl;
        }
    }
    cin >>i;
    //-----------------
    cout<<"\nTest zum Laden und speichern:"<<endl;
    cout<<"Die beiden folgenden Ergebnisse sollten gleich sein:"<<endl;
    cout<<"Ergebnis vor dem Speichern :"<<endl;
    for (ulong i=0; i<inputs.size();i++){
        snet->calc_output(inputs[i],ergebnis,false);
        cout<< i+1<<". Fall ("<<inputs[i]<<") :"<<ergebnis<<endl;
    }
    snet->save("tmps.net");
    snet->reset();
    snet->load("tmps.net");
    cout<<"Ergebnis nach Speichern und anschlieenden Laden:"<<endl;
    for (ulong i=0; i<inputs.size();i++){
        snet->calc_output(inputs[i],ergebnis,false);
        cout<< i+1<<". Fall ("<<inputs[i]<<") :"<<ergebnis<<endl;
    }
    delete snet;

    //Testen des XOR-Beispiels, dass mit dem SNNS mitgeliefert wird:
    //----------------------------------------------------------------
    CNeuronicSnnsNet ssnet;
    ssnet.load("/usr/share/doc/packages/snns/examples/xor_untrained.net");
    //FIXME: Das voreingestellte Lernverfahren Quickprop funktioniert (noch) nicht
    ssnet.set_learning_function("Std_Backpropagation");
    //Laden der Patterns
    CPatterns pats;
    pats.load_patterns("/usr/share/doc/packages/snns/examples/xor.pat");
    for (ulong i=0;i<5000;i++){
        ssnet.learn_all(pats,false);
        if ((i % 50) ==0){        
            for (ulong i=0; i<pats.inputs.size();i++){
                ssnet.calc_output(pats.inputs[i],ergebnis,false);
                cout<< i+1<<". Fall ("<<pats.inputs[i]<<") :"<<ergebnis<<endl;
            }
        }
    }
    //Ausgeben des Endergebnises
    for (ulong i=0; i<pats.inputs.size();i++){
        ssnet.calc_output(pats.inputs[i],ergebnis,false);
        cout<< i+1<<". Fall ("<<pats.inputs[i]<<") :"<<ergebnis<<endl;
    }

    //Testen des Letter-Beispiels, dass mit dem SNNS mitgeliefert wird:
    //----------------------------------------------------------------
    ergebnis.resize(26);  //da 26 Buchstaben
    ssnet.reset(true);
    ssnet.load("/usr/share/doc/packages/snns/examples/letters_untrained.net");
    //FIXME: Das voreingestellte Lernverfahren Quickprop funktioniert (noch) nicht
    //ssnet.set_learning_function("Std_Backpropagation");
    ssnet.set_learning_function("BackpropMomentum");
    //Laden der Patterns
    pats.delete_all();
    pats.load_patterns("/usr/share/doc/packages/snns/examples/letters.pat");
    for (ulong i=0;i<=100;i++){
        //Lernen fr alle Patterns
        ssnet.learn_all(pats,false);
        //Auswertung (nach einer best. Schrittzahl)
        if ((i % 30) ==0){
            cout<<"\n\nDURCHLAUF "<<i<<" :"<<endl;
            for (ulong i=0; i<pats.inputs.size();i++){
                ssnet.calc_output(pats.inputs[i],ergebnis,false);
                cout<< i+1<<". Fall"<<endl;
                for (long jx=0;jx<7;jx++){
                   for (long jy=0;jy<5;jy++) if (pats.inputs[i][jx*5+jy]==1) cout<<"X"; else cout<<".";
                   cout<<endl;
                }
                cout<<"------> "<<char('A'+ergebnis.elem_min_diff(1.0))<<endl<<endl;
            }
        }
    }
    //Ausgeben des Endergebnises
    cout<<"\n\nENDERGEBNIS :"<<endl;
    for (ulong i=0; i<pats.inputs.size();i++){
        ssnet.calc_output(pats.inputs[i],ergebnis,false);
        cout<< i+1<<". Fall"<<endl;
        for (long jx=0;jx<7;jx++){
           for (long jy=0;jy<5;jy++) if (pats.inputs[i][jx*5+jy]==1) cout<<"X"; else cout<<".";
           cout<<endl;
        }
        cout<<"------> "<<char('A'+ergebnis.elem_min_diff(1.0))<<endl<<endl;
    }
        
#endif


#ifdef TEST_VERGLEICH_SNNS_NEURONICA
//TEST ob SNNS und Neuronica identisch sind
    const double LEARNING_RATE =0.5;

    cout<<"\n--------------------------------------------------------------------\n";
    cout<<"Erster Test:\n";
    cout<<"Netz ber Neuronica erstellen, untrainiert als Neuronica-NET-Datei\n";
    cout<<"speichern, dann unabhngig 1000 Schritte mit dem SNNS und Neuronica\n";
    cout<<"trainieren und die Ergebnisse miteinander vergleichen.\n";
    cout<<"--------------------------------------------------------------------"<<endl;
    
    //Das Neuronica-BP-Netz erstellen und gleichzeitig als SNNS-Netz speichern
    CNeuronicNet bp_net;
    bp_net.add_layer(CInputNeuron::TYP,2);		//1. Schicht besteht aus 2 Inputneuronen  
    bp_net.add_layer(CBackpropNeuron::TYP,2);	// 2 Backpropneuronen in der verborgenen Schicht
    bp_net.add_layer(CBackpropNeuron::TYP,1);	// 1 Backpropneuron in der Output-Schicht
    bp_net.connect(CT_ALL_BEETWEEN_LAYERS,-1,1);
    //Normal und als SNNS-Netz speichern
    bp_net.save("xor_test1.nnd");
    bp_net.save("xor_test1.net","SNNSNET");
    //Die Muster fr ein XOR-Problem laden
    CPatterns pat;
    pat.load_patterns("xor.pat");
    //Training mit dem SNNS
    CNeuronicSnnsNet s_net;
    s_net.load("xor_test1.net");
    //Die Netze trainieren
    cout<<"Einstellen der gleichen Lernrate ("<<LEARNING_RATE<<") fr beide Netze:"<<endl;
    bp_net.set_learning_rate(LEARNING_RATE);
    s_net.set_learning_rate(LEARNING_RATE);
    //Das Training;
    cout<<"BP - Training   "<<flush;
    for (long i=1;i<1000;i++){
        bp_net.learn_all(pat,false);
        if ((i%100)==0) cout<<"."<<flush;
    }
    cout<<"done\n";
    cout<<"SNNS - Training "<<flush;
    for (long i=1;i<1000;i++){
        s_net.learn_all(pat,false);
        if ((i%100)==0) cout<<"."<<flush;
    }
    cout<<"done"<<endl;

    //Lernergebnis ausgeben
    cout<<"\nLernen liefert:"<<endl;
    for (ulong i=0; i<pat.inputs.size();i++){
        ergebnis.resize(0);   //muss nicht sein
        bp_net.set_input(pat.inputs[i],false);
        bp_net.feed_forward();
        bp_net.get_output(ergebnis,false);
        cout<< i+1<<". Fall BP   ("<<pat.inputs[i]<<"): "<<ergebnis<<endl;
        ergebnis.resize(0);   //muss nicht sein
        s_net.set_input(pat.inputs[i],false);
        s_net.feed_forward();
        s_net.get_output(ergebnis,false);
        cout<< i+1<<". Fall SNNS ("<<pat.inputs[i]<<"): "<<ergebnis<<endl;
    }
    
    cout<<"\n--------------------------------------------------------------------\n";
    cout<<"Zweiter Test:\n";
    cout<<"Netz ber Neuronica erstellen, trainieren und dannach mit dem SNNS\n";
    cout<<"weitertrainieren.\n";
    cout<<"--------------------------------------------------------------------"<<endl;
    //Den Zustand vor dem Training wieder herstellen
    bp_net.load("xor_test1.nnd");
    //Das Neuronica-BP-Netz trainieren
    for (long i=1;i<2000;i++){
        bp_net.learn_all(pat,false);    
    }     
    //Lernergebnis ausgeben 
    cout<<"\nBackprop-Netz-Lernen liefert:"<<endl;
    for (ulong i=0; i<pat.inputs.size();i++){
        bp_net.set_input(pat.inputs[i],false);
        bp_net.feed_forward();
        bp_net.get_output(ergebnis,false);
        cout<< i+1<<". Fall ("<<pat.inputs[i]<<"): "<<ergebnis<<endl;
    }
    //Netz abspeichern: einmal als Neuronica-NND-Netz und einmal als SNNN-NET-Netz    
    bp_net.save("xor_raw.nnd");
    bp_net.save("xor_raw.net","SNNSNET");

    //Vergleich Backprop - SNNS
    //Das eben gespeicherte Netz im SNNS-Net-Format ber den SNNS laden
    cout<<"Zum Vergleich: trainiertes Neuronica-Netz speichern, ber den SNNS\n";
    cout<<"laden und dann testen liefert:"<<endl;
    CNeuronicSnnsNet snns_net;  
    snns_net.load("xor_raw.net");    
    ergebnis.resize(0);
    //SNNS-Netz sollte die gleichen Werte liefern, wie vorher das Neuronica-BP-Netz
    for (ulong i=0; i<pat.inputs.size();i++){
        snns_net.set_input(pat.inputs[i],false);
        snns_net.feed_forward();
        snns_net.get_output(ergebnis,false);
        cout<< i+1<<". Fall ("<<pat.inputs[i]<<"): "<<ergebnis<<endl;
    }
    cout<<endl<<endl;
    //weiteren Trainingsschritten um zu zeigen, dass auch bei weiteren Training die selben Werte
    //erzeugt werden:
    cout<<"Weitertrainieren der beiden Netze unabhngig voneinander sollte die selben Ergebnisse liefern\n";
    cout<<"Einstellen der selben Lernrate ("<<LEARNING_RATE<<") fr beide Netze:"<<endl;
    bp_net.set_learning_rate(LEARNING_RATE);
    snns_net.set_learning_rate(LEARNING_RATE);
    //das eigentliche Training
    for (ulong j=1;j<=500;j++){
        if ((j % 50) ==0) cout<<endl<<"Lern-Schritt Nr. "<<j<<endl;
        for (ulong i=0; i<pat.inputs.size();i++){
            //Das SNNS-Netz trainieren
            snns_net.set_input(pat.inputs[i],false);
            snns_net.feed_forward();
            snns_net.get_output(ergebnis,false);
            snns_net.set_error(pat.outputs[i]-ergebnis,false);
            snns_net.feed_back();
            if ((j % 50) ==0) cout<< i+1<<". Fall SNNS ("<<pat.inputs[i]<<") :"<<ergebnis<<endl;
            ergebnis.resize(0); //Damit man sieht das dies kein Trick ist
            //Das NEuronica-BP-Netz trainieren
            bp_net.set_input(pat.inputs[i],false);
            bp_net.feed_forward();
            bp_net.get_output(ergebnis,false);
            bp_net.set_error(pat.outputs[i]-ergebnis,false);
            bp_net.feed_back();
            if ((j % 50) ==0) cout<< i+1<<". Fall BP   ("<<pat.inputs[i]<<") :"<<ergebnis<<endl;
            ergebnis.resize(0); //Damit man sieht das dies kein Trick ist
        }
    }
#endif

  
#ifdef BSP_BACKPROP
    // ##############################################################################
    // # BSP_BACKPROP:
    // # Backprop-Netz mit 1 versteckten Schicht, dass die XOR-Funktion vorhersagt
    // ##############################################################################

    // 1. Schritt: Anlegen des Netzes
    // ==============================

    //als erstes ein leeres Netz erzeugen = der FRAMEWORK, der das eigentliche Netz aufnimmt
    net=new CNeuronicNet();

    //die Schichten anlegen und zum Netz hinzufgen
    net->add_layer(CInputNeuron::TYP,2);		//1. Schicht besteht aus 2 Inputneuronen
    net->add_layer(CBackpropNeuron::TYP,2);	// 2 Backpropneuronen in der verborgenen Schicht
    net->add_layer(CBackpropNeuron::TYP,1);	// 1 Backpropneuron in der Output-Schicht
    /* so ginge es auch:
    CNeuronicLayer* layer=new CNeuronicLayer();
    layer->add_neurons(CInputNeuron::TYP,2);			//die erste Schicht muss aus Inputneuronen bestehen ...
    net->add_layer(layer);
    layer=new CNeuronicLayer();						  //die anderen Schichten knnen
    layer->add_neurons(CBackpropNeuron::TYP,2);
    net->add_layer(layer);
    layer=new CNeuronicLayer();	//die output layer
    layer->add_neurons(CBackpropNeuron::TYP,1);
    net->add_layer(layer);
    /**/

    //das Netz verbinden,
    //dabei wird (soweit nichts anderes angegeben zuflliges Gewicht zwischen -1 und 1 verwendet
    net->connect(CT_ALL_BEETWEEN_LAYERS,-1,1);
    //so ginge es auch:
    // net->layer(0)->connect_output(net->layer(1),-1,1);
    // net->layer(1)->connect_output(net->layer(2),-1,1);

    //optional noch ein paar Einstellungen vornehmen
    net->process_command("layer(*).neuron(*).slope",1.9);


    // 2. Schritt: Falls eine Skalierung verwendet werden soll, diese festelegen
    // ==============================
    // Achtung die zu verwendende Skalierung hngt zum Beispiel auch von der Aktivierungsfunktion ab
    // die Funktion zur Einstellung der Skalierung erwartet einen Vektor (CDoubleVector), der jeweils
    // die min und max-Werte fr jedes Neuron beinhaltet
    input_min.push_back(0);   input_max.push_back(1);  //das 1. Neuron in der Inputschicht akzeptiert Werte zwischen  0 und 1
    input_min.push_back(0);   input_max.push_back(1);  //das 2. Neuron in der Inputschicht akzeptiert Werte zwischen  0 und 1
    output_min.push_back(0);  output_max.push_back(1); //das Neuron der Outputschicht soll seine Ausgabe-Werte so skalieren, dass sie zw.-1 und 1 liegen
    // die vorbereiteten Werte werden gesetzt. dabei bedeuten die 0.1 und die 0.9 auf welche
    // Werte in den Neuronen die entsprechnenden min/max-Werte gesetzt werden.
    // hier wird also davon ausgegangen, dass eine Aktivierungsfunktion verwendet wird,
    // deren Werte zwischen 0 und 1 liegen.
    net->layer(0)->set_scaling(input_min,input_max,0.1,0.9);
    net->layer(2)->set_scaling(output_min,output_max,0.1,0.9);

    // 3. Schritt: Das Training vorbereiten
    // ===================================
    // zunchst werden die Werte vorbereitet, die dem Netz prsentiert werden.
    CNeuronica::load_patterns("xor.pat",inputs,outputs);
    assert(inputs.size()==outputs.size());

    //so speichert man die Pattern:
    CNeuronica::save_patterns("xor_neu.pat",inputs,outputs);

    /* //So knnte man die Werte per Hand eingeben
    //4 Trainingswerte
    inputs.resize(4);  outputs.resize(4);

    // XOR-Funktion
    //1.Trainingsbeispiel
    inputs[0].push_back(0);
    inputs[0].push_back(0);
    outputs[0].push_back(0);
    //2.Trainingsbeispiel
    inputs[1].push_back(0);
    inputs[1].push_back(1);
    outputs[1].push_back(1);
    //3.Trainingsbeispiel
    inputs[2].push_back(1);
    inputs[2].push_back(0);
    outputs[2].push_back(1);
    //4.Trainingsbeispiel
    inputs[3].push_back(1);
    inputs[3].push_back(1);
    outputs[3].push_back(0);
    */

    //Speichern im NND und im NET-Format muss nicht sein:
    net->save("xor_untrained.nnd");
    net->save("xor_untrained.net","SNNSNET");
    const char* dateiname;


    // 4.Schritt : das eigentliche Training
    // ===================================
    //Das Training
    long anzahl_trainingsiterationen=5000;
    for (long jj=0; jj<anzahl_trainingsiterationen;jj++){
        for (ulong i=0; i<inputs.size();i++){
            //dem Netz die Inputs repsentieren
            net->set_input(inputs[i],true);		//das true bedeutet, dass Skalierung verwendet werden soll
            //die Werte vorwrts propagieren
            net->feed_forward();
            //das Netz-Ergebnis auslesen
            net->get_output(ergebnis,true);		//das true bedeutet, dass Skalierung verwendet werden soll
            //den Fehler ermitteln und setzen
            net->set_error(outputs[i]-ergebnis,true);			//das true bedeutet, dass Skalierung verwendet werden soll
            //den Fehler zurck propagieren
            net->feed_back();
            if ((jj % 100) ==0) cout<< i+1<<". Fall ("<<inputs[i]<<") :"<<ergebnis<<endl;
        }
    }


    //5. Schritt: die Anwendung
    // ===================================
    //geht im Prinzip auch wie das Training, nur dass der feed-back-Schritt weggelassen wird
    CDoubleVector in,out;
    in.push_back(0);
    in.push_back(1);
    net->calc_output(in,out,true);
    cout<< "\nDas Ergebnis fr ("<<in<<") lautet "<<out;

    //so sehen Zugriffe auf die Innereien des Netzes aus
    cout<<"\nDas Gewicht der 1. Input-Synapse des 1. Neurons der 2. Schicht des Netzes lautet :"
        <<net->layer(1)->neuron(0)->input_synapse(0)->get_weight()<<endl;

    net->save("xor_trained.nnd");
    net->process_command("layer(*).neuron(*).ow=-5:5");
    delete net;
#endif



#ifdef BSP_KOHONEN
    // ##############################################################################
    // # BSP_KOHONEN:
    // # Kohonen-Map, der zufllige Inputs zwischen 0 und 1 reprsentiert werden
    // # nach dem Training sollte ein Gitter ber (0,0)-(1,1) ausgebildet sein
    // ##############################################################################

    // 1. Schritt: Anlegen des Netzes
    // ==============================
    CNeuronicNet* knet=new CNeuronicNet();
    //eine Schicht mit den Inputneuronen und eine Kohonenschicht (!!) mit den Kohonenneuronen
    //die Schichten anlegen und zum Netz hinzufgen
    CNeuronicLayer* layer=new CNeuronicLayer();
    layer->add_neurons(CInputNeuron::TYP,2);
    knet->add_layer(layer);
    CKohonenLayer* klayer=new CKohonenLayer();	//die eigentliche Kohonenschicht
    klayer->breite=4;
    klayer->add_neurons(CKohonenNeuron::TYP,16); //4x4 Neuronen
    knet->add_layer(klayer);

    //das Netz verbinden
    knet->connect(CT_ALL_BEETWEEN_LAYERS,0,0.9);


    // 2. Schritt: Training
    // ================================
    CDoubleVector inp;
    inp.resize(2);
    srand(2);
    for (long i=0; i<10000;i++){
        //>>> dient nur zur Veranschaulichung Ausdrucken der Gewichcte:
            // Zuerst die Gewichte bestimmen (vorher)
            if ((i % 100) ==0){
                CDoubleVector weights;
                for (long j=0; j<knet->layer(1)->get_anzahl_neurons();j++){
                    knet->layer(1)->neuron(j)->get_weight_vector(weights);
                    cout<< j+1<<". Neuron: "<<weights<<endl;
                }
            }
        //<<<<<

        inp[0]=((double)rand()/(double)RAND_MAX); //liefert eine Zufallszahl zw. 0 und 1
        inp[1]=((double)rand()/(double)RAND_MAX); //liefert eine Zufallszahl zw. 0 und 1
        knet->set_input(inp,false);  //keine Skalierung verwenden
        knet->feed_forward();
        knet->feed_back();

        //>>>>Ausgabe des Gewinners...
            if ((i % 100) ==0){
                //den Input + Gewinner ausgeben
                cout<<"----Input "<<inp<<endl;  //der Input der zu dem folgenden Gewichten fhrte
                for (long j=0; j<knet->layer(1)->get_anzahl_neurons();j++){
                    if (reinterpret_cast<CKohonenNeuron*>(knet->layer(1)->neuron(j))->is_winner) cout<<"And the winner is: "<<j+1<<endl;
                }
            }
        //<<<<<<
    }
    knet->save("kohonentest.nnd");
    delete knet;
#endif


#ifdef BSP_MULTI
    // ##############################################################################
    // # BSP_MULTI:
    // # Beispiel fr ein Multinetz (d.h. ein Netz, dass aus mehreren Teilnetzen besteht
    // ##############################################################################
    CNeuronicMultiNet* mnet=new CNeuronicMultiNet();

    //xor_untrained.nnd wird mit Beispiel BSP_BACKPROP erzeugt
    net=CNeuronica::load_new_net("xor_untrained.nnd");
    mnet->add_subnet(net);
    net=CNeuronica::load_new_net("xor_untrained.nnd");
    mnet->add_subnet(net);
    net=CNeuronica::load_new_net("xor_untrained.nnd");
    mnet->add_subnet(net);

    CNeuronica::load_patterns("xor.pat",inputs,outputs);

    //Das Training (trainiert alle 3 Netze als XOR-Netze
    for (long jj=0; jj<5000;jj++){
      for (ulong i=0; i<inputs.size();i++){
        for (long netnr=0; netnr<3; netnr++){
            //dem Netz die Inputs repsentieren, Output bestimmen, Fehler zurckpropagieren
            mnet->subnet(netnr)->set_input(inputs[i],true);		//das true bedeutet, dass Skalierung verwendet werden soll
            mnet->subnet(netnr)->feed_forward();
            mnet->subnet(netnr)->get_output(ergebnis,true);		//das true bedeutet, dass Skalierung verwendet werden soll
            mnet->subnet(netnr)->set_error(outputs[i]-ergebnis,true);			//das true bedeutet, dass Skalierung verwendet werden soll
            mnet->subnet(netnr)->feed_back();
            if ((jj % 100) ==0) cout<<"Netz "<<netnr <<": "<< i+1<<". Fall ("<<inputs[i]<<") :"<<ergebnis<<endl;
        }
      }
    }
    CDoubleVector tst,erg,soll;
    tst.resize(2); erg.resize(1); soll.resize(1);
    bool connect=true;
    //jetzt werden die Netze zusammengeschaltet und es wird getestet, ob die richtige logische Funktion berechnet wird
    //und zwar XOR(XOR(X1,X2),XOR(X3,X4)) liefert true, wenn x1<>x2 und x3=x4 oder wenn x1=x2 und x3<>x4
    if (connect){
    //zuerst die Sub-Netze gem der Formel verbinden:
    mnet->connect(0,2,OL0,IL0,0);	//verknpfe alle Neuronen der Outputlayer von Netz 0 mit der Inputlayer von Netz 2, beginne dabei mit dem Neuron 0 der Inputlayer
    mnet->connect(1,2,OL0,IL0,1); //verknpfe alle Neuronen der Outputlayer von Netz 1 mit der Inputlayer von Netz 2, beginne dabei mit dem Neuron 1 der Inputlayer
    cout<<"\n\n\n Testen der Funktionalitt :\n"<<endl;
    for (long x1=0;x1<=1;x1++)
      for (long x2=0;x2<=1;x2++)
        for (long x3=0;x3<=1;x3++)
          for (long x4=0;x4<=1;x4++){
              cout<<"X1: "<<x1<<", X2: "<<x2<<", X3: "<<x3<<", X4: "<<x4;
              tst[0]=x1; tst[1]=x2;
              mnet->set_input(tst,true,0);  //das erste Xor-NEtz fttern
              tst[0]=x3; tst[1]=x4;
              mnet->set_input(tst,true,1);  //das zweite Xor-NEtz fttern
              mnet->feed_forward();
              mnet->get_output(erg,true);   //Ergebnis auslesen
              //cout.setf(ios::fixed);
              //cout.precision(3);
              cout<<"      ERGEBNIS: "<<erg[0]<<endl;
          }
    //Ein Lernen oder Umlernen des zusammengeschalteten Netzes ist mglich;
    //ein Wert der Funktion soll umgelernt werden, und zwar wenn alle 0 sind, soll 1 ausgegeben werden
    cout<<"\n\n\nUmlernen:\n"<<endl;
    for (long i=0; i<5000;i++)
      for (long x1=0;x1<=1;x1++)
        for (long x2=0;x2<=1;x2++)
          for (long x3=0;x3<=1;x3++)
            for (long x4=0;x4<=1;x4++){
                tst[0]=x1; tst[1]=x2;
                mnet->set_input(tst,true,0);			  	//das erste Xor-NEtz fttern
                tst[0]=x3; tst[1]=x4;
                soll[0]=(x1 ^ x2) ^ (x3 ^ x4);				//XOR(XOR(X1,X2),XOR(X3,X4))
                if ((x1==x2) && (x2==x3) && (x3==x4)) soll[0]=1;
                mnet->set_input(tst,true,1);			  	//das zweite Xor-NEtz fttern
                mnet->feed_forward();
                mnet->get_output(erg,true);		//das true bedeutet, dass Skalierung verwendet werden soll
                mnet->set_error(soll-erg,true);			//das true bedeutet, dass Skalierung verwendet werden soll
                mnet->feed_back();
                if ((i % 200) ==0){
                    cout<<"X1: "<<x1<<", X2: "<<x2<<", X3: "<<x3<<", X4: "<<x4;
                    cout<<"      ERGEBNIS: "<<erg[0]<<",soll "<<soll[0]<<endl;
                }
            }
    }else{
        //nur zum testen: wenn man die Neuronen direkt verbindet, sollte es dasselbe Ergebnis liefern
        //funktioniert: connect auf false setzen
        long layneu;	//welche nr erhlt die Schicht  IL0 im subnet0
        layneu=CNeuronica::merge(mnet->subnet(0),mnet->subnet(2),OL0,IL0,0);
        assert(mnet->subnet(2)->get_anzahl_layers()==0);
        cout<<mnet->subnet(0)->get_anzahl_layers()<<endl;
        cout<<mnet->subnet(0)->layer(layneu)->get_anzahl_neurons()<<endl;
        cout<<dynamic_cast<CInputNeuron*>(mnet->subnet(0)->layer(layneu)->neuron(0))<<endl;

        //die erste Schicht des bisherigen Netzes ist eine InputSchicht fr das XOR:
        layneu=mnet->subnet(1)->get_anzahl_layers();
        CNeuronica::merge(mnet->subnet(1),mnet->subnet(0),OL0,static_cast<ELayerCode>(layneu),0);	//layneu ist die nr. der zweiten input-Layer
        assert(mnet->subnet(0)->get_anzahl_layers()==0);
        cout<<dynamic_cast<CInputNeuron*>(mnet->subnet(1)->layer(layneu)->neuron(0))<<endl;
        //das Training
        for (long i=0; i<2000;i++)
          for (long x1=0;x1<=1;x1++)
            for (long x2=0;x2<=1;x2++)
              for (long x3=0;x3<=1;x3++)
                for (long x4=0;x4<=1;x4++){
                    tst[0]=x1; tst[1]=x2;
                    mnet->subnet(1)->set_input(tst,true,layneu);			  	//die erste INput-Layer fttern
                    tst[0]=x3; tst[1]=x4;
                    soll[0]=(x1 ^ x2) ^ (x3 ^ x4);
                    if ((x1==x2) && (x2==x3) && (x3==x4)) soll[0]=1;
                    mnet->subnet(1)->set_input(tst,true,0);			  	//die zweite Input-Layer fttern
                    mnet->subnet(1)->feed_forward();
                    mnet->subnet(1)->get_output(erg,true);		//das true bedeutet, dass Skalierung verwendet werden soll
                    mnet->subnet(1)->set_error(soll-erg,true);			//das true bedeutet, dass Skalierung verwendet werden soll
                    mnet->subnet(1)->feed_back();
                    if ((i % 100) ==0){
                        cout<<"X1: "<<x1<<", X2: "<<x2<<", X3: "<<x3<<", X4: "<<x4;
                        cout<<"      ERGEBNIS: "<<erg[0]<<",soll "<<soll[0]<<endl;
                    }
                }

    }


    //nun kommt ein Check, der testet, ob das Speichern auch richtig funktioniert
    //die beiden sich ergebenden Dateien mssen gleich sein
    mnet->save("tmp1.nnd");
    mnet->subnet(0)->save("tmp_s0.nnd");
    mnet->subnet(1)->save("tmp_s1.nnd");
    //Achtung zum Laden darf nicht nur load verwendet werden, da zustzlich die Verbindungen
    //zwischen den Netzen noch mit beachtet werden mssen
    mnet->load_subnet(0,"tmp_s0.nnd");
    mnet->save("tmp2.nnd");
    mnet->load_subnet(1,"tmp_s1.nnd");
    mnet->save("tmp3.nnd");
#endif

#ifdef BENCHMARK1
    // ##############################################################################
    // # BENCHMARK1:
    // # Benchmark zum Vergleich mit dem SNNS	- XOR
    // ##############################################################################

      //als erstes ein leeres Netz erzeugen...
    net=new CNeuronicNet();

    //Netz laden ( xor_untrained.nnd wird in BSP_BACKPROP erzeugt)
    if(!net->load("xor_untrained.nnd")) cerr<<"Fehler beim Laden der Datei.\n";
    //Lernrate auf 0.9 einstellen
    net->set_learning_rate(0.9);
    //Patterns laden
    patterns.load_patterns("xor.pat");

    //Trainieren
    long anzahl_trainingsiterationen_bench=8000;
    for (long jj=0; jj<anzahl_trainingsiterationen_bench;jj++){
        net->learn_all(patterns);
        cout<< net->get_sse()<<" \t ";
    }
    delete net;
#endif


#ifdef PROGRAMM_NND2NET
    if (argc!=3){
        cout << "nnd2net: wrong number of parameters\n";
        cout << "Usage: nnd2net nnd_file snns_net_file_name\n\n";
        return -1;
    }
    return CNeuronica::nnd2net(argv[1],argv[2]);
#endif
  
#ifdef PROGRAMM_PATTERNSHUFFLE
    if (argc!=3){
        cout << "patternshuffle: wrong number of parameters\n";
        cout << "Usage: patternshuffle pattern_file new_pattern_file\n\n";
        return -1;
    }
    CPatterns pat;
    cout<<"Reading "<<argv[1]<< " ..."<<endl;
    pat.load_patterns(argv[1]);
    //Zufallszahlen initialisieren je nach STL-Implementation muss eine andere Funktion aufgerufen werden
#ifndef __STL_NO_DRAND48
    srand48((unsigned int)time((time_t *)NULL));
#else
    srand((unsigned int)time((time_t *)NULL));
#endif
    cout<<"Shuffling... "<<endl;
    pat.shuffle();
    cout<<"Saving "<<argv[2]<<" ..."<<endl;
    return pat.save_patterns(argv[2]);
#endif

    return EXIT_SUCCESS;
  
}
