/* **************************************************************************
                          cbackpropneuron.cpp  -  description
                             -------------------
    begin                : Mon Dec 11 2000
    copyright            : (C) 2000 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cbackpropneuron.h"
#include <math.h>	//for M_E
#include "csynapse.h"
#include "cdoublevector.h" //for Debug
/* **************************************************************************
 * CBackpropNeuron                                                                        *
 ***************************************************************************/
CBackpropNeuron::CBackpropNeuron(){
    feedback_allowed=true;
    learning_allowed=true;
    error_activation=0;
    old_error_activation=0;
    slope=1;
    //fr die tanh-Funktion
    set_activation_function(AF_TANGENSH,true); //die Parameter werden mit eingestellt: learning_rate=0.5
    //fr die logistische Funktion als Default
    //set_activation_function(AF_LOGISTIC,true); //die Parameter werden mit einstellt: learning_rate=0.7

    //Festlegen der Schwelle
    schwelle=0.0;    
}

CBackpropNeuron* CBackpropNeuron::TYP=0;
//-----------------------------------------------------------------------------
CBackpropNeuron::~CBackpropNeuron(){
}

//-----------------------------------------------------------------------------
/** fhrt einen Feedbackschritt aus
 */
void CBackpropNeuron::feed_back(TNTimeUnit time=-1)
{
	//als erstes den Fehler ermitteln und festsetzen und die error_activation zurcksetzen
	calc_error_output();
	
	//nun den Fehler zurckpropagieren
	if (feedback_allowed)
	{
		for (long i=0;i<get_anzahl_inputs();i++)
		{
				input_neuron(i)->add_error_activation(input_synapse(i)->get_post_synaptic_value(get_last_error_output()));
		}
	}
	//dann die Gewichte updaten (und die Schwelle)
	if (learning_allowed)
	{
		learn_step();
	}
}

//-----------------------------------------------------------------------------
/** Updated die Neuronen entsprechend der Fehleraktivierung */
void CBackpropNeuron::learn_step(){
    //ndere die Gewichte zu den Vorgngerneuronen
    double eo=get_last_error_output();
    for (long i=0;i<get_anzahl_inputs();i++){
        assert(input_neuron(i));
        //
        double delta_weight=learning_rate * eo * input_neuron(i)->get_last_output();
        input_synapse(i)->add_weight(delta_weight);
    }
    //ndere die Schwelle
    double delta_schwelle=learning_rate*eo*1;
    //    cout<<" B"<<long(delta_schwelle*1000)<<endl;
    schwelle+=delta_schwelle;
    //
    learning_rate.step();
}
//-----------------------------------------------------------------------------
/** liefert den Fehler den das Neuron ausweist.
Der Fehler ergibt sich aus der error_activation mal der Ableitung der letzten Aktivierung.
Die Fehleraktivierung ergibt sich aus dem Vorhersagefehler (soll-ist) falls das Neuron in der Outputschicht ist.
Und falls das Neuron in einer Hidden Layer ist aus der Summe von get_error_output()der Nachfolgeneuronen*Gewicht der Synapse zu den Nachfolgerneuronen */
TNErrorValue CBackpropNeuron::calc_error_output(){
    error_output=error_activation*derivative_activation_function(old_activation);
    old_error_activation=error_activation;
    error_activation=0;
    return error_output;
}
//-----------------------------------------------------------------------------
#ifdef USE_CCOMMUNICATION
/** ausfhren eines Kommandos*/
bool CBackpropNeuron::process_cmd(const char* a_command, const void* data,TPtrVector* add_data=0,long id=0){
    if (check_command(a_command,"allow_learning","Falls auf 0 gesetzt drfen die Inputsynapsen nicht mehr lernen (Standard:1)",id) ||
        check_command(a_command,"learning","Falls auf 0 gesetzt drfen die Inputsynapsen nicht mehr lernen (Standard:1)",id))
    {
        if (data){
            const double* new_data=reinterpret_cast<const double*>(data);
            if (*new_data==0) learning_allowed=false;
              else if (*new_data==1)learning_allowed=true;
                else infos("Fehlerhafte Parameter: erlaubt:(0|1)\n");
        }
        if (learning_allowed) infos("Lernen ist aktiviert.\n");
          else infos("Lernen ist deaktiviert.\n");
        return true;
    }
    // die Parentklasse wei hoffentlich was damit anzufangen
    return CNeuron::process_cmd(a_command,data,add_data,id);
}
#endif

#ifdef USE_LIBXML
//-----------------------------------------------------------------------------
/** speichert in das XML-File, zu dem tree gehrt */
xmlNodePtr CBackpropNeuron::save_to_xml(xmlNodePtr tree){
    assert(tree);
    //einen neuen Knoten anlegen (mit dem Klassennamen)
    xmlNodePtr subtree;
    subtree = xmlNewChild(tree, NULL, (const xmlChar *)"CBackpropNeuron",NULL);
    xmlSetProp(subtree, (const xmlChar *)"type",(const xmlChar *) "neuron");	//es handelt sich um ein Element vom Typ Neuron
    //Zuerst die Basisklasse abspeichern
    CNeuron::save_to_xml(subtree);
    return subtree;
}
//-----------------------------------------------------------------------------
/** ldt aus dem XML-File, zu dem tree gehrt */
void CBackpropNeuron::load_from_xml(xmlNodePtr tree){
    assert(tree);
    //in den Knoten gehen (mit dem Klassennamen)
    if (strcmp((const char*)tree->name, "CBackpropNeuron")) throw(666);
    xmlNodePtr subtree;
    subtree=tree->xmlChildrenNode;
    while (subtree!=0)
    {
        if (!strcmp((const char*)subtree->name,"CNeuron")) CNeuron::load_from_xml(subtree);
        //Aus Kompatibilittsgrnden beibehalten:-----start--->
        xmlLoadDouble(subtree,"error_activation",error_activation);
        xmlLoadDouble(subtree,"old_error_activation",old_error_activation);
        xmlLoadDouble(subtree,"error_output",error_output);
        //<------------ende
        subtree=subtree->next;
    }
}
#endif
