/* **************************************************************************
                          cbackpropneuron.h  -  description
                             -------------------
    begin                : Mon Dec 11 2000
    copyright            : (C) 2000 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CBACKPROPNEURON_H
#define CBACKPROPNEURON_H

#include "cbasicneuron.h"

/** Implementiert ein Backprop-Neuron.

    Zur Zeit ist das Standard-Backprop-Verfahren implementiert. Die Verarbeitung erfolgt lokal auf
    der Ebene des Neurons. Letztendlich ergibt sich aber trotzdem die (nicht lokale) Standardformel
    fr Backprop (vgl. Zell, "Simulation Neuronaler Netze").
    Auf Grund der Art der Implementation ist es mglich die Lernparameter (Rate, Anstieg) fr jedes Neuron
    einzeln anzupassen. Weitere Erweiterungen (Momentum) sind bis jetzt noch nicht untersttzt.

    <b> TODO: </b>
    <li> verschiedene Erweiterungen des Lernverfahren untersttzen, insbesondere Momentum-Term (ev. Weight-Decay)
    <li> Resilent Backpropagation implementieren (neue Klasse?)
    </li>

    <b> BUGS: </b>
    <li> Beim Destruktor-Aufruf werden bestehende Verbindungen nicht korrekt weggelscht
    </li>

    @author Jon Hennig
*/
class CBackpropNeuron : public CNeuron
{
public:
  CBackpropNeuron();
  virtual ~CBackpropNeuron();
		
 /** fhrt einen Feedbackschritt aus. Berechnet dazu den Error-Output und sendet diesen an alle
     Vorgnger-Neuronen weiter (unter der Beachtung der Gewichte der Synapsen).
     */
  virtual void feed_back(TNTimeUnit time=-1);

  /** liefert den Fehler den das Neuron ausweist.
      Der Fehler ergibt sich aus der error_activation mal der Ableitung der letzten Aktivierung
      Die Fehleraktivierung ergibt sich aus dem Vorhersagefehler (soll-ist) falls das Neuron in der Outputschicht ist.
      Und falls das Neuron in einer Hidden Layer ist aus der Summe von get_error_output()der Nachfolgeneuronen*Gewicht der Synapse zu den Nachfolgerneuronen
      */
  TNErrorValue calc_error_output();

#ifdef USE_CCOMMUNICATION
    /** Ausfhren eines Kommandos. Folgende stehen zur Verfgung: <b>
    		<li> allow_learning: deaktiviert/aktiviert das Lernen fr das Neuron.
    		</li>
    		@see CCommunication::process_cmd
      */
    virtual bool process_cmd(const char* command, const void* data,TPtrVector* add_data=0,long id=0);
#endif	

#ifdef USE_LIBXML
    /** speichert in das XML-File, zu dem tree gehrt (in den entsprechenden Zweig)
      */
    virtual xmlNodePtr save_to_xml(xmlNodePtr tree);
  
    /** ldt aus dem XML-File, zu dem tree gehrt (in den entsprechenden Zweig)
      */
    void load_from_xml(xmlNodePtr tree);
#endif

  /** diese Variable wird nur bentigt, um den Typ dieses Neurons in add_neurons von CNeuronicLayer
      bestimmen zu knnen.
      */
  static CBackpropNeuron* TYP;

protected:
    //Einstellungen

    /** Aktualisiert die Neuronen entsprechend der Fehleraktivierung
       */
    virtual void learn_step();


};

#endif