/***************************************************************************
                          cBasiceuron.cpp  -  description
                             -------------------
    begin                : Tue Jun 6 2000
    copyright            : (C) 2000 by Jon Hennig
    email                :
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cbasicneuron.h"
#include <math.h>             //for M_E
#include <values.h>           //for MAXFLOAT
#include <iomanip>            //for setprecision()
#include "csynapse.h"
#include "cdoublevector.h"

#include "tools.h"            //for xmlSave*()

#ifdef USE_LIBXML
#include <libxml/parser.h>
#endif

/***************************************************************************
 * CNeuron                                                                        *
 ***************************************************************************/
CNeuron* CNeuron::TYP=0;	//als Kennzeichnung
CNeuron::CNeuron(){
    //die Standardwerte einstellen
    act_activation_function=AF_IDENT;
    slope=1;
    //Festlegen der Schwelle
    schwelle=0;
    learning_rate=0.5;
    learning_allowed=true;
    output_set=false;
    min_error=0;
    activation=0;
    old_activation=0;
}
//-----------------------------------------------------------------------------
/** Destruktor, entfernt alle Verbindungen zu dem Neuron*/
CNeuron::~CNeuron(){
    unconnect_all_outputs();
    unconnect_all_inputs();
}


//-----------------------------------------------------------------------------
void CNeuron::connect_output(CNeuron* other, TNSynapticWeight a_weight){
    assert(other);
    other->create_input_synapse(this,a_weight);
}
//-----------------------------------------------------------------------------
/** Entfernt alle Verbindungen mit dem anderen Neuron, falls dieses angegeben ist.
Ist kein anderes Neuron angegeben, werden alle Verbindungen gelscht. */
void CNeuron::unconnect_all_outputs(CNeuron * other=0){
    	for (long i=0; i<get_anzahl_outputs();i++){
    		if (output_neuron(i)==other || other==0){
    			//weg damit:
    			output_neuron(i)->remove_input(output_synapse(i));
    			delete output_synapse(i);
    			outputs.erase(outputs.begin()+i);
    			i--;	//da der Eintrag auf den i zeigt gerade gelscht wurde
    		}
    	}
}
//-----------------------------------------------------------------------------
/** Entfernt alle Verbindungen mit dem anderen Neuron, falls dieses angegeben ist.
Ist kein anderes Neuron angegeben, werden alle Verbindungen gelscht. */
void CNeuron::unconnect_all_inputs(CNeuron * other=0){
    for (long i=0; i<get_anzahl_inputs();i++){
    	if (input_neuron(i)==other || other==0){
    		//weg damit:
    		input_neuron(i)->remove_output(input_synapse(i));
    		delete input_synapse(i);
    		inputs.erase(inputs.begin()+i);
    		i--;	//da der Eintrag auf den i zeigt gerade gelscht wurde
    	}
    }
}
//-----------------------------------------------------------------------------
/**informiert das Neuron das es ein neues Input aufnehmen solll*/
void CNeuron::add_input(CSynapse* synapse){
    assert(synapse);
    assert(synapse->post_synaptic_neuron==this);
    inputs.push_back(synapse);
}
//-----------------------------------------------------------------------------
/**informiert das Neuron das es ein neues Output aufnehmen solll*/
void CNeuron::add_output(CSynapse* synapse){
    assert(synapse);
    assert(synapse->pre_synaptic_neuron==this);
    outputs.push_back(synapse);
}
//-----------------------------------------------------------------------------
/**informiert das Neuron das es ein neues Input entfernen soll*/
void CNeuron::remove_inoutput(CSynapse* syn,std::vector<CSynapse *> & inoutputs){
    assert(syn);
    assert((syn->post_synaptic_neuron==this) || (syn->pre_synaptic_neuron==this));

    std::vector<CSynapse*>::iterator pos = inoutputs.begin();
    std::vector<CSynapse*>::iterator last = inoutputs.end();
    while (pos != last) {
              std::vector<CSynapse*>::iterator next = pos;
              ++next;
              if (*pos == syn)
                  inoutputs.erase(pos);
              pos = next;
    }
}
//-----------------------------------------------------------------------------
/**informiert das Neuron das es ein Input entfernen soll*/
void CNeuron::remove_input(CSynapse* syn){
    remove_inoutput(syn,inputs);
}
//-----------------------------------------------------------------------------
/**informiert das Neuron das es ein Output entfernen soll*/
void CNeuron::remove_output(CSynapse* syn){
    remove_inoutput(syn,outputs);
}

//-----------------------------------------------------------------------------
void CNeuron::add_excitation(TNPostSynapticValue a_excitation,TNTimeUnit time){
    activation+=a_excitation;
}
//-----------------------------------------------------------------------------
void CNeuron::feed_forward(TNTimeUnit time){
    //als erstes den Output festsetzen (falls ntig)
    if (!output_set) calc_output();
    output_set=false;
    //fr jede Synapse aus den OUTPUTS
    for (long i=0;i<get_anzahl_outputs();i++){
        output_synapse(i)->process_input(get_last_output());
    }
}
//-----------------------------------------------------------------------------
/** nicht implementiert, gegebenenfalls in den abgeleiteten Klassen zu implementieren*/
void CNeuron::feed_back(TNTimeUnit time=-1){
    //im Standardfall gar nichts machen
    //diese Funktion wurde trotzdem hier implementiert,
    //damit abgeleitete Klassen, die diese Funktion ebenfalls
    //nicht bentigen diese nicht nochmal deklarieren mssen
}
//-----------------------------------------------------------------------------
void CNeuron::add_error_activation(TNErrorValue a_error){
    error_activation+=a_error;
}

//-----------------------------------------------------------------------------
/** liefert die Zahl der Input-Synapsen */
long CNeuron::get_anzahl_inputs(){
    return inputs.size();
}
//-----------------------------------------------------------------------------
/** liefert die Zahl der Output-Synapsen */
long CNeuron::get_anzahl_outputs(){
    return outputs.size();
}
//-----------------------------------------------------------------------------
/** liefert das Inputneuron mit der Nummer number */
CNeuron* CNeuron::input_neuron(long number){
    assert(number>=0);
    if (number>=(long)inputs.size()) return 0; //nicht vorhanden
    assert( inputs[number]);
    assert(inputs[number]->pre_synaptic_neuron);
    //
    return inputs[number]->pre_synaptic_neuron;
}
//-----------------------------------------------------------------------------
/** liefert das Outputneuron mit der Nummer number */
CNeuron* CNeuron::output_neuron(long number){
    assert(number<(long)outputs.size());
    assert(number>=0);
    assert( outputs[number]);
    assert(outputs[number]->post_synaptic_neuron);
    //
    return outputs[number]->post_synaptic_neuron;
}
//-----------------------------------------------------------------------------
/** liefert die InputSynapse mit der Nummer number. */
CSynapse* CNeuron::input_synapse(long number){
    assert(number<(long)inputs.size());
    assert(number>=0);
    assert(inputs[number]);
    //
    return inputs[number];
}
//-----------------------------------------------------------------------------
/** liefert die OutputSynapse mit der Nummer number. */
CSynapse* CNeuron::output_synapse(long number){
    assert(number<(long)outputs.size());
    assert(number>=0);
    assert(outputs[number]);
    //
    return outputs[number];
}
//-----------------------------------------------------------------------------
TNPreSynapticValue CNeuron::calc_output(){
    //zur aktivierung noch die Schwelle hinzufgen
    activation+=schwelle;
    //den Output berechnen
    output=activation_function(activation);
    //die Aktivierung zurcksetzen und den alten Wert der Aktivierung abspeichern
    old_activation=activation;
    activation=0;
    return output;
}
//-----------------------------------------------------------------------------
/** liefert den letzten Output des Neurons*/
TNPreSynapticValue CNeuron::get_last_output(){
    return output;
}
//-----------------------------------------------------------------------------
/** setzt den Output des Neurons auf den durch a_value bestimmten Wert */
void CNeuron::set_output(double a_value){
    output=a_value;
    output_set=true;
}

//-----------------------------------------------------------------------------
/** setzt den Fehlerwert (error_activation) fr das Neuron */
void CNeuron::set_error(TNErrorValue error){
    if (fabs(error)>min_error){
        error_activation=error;
    }else{
        error_activation=0;
    }
}	
//-----------------------------------------------------------------------------
/** liefert den Fehler den das Neuron ausweist.
Der Fehler ist die Fehleraktivierung
Die Fehleraktivierung ergibt sich normalerweise aus dem Vorhersagefehler (soll-ist) falls das Neuron in der Outputschicht ist.
Und falls das Neuron in einer Hidden Layer ist aus der Summe von get_error_output()der Nachfolgeneuronen*Gewicht der Synapse zu den Nachfolgerneuronen */
TNErrorValue CNeuron::calc_error_output(){
    error_output=error_activation;
    old_error_activation=error_activation;
    error_activation=0;
    return error_output;
}
//-----------------------------------------------------------------------------
/** liefert den letzten festgestellten error_output ('ohne ihn neu zu ermitteln) */
TNErrorValue CNeuron::get_last_error_output(){
    return error_output;
}
//-----------------------------------------------------------------------------
/** liefert den (letzten) Fehler des Neurons. D.h. der vor dem letzten Feedback-Schritt. */
TNErrorValue CNeuron::get_last_error(){
    return old_error_activation;
}

//-----------------------------------------------------------------------------
/** setzt die Aktivierung des Neurons auf einen bestimmten Wert */
void CNeuron::set_activation(TNPostSynapticValue a_activation){
    activation=a_activation;
}
//-----------------------------------------------------------------------------
/** liefert die aktuelle Aktivierung des Neurons */
TNPostSynapticValue CNeuron::get_activation(){
    return activation;
}

//-----------------------------------------------------------------------------
/** die aktivierungsfunktion fr das neuron,standardmig wird einfach activation als Wert zurckgegeben*/
TNPostSynapticValue CNeuron::activation_function(TNActivationValue activation)
{
   switch (act_activation_function)
   {
       case AF_LOGISTIC:
       {   //die logistische Funktion
           return(  (1.0/(1.0+exp(-slope*activation))) );
           //SNNS verwendet hier: (1.0 / (1.0 + exp_s( -activation - bias )))) );
       }break;
       case AF_LOGISTIC_SNNS:
       {   //die logistische Funktion wie im SNNS verwendet:
           if (activation > 88.72) return( MAXFLOAT );
           else if (activation < -88.0) return( 0.0 );
           return( (1.0 / (1.0 + exp( -activation ))));
       }break;
       case AF_TANGENSH:{
           //TangensHyperbolicus
           return tanh(slope*activation);
           //SNNS verwendet hier: tanh(activation+bias);
           //die alte Variante
           //double e_hoch_x=exp(slope*activation);
           //double e_hoch_minus_x=exp(-slope*activation);	
           //return (e_hoch_x-e_hoch_minus_x)/(e_hoch_x+e_hoch_minus_x);
       }break;		
       case AF_IDENT:{
           //Rckgabewert ist identisch der Aktivierung (mit Faktor slope)
           return slope*activation;
       }break;
       default:      
           throw_error(666,"Unknown activation function"," CNeuron::activation_function");
   }
   return 0; //dem Compiler zuliebe
}
//-----------------------------------------------------------------------------
/** Die Ableitung der Aktivierungsfunktion */
double CNeuron::derivative_activation_function(double x)
{
    switch (act_activation_function)
    {
        case AF_LOGISTIC:
        case AF_LOGISTIC_SNNS:{
            double af=activation_function(x);
            return af*(1.0-af);
            //SNNS verwendet hier:  return( GET_UNIT_ACT( unit_ptr ) * (1.0 - GET_UNIT_ACT( unit_ptr )) );
            //anscheinend ist GET_UNIT_ACT( unit_ptr ) gleich dem output, also aktivierungsfunktion und bias schon bercksichtigt
        }break;
        case AF_TANGENSH:{
            double af=activation_function(x);
            //return slope*(1-(af*af)); =falsch
            return (1.0-(af*af));
            //SNNS verwendet hier:    return(1.0-GET_UNIT_ACT( unit_ptr ) * (GET_UNIT_ACT( unit_ptr )) );
        }break;
        default:
            throw_error(666,"Unknown activation function","CNeuron::derivative_activation_function(double x)");
    }
    return 0; //dem Compiler zuliebe
}


//-----------------------------------------------------------------------------
/** liefert den minimalen Wert, den die Aktivierungsfunktion liefern kann
  * bei der Ident-Aktivierungsfunktion kann es natrlich keinen echten min- oder max-Wert geben*/
double CNeuron::get_min_output(){
   switch (act_activation_function){
       case AF_LOGISTIC: 
       case AF_LOGISTIC_SNNS: return 0;
       case AF_TANGENSH: return -1;
       case AF_IDENT: assert(false);//nur zu DEBUG-Grnden um sicherzustellen, dass es nicht aufgerufen wird
                      return -MAXDOUBLE;
       default:
          //idealerweise sollte per switch-Anweisung die Grenzwerte vorgegeben werden
          //aber es reicht auch das funktionsergebnis fr einen sehr grossen Wert zu ermitteln
          return activation_function(-500);
   }
}
//-----------------------------------------------------------------------------
/** liefert den maximalen Wert, den die Aktivierungsfunktion liefern kann
  * bei der Ident-Aktivierungsfunktion kann es natrlich keinen echten min- oder max-Wert geben*/
double CNeuron::get_max_output(){
   switch (act_activation_function){
       case AF_LOGISTIC:
       case AF_LOGISTIC_SNNS: 
       case AF_TANGENSH: return 1;
       case AF_IDENT: return MAXDOUBLE;
       default:
          //idealerweise sollte per switch-Anweisung die Grenzwerte vorgegeben werden
          //aber es reicht auch das funktionsergebnis fr einen sehr grossen Wert zu ermitteln
          return activation_function(500);
   }
}

//-----------------------------------------------------------------------------
/** erzeugt eine Synapse, die mit dem anderen Neuron verbunden ist */
CSynapse* CNeuron::create_input_synapse(CNeuron* other,TNSynapticWeight weight){
    assert(other);
    return new CSynapse(other,this,weight);
}
//-----------------------------------------------------------------------------
/**  liefert den Vektor der Inputgewichte des Neurons */
void CNeuron::get_weight_vector(CDoubleVector& vec){
    vec.resize(0);
    for (long i=0; i<get_anzahl_inputs(); i++){
        vec.push_back(input_synapse(i)->get_weight());
    }
}
//-----------------------------------------------------------------------------
/** falls das die Funktion mit Argument true aufgerufen wird,
wird das Lernen fr dieses Neuron erlaubt, ansonsten nicht. */
void CNeuron::allow_learning(bool allow){
    learning_allowed=allow;
}
//-----------------------------------------------------------------------------
void CNeuron::allow_feedback(bool allow){
    feedback_allowed=allow;
}

//-----------------------------------------------------------------------------
#ifdef USE_CCOMMUNICATION
/** ausfhren eines Kommandos*/
bool CNeuron::process_cmd(const char* a_command, const void* data,TPtrVector* add_data=0,long id=0)
{
    if (check_command(a_command,"learning_rate","bestimmt die Lernrate des Neurons",id) ||
      check_command(a_command,"lr","bestimmt die Lernrate des Neurons",id)){
        if (data){
            double wert=get_doub(data,add_data);
            learning_rate=wert;
        }else infos(">>>");
        infos("Lernrate ist nun :");
        infos(learning_rate);
        infos("\n");
        return true;
    }
    if (check_command(a_command,"learning_rate_goal","bestimmt den Wert, auf den die Lernrate konvergieren soll.",id) ||
        check_command(a_command,"lr_goal","siehe learning_rate_goal",id))
    {
        double wert=get_doub(data,add_data);
        learning_rate.set_goal(wert);
        infos("Konvergenzziel fr die Lernrate ist nun :");
        infos(wert);
        infos("\n");
        return true;
    }
    if (check_command(a_command,"learning_rate_factor","Der Faktor mit dem sich die Lernrate dem Konvergenzziel nhert. (blicherweise >0.99 und <1)",id) ||
      check_command(a_command,"lr_factor","siehe learning_rate_factor",id)){
          double wert=get_doub(data,add_data);
          learning_rate.set_change_factor(wert);
          infos("Der Faktor fr die Lernrate ist nun :");
          infos(wert);
          infos("\n");
          return true;
    }
    if (check_command(a_command,"slope","Der Anstieg der Aktivierungsfunktion des Neurons.",id) ||
        check_command(a_command,"sl","siehe slope",id))
    {
        if (data){
            slope=get_doub(data,add_data);
        }else infos(">>>");
        infos("Der Anstieg ist nun :");
        infos(slope);
        infos("\n");
        return true;
    }
    if (check_command(a_command,"min_error","Fehler kleiner als min_error, werden auf 0 gesetzt. D.h. fr diese Werte erfolgt kein Lernen mehr.(Vermeidung von Overtraining)",id) ||
        check_command(a_command,"me","siehe min_error",id))
    {
        double wert=get_doub(data,add_data);
        min_error=(wert);
        infos("min_error ist nun :");
        infos(wert);
        infos("\n");
        return true;
    }
    if (check_command(a_command,"output_weights","Setzt die Gewichte der ausgehenden Synapsen auf einen zuflligen Wert zw. min und max.(Standard -1,1)",id) ||
        check_command(a_command,"ow","siehe output_weights",id))
    {
        if (!data){
            //Standardwerte verwenden
            set_output_weights();
        }else{
            double min,max;
            get_doub(data,add_data,&min,&max);
            assert( min<=max);
            set_output_weights(min,max);
        }
        infos("Gewichte gendert.");
        infos("\n");
        return true;
    }
    if (check_command(a_command,"activation_function","Setzt die Aktivierungsfunktion. (tan-tanH, log-Logist. Akt.ftk.)",id) ||
        check_command(a_command,"af","siehe activation_function",id))
    {
        if (data){
            if (!strcmp((char*) data,"tan")) set_activation_function(AF_TANGENSH);
            else if (!strcmp((char*) data,"log")) set_activation_function(AF_LOGISTIC);
            else set_activation_function(EActivationFunction(string_to_long((char*) data)));
        }
        //Standardwerte verwenden
        infos("Activation funtion: ");
        infos(act_activation_function);
        infos("\n");
        return true;
    }
    // die Parentklasse wei hoffentlich was damit anzufangen
    return CCommunication::process_cmd(a_command,data,add_data,id);
}
#endif
//-----------------------------------------------------------------------------
/** gibt zu einem Neuron dessen inputgewichte (als Zeilenvektor) aus */
void CNeuron::print_input_vector(){
    CDoubleVector vec;
    get_weight_vector(vec);
    cout.setf(ios::fixed);
    cout<<setprecision(2)<<vec;
}
#ifdef USE_LIBXML
//-----------------------------------------------------------------------------
/** speichert in das XML-File, zu dem tree gehrt */
xmlNodePtr CNeuron::save_to_xml(xmlNodePtr tree){
    assert(tree);
    //einen neuen Knoten anlegen (mit dem Klassennamen)
    xmlNodePtr subtree;
    subtree = xmlNewChild(tree, NULL, (const xmlChar *)"CNeuron",NULL);
    xmlSetProp(subtree, (const xmlChar *)"type",(const xmlChar *) "neuron");	//es handelt sich um ein Element vom Typ Neuron
    xmlSaveDouble(subtree,"activation",activation);
    xmlSaveLong(subtree,"act_activation_function",act_activation_function);
    xmlSaveBool(subtree,"feedback_allowed",feedback_allowed);
    xmlSaveBool(subtree,"learning_allowed",learning_allowed);
    xmlSaveCProp(subtree,"learning_rate",learning_rate);
    xmlSaveDouble(subtree,"old_activation",old_activation);
    xmlSaveDouble(subtree,"output",output);
    xmlSaveBool(subtree,"output_set",output_set);
    xmlSaveDouble(subtree,"schwelle",schwelle);
    xmlSaveDouble(subtree,"slope",slope);
    xmlSaveDouble(subtree,"min_error",min_error);
    return subtree;
}
//-----------------------------------------------------------------------------
/** ldt aus dem XML-File, zu dem tree gehrt */
void CNeuron::load_from_xml(xmlNodePtr tree){
    assert(tree);
    //in den Knoten gehen (mit dem Klassennamen)
    if (strcmp((const char*)tree->name, "CNeuron")) throw(666);
    xmlNodePtr subtree;
    subtree=tree->xmlChildrenNode;
    int i;
    while (subtree!=0){
        if (xmlLoadInt(subtree,"act_activation_function",i)) act_activation_function=(EActivationFunction)i; ;
        xmlLoadDouble(subtree,"activation",activation);
        xmlLoadBool(subtree,"feedback_allowed",feedback_allowed);
        xmlLoadBool(subtree,"learning_allowed",learning_allowed);
        xmlLoadCProp(subtree,"learning_rate",learning_rate);
        xmlLoadDouble(subtree,"old_activation",old_activation);
        xmlLoadDouble(subtree,"output",output);
        xmlLoadBool(subtree,"output_set",output_set);
        xmlLoadDouble(subtree,"schwelle",schwelle);
        xmlLoadDouble(subtree,"slope",slope);
        xmlLoadDouble(subtree,"min_error",min_error);
        subtree=subtree->next;
    }
}
#endif


//-----------------------------------------------------------------------------
/** setzt die Gewichte der Output-Synapsen auf einen zuflligen Wert zwischen  min und max */
void CNeuron::set_output_weights(TNSynapticWeight min_weight=-1, TNSynapticWeight max_weight=1){
    assert(min_weight<=max_weight);
    for (long i=0; i<get_anzahl_outputs();i++){
        output_synapse(i)->set_random_weight(min_weight,max_weight);
    }
}
//-----------------------------------------------------------------------------
/** legt die Aktivierungsfunktion fr das Neuron fest. Achtung dies sollte normalerweise 
vor dem Trainieren des Netzes geschehen (beim Erstellen). Wenn fr eine Output-
Schicht eine Skalierung gesetzt werden soll, so muss darauf geachtet werden,
in welchem Bereich die Aktivierungsfunktion Werte liefert. Zum Beispiel 0..1 fr 
AF_LOGISTIC und -1..1 fr AF_TANH */
void CNeuron::set_activation_function(EActivationFunction act_func, bool set_default_params=false){
      act_activation_function=act_func;
      if (set_default_params){
          switch (act_func){
              case AF_TANGENSH: learning_rate=0.5;  break;
              case AF_LOGISTIC:
              case AF_LOGISTIC_SNNS: learning_rate=0.7;  break;
              default: ;//nichts ndern;
          }
      }
}
