/***************************************************************************
                          ccommunication.cpp  -  description
                             -------------------
    begin                : Tue Jan 23 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
 /***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <typeinfo>             //fr typeid
#include "ccommunication.h"
#include "tools.h"              //fr CString


CCommunication::CCommunication(){
    showinfo=true;
}
CCommunication::~CCommunication(){
}
bool CCommunication::global_showinfo=true;
/** stellt eine Procedur zum Verarbeiten von Befehlen zur
Verfgung. Muss (bei Bedarf) in den abgeleiteten Klassen berschrieben
werden. command enthlt den Befehl (als String),data enthlt einen Zeiger auf 
die bergebenen Daten (id dient zur Identifikation von Antworten) */
bool CCommunication::process_cmd(const char* command,const void* data,TPtrVector* additional_data=0,long id=0){
    CString a_command=command;
    if (a_command=="list_commands" || a_command=="?"){
        infos("\nListe der verfgbaren Kommandos : \n=================================================\n\n");
        //Nochmal das Kommando an alle Schicken, aber mit der ID -1, so dass Check_command die
        //Erklrungen ausgibt
        process_cmd(" "," ",0,ID_HELP);
        infos("list_commands:\nListet die verfgbaren Kommandos auf.\n\n");
    }
    if ((id!=ID_HELP) && (id!=ID_NO_ERROR_MSG)){  //abchecken aller Kommandos -> Ausgabe unterdrcken
        infos("Kein Empfnger gefunden fr : ");
        infos(command);
        infos("\n");
    }
    return false;
}

/** wie @see process_cmd(command,data,...) nur dass data auf eine double-Zahl festgelegt ist
    */
bool CCommunication::process_command(const char* cmd, double data){
    return process_cmd(cmd,&data);
}


/** wie process_command(command,data) nur dass data als Teil des Kommandos bergeben wird*/
bool CCommunication::process_command(const char* orig_cmd){
    char cmd[2000];
    strcpy(cmd,orig_cmd); //cmd wird durch strtok zerstrt
    char* command;
    char* data;
    char* ddata1;
    char* ddata2;
    char* str_data;
    double double_data1;
    double double_data2;

    command=strtok(cmd, "=");
    data=strtok(NULL, "=");

    if (data){  //ein Parameter wurde mit bergeben
        if (data[0]=='"')
        { //es ist ein Stringwert
        	str_data=data+1;
        	long pos=strlen(str_data)-1;
        	if (str_data[pos]=='"') str_data[pos]=0;
        	return process_cmd(command,str_data);
        }else
        { //ein Doublewert (oder 2)
            ddata1=strtok(data,":");
            ddata2=strtok(NULL,":");
            if (ddata2){ //es sind 2 Doublewerte
            	double_data1=atof(ddata1);
            	double_data2=atof(ddata2);
            	TPtrVector tmp;
            	tmp.push_back(&double_data2);
            	return process_cmd(command,&double_data1,&tmp);
            }else{	//nur 1 Doublewert (2mal den selben Wert bergeben)
            	double_data1=atof(ddata1);
            	return process_cmd(command,&double_data1);
            }
        }
    }else{
        return process_cmd(cmd,0);
    }
}



/** zeigt den Text an, falls showinfo=true ist*/
void CCommunication::infos(const CString& info_text){
    CString tmp(info_text);	//da sonst ohne abschlieendes Leerzeichen
    tmp.push_back(0);
    if (showinfo && global_showinfo) cout <<tmp;
}
/** zeigt den Text an, falls showinfo=true ist */
void CCommunication::infos(const double info_text){
    if (showinfo && global_showinfo) cout <<info_text;
}

/** liefert true, falls to_check mit command bereinstimmt
  * wird in process_cmd verwendet*/
bool CCommunication::check_command(const char* to_check, const char * command, const char* description, long id=0){
    if (id==ID_HELP){        //hier wird id missbraucht um eine Beschreibung zu dem Befehl auszugeben
        infos(command);
        infos(":\t");
        infos(description);
        infos("\n");
        return false;
    }else{
        int to_check_len=strlen(to_check);
        int command_len=strlen(command);
        if (to_check_len<command_len) return false;
        long i;
        for (i=0; i<command_len;i++){
            if (to_check[i]!=command[i]) return false;
        }
        //ok bis hierhin stimmt alles berein wenn jetzt noch das folgende Zeichen eine Klammer oder 0 oder ein Punkt  ist
        //haben wir gewonnnen
        if (to_check[i]==0){
            infos(command);
            infos(": ");
            return true;
        }
        if (to_check[i]=='(')
        {
            infos(command);
            return true;
        }
        if (to_check[i]=='.')
        {
            infos(command);
            return true;
        }
        else return false;
    }
}

/** prft den Befehl command_to_check welches 
Kommando weitergegeben werden soll und 
an welche Elemente
Start und stop geben das Anfangs und Endelement zurck.
In min und max  wird die nummer des kleinsten und grten elements bergeben.
In new_command steht das weiterzugebende Kommando. */
bool CCommunication::parse_command(const char * command_to_check,char*& new_command,long& start,long& stop,long min, long max){
    char* tmp_str;
    char* tmp_str2;
    tmp_str2=strchr(command_to_check,'.');
    tmp_str=strchr(command_to_check,'(');
    if ((!tmp_str) || (strlen(tmp_str)<strlen(tmp_str2)) ){  //ein Punkt steht vor der Klammer
        //keine Klammer enthalten, also an alle weiterreichen
        //den Punkt finden
        if (!tmp_str2){
            return false; //Fehlerhaftes Kommando
        }
        start=min; stop=max;
        new_command=&tmp_str2[1];
        infos(format_string("(%d).",start));
        //infos("(*).");
        return true;
    }
    if (tmp_str[1]=='*'){
        //Alle ausgeben		beispiel(*).test
        start=min; stop=max;
        new_command=&tmp_str[4];
        infos(format_string("(%d).",start));
        //infos("(*).");
        return true;
    }
    for (unsigned long i=2; i<strlen(tmp_str)-2;i++){
        if (tmp_str[i]==')'){
            char zahl_str[20]="";
            strncpy(zahl_str,&tmp_str[1],  i-1);
            start=atoi(zahl_str);
            stop=start;
            if (start<min || stop>max){
                infos("\nFehlerhafter Breich: ");
                return false;
            }
            new_command=&tmp_str[i+2];
            infos("(");
            infos(start);
            infos(").\n");
            return true;
        }
    }
    return false;
}


/** Hilfsfunktion, die zu data und add_data den Wert ermittelt, der verwendet werden soll
		*/
double CCommunication::get_doub(const void* data,TPtrVector* add_data,double *min_val=0,double* max_val=0){
    double erg;
    double min,max;
    if (!data) throw_error(666,"Keine Daten bergeben.", "CCommunication::process_command");
    erg=(*reinterpret_cast<const double*>(data));
    //Sonderflle (zustzliche Daten) abchecken
    if (add_data){
      if ((*add_data).size()>0){
        if ((*add_data)[1]){
        	//wenn in add_data[0] Daten stehen, gibt *data die Untergrenze, *add_data[0] die Obergrenze an
        	//und der Wert ergibt sich aus einem Zufallswert zwischen den beiden Grenzen
        	TPtrVector tmp=*add_data;
        	max=(*reinterpret_cast<const double*>(tmp[0]));
        	min=erg;
        	erg=get_random_value(min,max);
        }
      }
    }else{ //keine zustzlichen Daten fertig
        min=max=erg;
    }
    //Rckgabe der Min/Max-Werte
    if (min_val) *min_val=min;
    if (max_val) *max_val=max;
    return erg;
}
