/***************************************************************************
                          cdoublevector.cpp  -  description
                             -------------------
    begin                : Tue Jun 20 2000
    copyright            : (C) 2000 by Jon Hennig
    email                : 
 ***************************************************************************/

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +                                                                         +
// +   This program is free software; you can redistribute it and/or modify  +
// +   it under the terms of the GNU General Public License as published by  +
// +   the Free Software Foundation; either version 2 of the License, or     +
// +   (at your option) any later version.                                   +
// +                                                                         +
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++/

#include "cdoublevector.h"
#include <values.h> //for MAXDOUBLE
#include <iomanip>  //for setw()

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// CDoubleVector
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

CDoubleVector::CDoubleVector(){
}
CDoubleVector::CDoubleVector(basic_string<char> str,basic_string<char> delimiter=" " ){
    long	start=0;
    long	ende=0;
    double dou;
    while(str.length()>0 && start>=0){
        start=str.find_first_not_of(' ');		//erstes -Nicht-Leerzeichen finden
        if (start>=0){
            ende=str.find(delimiter,start);
            if (ende<0) ende=str.length();
            if (start<ende){ //es wurde etwas zwischen 2 Leerzeichen gefunden
                dou=strtod(str.substr(start,ende).c_str(),0);
                push_back(dou);
            }
            str=str.substr(ende);
        }
    }
}
CDoubleVector::~CDoubleVector(){
}


CDoubleVector operator-(CDoubleVector& x,CDoubleVector& y){
	assert(x.size()==y.size());
	
	CDoubleVector ergebnis;
	for (long i=0; i<(long)x.size();i++)
	{
		ergebnis.push_back(x[i]-y[i]);
	}
	
	assert(x.size()==ergebnis.size());
	return ergebnis;
}


CDoubleVector operator+(CDoubleVector& x,CDoubleVector& y)
{
	assert(x.size()==y.size());
	
	CDoubleVector ergebnis;
	for (long i=0; i<(long)x.size();i++)
	{
		ergebnis.push_back(x[i]+y[i]);
	}
	
	assert(x.size()==ergebnis.size());
	return ergebnis;
}

CDoubleVector operator*(CDoubleVector& x,double l)
{
	CDoubleVector ergebnis;
	for (long i=0; i<(long)x.size();i++)
	{
		ergebnis.push_back(x[i]*l);
	}
	
	assert(x.size()==ergebnis.size());
	return ergebnis;
}


bool operator<(CDoubleVector& x,CDoubleVector& y)
{
	assert(x.size()==y.size());
	
	bool ergebnis=true;
	for (long i=0; i<(long)x.size();i++)
	{
		if (x[i]>=y[i]) ergebnis=false;
	}
	return ergebnis;
}



ostream& CDoubleVector::write(ostream& stream,long width=0,char delimiter=' '){
  if (size()==0)
      return stream<<"<empty>";
  else
      return write_partially(stream,0,size()-1,width,delimiter);
}


ostream& CDoubleVector::write_partially(ostream& stream,long from, long to, long width=0,char delimiter=' '){
	assert(from >=0);
	assert(to<(long)size());
	assert(to>=from);
	
	long i;
	for ( i=from;i<to;i++)
	{
		stream<<setw(width)<< (*this)[i] <<delimiter;
	}
	return stream<<setw(width)<<(*this)[i];
}



ostream& operator<<(ostream& stream,CDoubleVector& x)
{
	return x.write(stream,7,',');
}

istream& operator>>(istream& stream,CDoubleVector& x)
{
	double dou;
	x.resize(0);
	while (stream>>dou){
		x.push_back(dou);
	}
	return stream;
}

/** dividiert jedes Element des Vektors durch den
angegebenen Wert */
void CDoubleVector::divide(double value){
    for (ulong i=0; i<this->size(); i++){
        (*this)[i]/=value;
    }
}

/** multipliziert jedes Element des Vektors mit dem 
  *angegebenen Wert */
void CDoubleVector::multiply(double value){
    for (ulong i=0; i<this->size(); i++){
        (*this)[i]*=value;
    }
}
/** addiert zu dem Vector den anderen Vektor (skalar) hinzu */
void CDoubleVector::add(CDoubleVector& other)
{
    assert(other.size()==this->size());
    for (ulong i=0; i<this->size(); i++){
        (*this)[i]+=other[i];
    }	
}
/** multipliziet den Vector mit dem anderen Vektor (skalar) */
void CDoubleVector::multiply(CDoubleVector& other)
{
	assert(other.size()==this->size());
	
	for (ulong i=0; i<this->size(); i++)
	{
		(*this)[i]*=other[i];
	}	
}
/** addiert value zu jedem Element des Vektors hinzu */
void CDoubleVector::add(double value)
{
	for (ulong i=0; i<this->size(); i++)
	{
		(*this)[i]+=value;
	}	
}

void CDoubleVector::push_back_doublevector(const CDoubleVector &vec){
	for (ulong i=0; i<vec.size(); i++) push_back(vec[i]);
}



double CDoubleVector::distance_to(CDoubleVector &other, int type){
    assert (other.size()==this->size());
    double val=0;
    for (ulong i=0; i<other.size();i++)
    {
        val+=fabs((*this)[i]-other[i]);
    }
    return val;
}


/** liefert das Element des Vektors, der die minimale Differenz zu dem Wert (diff_value) besitzt */
ulong CDoubleVector::elem_min_diff(double diff_value){
  double min;
  long min_elem;
  min = MAXDOUBLE;
  min_elem=0;
  for (ulong i=0;i<size();i++){
      if (fabs((*this)[i]-diff_value)<min){
          min=fabs((*this)[i]-diff_value);
          min_elem=i;
      }
  }
  return min_elem;
}


//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// CDoubleVectorVector
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

CDoubleVector CDoubleVectorVector::get_avg(){
    CDoubleVector ergebnis;
    if (size()<1) throw(666); //,"Vector with no values. Operation impossible","CDoubleVectorVector::get_avg");
    long noof_elems=(*this)[0].size();
    if (noof_elems<1) throw(666); //,"Vector with no values. Operation impossible.","CDoubleVectorVector::get_avg");
    ergebnis.resize(noof_elems);
    for (ulong i=0;i<size();i++){
        ergebnis+=(*this)[i];
    }
    //Achtung diese Art der Berechnung kann zu Rundungsfehlern durch
    //den beschrnkten Zahlenbereich/Auflsung fhren, wenn der Durchschnitt
    //ber viele Werte gebildet werden soll:
    ergebnis/=size();
    return ergebnis;
}








//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// CLongSet
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


/** entfernt alle Werte aus values_to_remove aus dem aktuellen Vektor*/
void CLongSet::remove(CLongSet& values_to_remove)
{
  //CLongSet::iterator this_it;
	for (CLongSet::iterator it=values_to_remove.begin();it!=values_to_remove.end();it++){
		erase(*it);
	}
}

/** fgt alle Werte aus values_to_add zu dem aktuellen Set hinzu */
void CLongSet::add(CLongSet& values_to_add)
{
	for (CLongSet::iterator it=values_to_add.begin();it!=values_to_add.end();it++){
			insert(*it);
	}
}


ostream& operator<<(ostream& stream,CLongSet& x)
{
  for (CLongSet::iterator it=x.begin();it!=x.end();it++)
  	stream <<*it<<", ";
  return stream;//<<(*it);
}

