/***************************************************************************
                          cinputneuron.cpp  -  description
                             -------------------
    begin                : Wed Jun 21 2000
    copyright            : (C) 2000 by Jon Hennig
    email                :

    Input-Neuron:
    Standardmig werden dort die Input-Werte eingegeben.
    Die Neuronen knnen aber auch mit einem "eingebetteten" Neuron verbunden
    werden. Dann werden die Outputs von denen der "eingebetteten" Neuronen
    verwendet.
    add_excitation - Aufrufe werden an die "eingebetteten" Neuronen weitergeleitet
    feed_back-Aufrufe werden nicht weitergeleitet, die mssen fr die eingebetteten
    Neuronen direkt aufgerufen werden.
    feed_forward() ruft calc_output() auf und damit werden die Outputs vom eingebetteten
    Neuron bernommen. Achtung: Der Output des eingebetteten Neurons muss schon be-
    rechnet sein.
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cinputneuron.h"

CInputNeuron::CInputNeuron()
{
	act_activation_function=AF_IDENT;  //Output entspricht der Aktivierung
	slope=1;	//Faktor 1
	schwelle=0;
}
CInputNeuron* CInputNeuron::TYP=0;
CInputNeuron::~CInputNeuron(){
}

#ifdef USE_LIBXML
//-----------------------------------------------------------------------------
/** ldt aus dem XML-File, zu dem tree gehrt */
void CInputNeuron::load_from_xml(xmlNodePtr tree){
	assert(tree);
	//in den Knoten gehen (mit dem Klassennamen)
	if (strcmp((const char*)tree->name, "CInputNeuron")) throw(666);
	xmlNodePtr subtree;
	subtree=tree->xmlChildrenNode;
	while (subtree!=0)
	{
		if (!strcmp((const char*)subtree->name,"CNeuron")) CNeuron::load_from_xml(subtree);
		subtree=subtree->next;
	}
}
//-----------------------------------------------------------------------------
/** speichert in das XML-File, zu dem tree gehrt */
xmlNodePtr CInputNeuron::save_to_xml(xmlNodePtr tree){
	assert(tree);
	//einen neuen Knoten anlegen (mit dem Klassennamen)
	xmlNodePtr subtree;
	subtree = xmlNewChild(tree, NULL, (const xmlChar *)"CInputNeuron",NULL);
	xmlSetProp(subtree, (const xmlChar *)"type",(const xmlChar *) "neuron");	//es handelt sich um ein Element vom Typ Neuron
	//Zuerst die Basisklasse abspeichern
	CNeuron::save_to_xml(subtree);
	return subtree;
}
#endif
//-----------------------------------------------------------------------------
/** fuegt der Error-aktivierung des gkapselten Neurons
den entsprechenden Wert hinzu, falls dieses existiert */
void CInputNeuron::add_error_activation(TNErrorValue error){
    if (encapsulated_nrn() && feedback_allowed){
       encapsulated_nrn()->add_error_activation(error);
    }else{
        CNeuron::add_error_activation(error);
    }
}
//-----------------------------------------------------------------------------
/** liefert einen Verweis auf das gekapselte Neuron */
CNeuron* CInputNeuron::encapsulated_nrn(){
	return input_neuron(0);
}
//-----------------------------------------------------------------------------
/** setzt den Output des Neurons, liest dazu den Output-Wert des gekpselten Neurons ein
	  * und speichert  ihn in output. Die Aktivierung wird nicht beachtet
	  ACHTUNG: calc_output wird nur von feed_forward aufgerufen, wenn das Output nicht gesetzt wurde.
 */
TNPreSynapticValue CInputNeuron::calc_output(){
	if (encapsulated_nrn()){
		output=encapsulated_nrn()->get_last_output();
	}else{
		//sonst gar nichts machen. fr die Input-Neuronen sollte immer das Output gesetzt werden -
		//nicht die Aktivierung
		//set_activation ist fr die Input-Neuronen daher auch nicht erlaubt.
		//Um die Funktion nicht virtuell def. zu mssen wird hier getestet, ob die Aktivierung <>0 ist
		assert(activation==0);
	}
	return output;
	//folgendes entsprche der implementierung in CNeuron -> kann aber Probleme geben, z.B. bei layer(*).neuron(*).slope=xxx
//	if (encapsulated_nrn()){
//		output=activation_function(encapsulated_nrn()->get_last_output());
//		old_activation=activation;
//		activation=0;
//		return output;
//	}else{
//		return CNeuron::calc_output();
//	}
}
//-----------------------------------------------------------------------------
/** erzeugt eine Synapse, die mit dem anderen Neuron verbunden ist
stellt sicher, dass max. eine solche Verbindung ex. */
CSynapse * CInputNeuron::create_input_synapse(CNeuron * other, TNSynapticWeight weight){
	if (get_anzahl_inputs()!=0) throw 666; //fr diesen Neuronentyp ist max. 1  input-Neuron (das eingebettete) erlaubt
	return CNeuron::create_input_synapse(other,0);
}
