/* **************************************************************************
                          cinputneuron.h  -  description
                             -------------------
    begin                : Wed Jun 21 2000
    copyright            : (C) 2000 by Jon Hennig
    email                : 
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CINPUTNEURON_H
#define CINPUTNEURON_H

#include "cbasicneuron.h"

/** realisiert ein einfaches Input-Neuron, das als Output immer den mit set_activation gesetzten Wert liefert.
		Das heit in dem Neuron wird einfach die Netzeingabe (bzw. ein Teil davon) abgespeichert.
		
		Das Input-Neuron wurde implementiert, um den modularen Charakter der Bibliothek zu erhalten.
		Normalerweise knnte das Input-Neuron	auch eingespart, und die Werte direkt in das Nachfolger-Neuron
		eingelesen werden. Das Nachfolger-Neuron kann/soll aber nicht wissen, ob das Signal, das es
		erhlt von auen, oder von einem anderen Neuron kommt.
		
		Ein Input-Neuron kann auch ein anderes Neuron kapseln. In diesem Fall wird der Output des Input-Neurons
		dem des anderen Neurons gleichgesetzt. Diese Kapselung ist dann wichtig, wenn verschiedene Netze zusammmengeschaltet
		werden sollen, wobei jedes Teilnetz seine Struktur beibehalten soll (z.B. zum spteren getrennten abspeichern).
		Das wird zur Zeit in @ref CNeuronicMultiNet verwendet.
		
    <b> TODO: </b>
	  <li> spezielle Funktionen zum Kapseln eines Neurons (zur Zeit nicht explitzit mgl. siehe @ref CNeuronicMultiNet)
	  <li> normalerweise sollten activation und output immer die selben Werte beinhalten.
	  <li> calc_output sollte output nicht berechnen,sondern im Standardfall gar nichts machen
	  </li>
	
	  <b> BUGS: </b>
	  <li> keine bekannt
	  </li>

    @author Jon Hennig
  */

class CInputNeuron : public CNeuron  {
public: 
	CInputNeuron();
	virtual ~CInputNeuron();
	
  /** liefert den minimalen Wert, den die Aktivierungsfunktion liefern kann
			der Wert wird verwendet, um die Skalierung einzustellen.
			Da zu groe Inputs nicht gut sind, wird ein Wert von 1 als Max-Wert vorgegeben
			*/
  virtual double get_min_output() {return 0;};

  /** liefert den minimalen Wert, den die Aktivierungsfunktion liefern kann.
			Da zu kleine Inputs nicht gut sind, wird ein Wert von 0 als Min-Wert vorgegeben
			*/
  virtual double get_max_output() {return 1;};

  /** ldt die Daten zu dem Input-Neuron aus dem XML-File, zu dem tree gehrt
  		*/
  virtual void load_from_xml(xmlNodePtr tree);

	/** speichert in das XML-File, zu dem tree gehrt
			*/
	virtual xmlNodePtr save_to_xml(xmlNodePtr tree);
	
  /** liefert einen Verweis auf das gekapselte Neuron
  		@returns Zeiger auf das Neuron, bzw. 0 wenn nicht vorhanden.
  		*/
  CNeuron* encapsulated_nrn();

  /** fuegt der Error-aktivierung des gekapselten Neurons
			den entsprechenden Wert hinzu, falls dieses existiert und falls
      allow_feed_back true ist. Im anderen Fall wird die Aktivierung nur
      intern gesetzt.
			*/
  void add_error_activation(TNErrorValue error);

  /** setzt den Output des Neurons, liest dazu den Output-Wert des gekpselten Neurons ein
	    und speichert  ihn in output. Die Aktivierung wird nicht beachtet
	  	*/
  TNPreSynapticValue calc_output();

  /** diese Variable wird nur bentigt, um den Typ dieses Neurons in add_neurons von CNeuronicLayer
    	bestimmen zu knnen.
    	*/
  static CInputNeuron* TYP;

 protected:
  /** erzeugt eine Synapse, die mit dem anderen Neuron verbunden ist
			stellt sicher, dass max. eine solche Verbindung existiert. (= Implementation f. gekapseltes Neuron)
			*/
  CSynapse * create_input_synapse(CNeuron * other, TNSynapticWeight weight);

};

//extern CInputNeuron* N_INPUTNEURON;

#endif
