/* **************************************************************************
                          ckohonenlayer.h  -  description
                             -------------------
    begin                : Mon Feb 19 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CKOHONENLAYER_H
#define CKOHONENLAYER_H

#include "cneuroniclayer.h"

class CLongSet;


/** die verschiedenen Neighborhood-Kernel-Funktionen (=Distanzfunktionen)
    siehe Zell, "Simulation Neuronaler Netze", Kap. 15.2
    */
enum ENeighborhoodKernel		
{
	NK_CYLINDER,
	NK_CONE,
	NK_COS,
	NK_GAUSS1,
};

/** BETA:
		Schicht, implementiert eine selbstorganisierende Karte (Kohonen map).
		
		<b> TODO: </b>
	  <li> muss noch weiter getestet werden
	  <li> 3 Dim. relativ ungetestet
	  </li>
	
	  <b> BUGS: </b>
	  <li> Faktor nicht korrekt implementiert max 2 Abstnde lernen mit
	  </li>

   	@author Jon Hennig
  */
class CKohonenLayer : public CNeuronicLayer  {
public: 
	CKohonenLayer();
	~CKohonenLayer();
	
	/** liefert das Output der KohonenMap an die nachfolgenden Neuronen weiter
			*/
	void feed_forward(TNTimeUnit time);
	
	/** fhrt das Lernen durch. Auf Basis des Fehlers.
			*/
	void feed_back(TNTimeUnit time);
	
	/** liefert den Index des Winners der KohonenLayer, basierend auf den aktuellen
	  	Aktivierungswerten der Neuronen. Nach einem Feedback-Schritt ist ein Aufruf
	  	dieser Funktion ungltig, da die Aktualisierungen zurckgesetzt wurden.
	  	In diesem Fall msste man get_last_winner verwenden
		  */
	long get_winner();
	
	/** setzt @ref is_winner beim Gewinner auf true, auf false bei den brigen
			*/
	void set_winner();
	
  /** liefert den Abstand zwischen den beiden Neuronen
    	zur Zeit wird die oo-Norm verwendet
    	*/
  double get_abstand(long n1, long n2);

	#ifdef USE_CCOMMUNICATION
	  /** Ausfhren eines Kommandos. Folgende stehen zur Verfgung: <b>
	  		<li> "radius" ... bestimmt den Radius, um den andere Neuronen um das Gewinnerneuron mitlernen (Startwert)
	  		<li> "radius_goal" ... bestimmt den Wert, auf den der Radius konvergieren soll
	  		<li> "radius_factor" ... der Faktor mit dem sich der Radius dem Konvergenzziel nhert. (blicherweise >0.99 und <1)
	  		</li>
	  		@see CCommunication::process_cmd
		  */
	  virtual bool process_cmd(const char* command, const void* data,TPtrVector* add_data=0,long id=0);
  #endif	

  /** liefert die Nummer des Neurons, das beim letzten FeedForward-Schritt der Gewinner war
  		*/
  long get_last_winner();

protected:
  /** liefert zu dem Neuron nr die spalte
      in der entsprechenden Dimension des Gitters
      */
  long get_spalte(long nr,long dim=2);

  /** liefert eine Zahl zwischen 0 und 1, die angibt, wie stark
			ein entsprechendes Neuron im Abstand abstand mitlernt
			siehe Zell "Simulation Neuronaler Netze", Kap 15.2
			*/
  double neighborhood_kernel(double abstand, double max_radius);

	/** das letzte Gewinnerneuron (wird bei feed_forward() ermittelt - Abfrage mit get_last_winner)
			*/
	long winner;

  #ifdef USE_LIBXML
	  /** speichert in das XML-File, zu dem tree gehrt
	  		*/
  	virtual xmlNodePtr save_to_xml(xmlNodePtr tree);
    /** ldt aus dem XML-File, zu dem tree gehrt
    		*/
    void load_from_xml(xmlNodePtr tree);
  #endif

public:  //FIXME: besser schtzen ??
  /** die aktuell zu verwendende Funktion zur Bestimmung des Nachbarschaftskernels
  		*/
  ENeighborhoodKernel act_neighborhood_kernel;
	
	/** die Dimension der Map
			*/
	long dimension;
	/** breite der Map falls mind. 2-Dim. FIXME: unbegrenzte Dimensionen ermglichen
			*/
	long breite;
	/** tiefe der Map falls mind. 3-Dim. FIXME: unbegrenzte Dimensionen ermglichen
			*/
	long tiefe;
	
	/** Radius um den die Neuronen um den Gewinner mit lernen
			*/
	CChangingProperty radius;
};

#endif
