/***************************************************************************
                          ckohonenneuron.cpp  -  description
                             -------------------
    begin                : Mon Feb 19 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
 /***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "ckohonenneuron.h"

#include "ckohonensynapse.h"
#include "cdoublevector.h"

CKohonenNeuron::CKohonenNeuron()
{
	//die Standardwerte einstellen
	//Alle Fehler von den (vergleichenden) Synapsen werden aufsummiert (= oo-Norm)
	act_activation_function=AF_IDENT;
	slope=1;
	//Ziellernrate ist 0.1
	learning_rate=0.7;
	learning_rate.set_goal(0.1);
	learning_rate.set_change_factor(0.995);
	
	//eine Schwelle wird nicht verwendet
	schwelle=0;
}
CKohonenNeuron* CKohonenNeuron::TYP=0;
CKohonenNeuron::~CKohonenNeuron(){
}

/** erzeugt eine Synapse, die mit dem anderen Neuron verbunden ist
  * hier wird standardmig eine Kohonensynapse (CKohonensynapse) erzeugt*/
CSynapse* CKohonenNeuron::create_input_synapse(CNeuron* other,TNSynapticWeight weight)
{
	assert(other);
	return new CKohonenSynapse(other,this,weight);
}

/** erzeugt und liefert einen Doublevector mit dem letzten Inputvektor des Neurons*/
void CKohonenNeuron::get_last_input(CDoubleVector& vec)
{
	vec.resize(0);
	CKohonenSynapse* syn;
	for (long i=0; i<get_anzahl_inputs();i++)
	{	
		//FIXME: ist das der richtige Cast??
		syn=reinterpret_cast<CKohonenSynapse*>(input_synapse(i));
		vec.push_back(syn->get_last_input());
	}
}

//-----------------------------------------------------------------------------
TNPreSynapticValue CKohonenNeuron::calc_output()
{
	if (is_winner) output=1;
		else output=0;
	
	old_activation=activation;
	activation=0;
	return output;
}

//-----------------------------------------------------------------------------
/** bewegt das Neuron (genauer seine Gewichte) etwas in die gewnschte Richtung,
d.h. auf den vector vec zu. (gem. factor) */
void CKohonenNeuron::learn_step(CDoubleVector& vec,double faktor)
{
    if (learning_allowed){
        assert(vec.size()==inputs.size());
        for (long i=0; i<get_anzahl_inputs(); i++)
        {
        	double move=((vec[i]-input_synapse(i)->get_weight())*faktor*learning_rate);
        	input_synapse(i)->add_weight(move);
        }
        learning_rate.step();
    }
}
#ifdef USE_LIBXML
//-----------------------------------------------------------------------------
/** speichert in das XML-File, zu dem tree gehrt */
xmlNodePtr CKohonenNeuron::save_to_xml(xmlNodePtr tree){
	assert(tree);
	//einen neuen Knoten anlegen (mit dem Klassennamen)
	xmlNodePtr subtree;
	subtree = xmlNewChild(tree, NULL, (const xmlChar *)"CKohonenNeuron",NULL);
	xmlSetProp(subtree, (const xmlChar *)"type",(const xmlChar *) "neuron");	//es handelt sich um ein Element vom Typ Neuron
	//Zuerst die Basisklasse abspeichern
	CNeuron::save_to_xml(subtree);
	xmlSaveBool(subtree,"is_winner",is_winner);
	return subtree;
}
//-----------------------------------------------------------------------------
/** ldt aus dem XML-File, zu dem tree gehrt */
void CKohonenNeuron::load_from_xml(xmlNodePtr tree){
	assert(tree);
	//in den Knoten gehen (mit dem Klassennamen)
	if (strcmp((const char*)tree->name, "CKohonenNeuron")) throw(666);
	xmlNodePtr subtree;
	subtree=tree->xmlChildrenNode;
	while (subtree!=0)
	{
		if (!strcmp((const char*)subtree->name,"CNeuron")) CNeuron::load_from_xml(subtree);
		xmlLoadBool(subtree,"is_winner",is_winner);
		subtree=subtree->next;
	}
}
#endif
