/* **************************************************************************
                          ckohonenneuron.h  -  description
                             -------------------
    begin                : Mon Feb 19 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CKOHONENNEURON_H
#define CKOHONENNEURON_H

#include "cbasicneuron.h"

class CDoubleVector;

/** Neuron, dass als Inputsynapsen Kohonensynapsen (siehe @ref CKohonenSynapse) besitzt,
		das heit die an der Synapse ankommenden Signale werden nicht	mit dem Gewicht multipliziert,
		sondern mit dem Gewicht verglichen.
		Von der Synapse wird die Differenz zurckgegeben, und vom KohonenNeuron
		(entsprechend einer Norm) aufsummiert. Dadurch ergibt sich (unter Zuhilfenahme der Funktionen
		in @ref CKohonenLayer der Standard-Kohonen-Algorithmus.

		<b> TODO: </b>
	  <li> weiter testen
	  </li>
	
	  <b> BUGS: </b>
	  <li> ???
	  </li>		

	  @author Jon Hennig
  */
class CKohonenNeuron : public CNeuron  {
public: 
	CKohonenNeuron();
	~CKohonenNeuron();
	
	/** erzeugt eine Synapse, die mit dem anderen Neuron verbunden ist
  		hier wird standardmig eine Kohonensynapse (CKohonensynapse) erzeugt
  		*/
	CSynapse* create_input_synapse(CNeuron* other,TNSynapticWeight weight);
	
	/** liefert einen Doublevector mit dem letzten Inputvektor des Neurons
			*/
	void get_last_input(CDoubleVector& vec);
	
  /** bewegt das Neuron (genauer seine Gewichte) etwas in die gewnschte Richtung,
			d.h. auf den vector vec zu. (gem. factor)
			*/
  void learn_step(CDoubleVector& vec,double faktor);
		
	/** ist wahr, wenn das Neuron der Gewinner der Schicht ist (von CKohonenLayer od eq. gesetzt)
			*/
	bool is_winner;
	
	/** berechnet und legt die Ausgabe des Neurons fest
			*/
	virtual TNPreSynapticValue calc_output();
	
#ifdef USE_LIBXML
	  /** speichert in das XML-File, zu dem tree gehrt
	  		*/
  	virtual xmlNodePtr save_to_xml(xmlNodePtr tree);
  	/** ldt aus dem XML-File, zu dem tree gehrt
  			*/
  	void load_from_xml(xmlNodePtr tree);
#endif
	
	//Sonstiges
	//==================
  /** diese Variable wird nur bentigt, um den Typ dieses Neurons in add_neurons von CNeuronicLayer
    	bestimmen zu knnen
    	*/
  static CKohonenNeuron* TYP;
};

#endif
