/* **************************************************************************
                          ckohonensynapse.h  -  description
                             -------------------
    begin                : Thu Feb 15 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CKOHONENSYNAPSE_H
#define CKOHONENSYNAPSE_H

#include "csynapse.h"

/** ist eine Synapse, wie sie in einem Kohonenneuron verwendet wird.
    Als Output wird nicht der Input mal Gewicht (wie bei einer Standardsynapse)
    sondern der Betrag der Differenz zwischen Input und Gewicht zurckgegeben

    @see CKohonenNeuron
    @author Jon Hennig
  */

class CKohonenSynapse : public CSynapse  {
public: 
	/** erzeugt die Synapse als Verbindung zwischen den beiden Neuronen.
			@param weight wird als Gewicht der Verbindung eingestellt.
			*/
  CKohonenSynapse(CNeuron* pre_neuron, CNeuron* post_neuron, TNSynapticWeight a_weight);
	
	/** die einzige Funktion, die im Prinzip verndert werden muss.
	    als Ergebnis wird der Betrag der Differenz zwischen in und Output zurckgeliefert.
	    Dadurch erfolgt der fr den Kohonen-Algorithmus ntige Vergleich.
	    */
  virtual TNPostSynapticValue get_post_synaptic_value(TNPreSynapticValue a_input);

	/** liefert den letzen Input fr diese Synapse
  		das entspricht den letzten Output des Vorgngerneurons
  		ACHTUNG!! Das widerspricht etwas dem lokalen Charakter der brigen
  		Implementierung. Es sollte daher sichergestellt sein, dass zum Zeitpunkt
  		der Abfrage die vorhergehenden Neuronen die gleichen Werte haben
  		wie zum Zeitpunkt der Erregung der Synapse. Das ist fr die verwendeten Probleme der Fall.
  		Ansonsten msste der Wert einfach zwischengespeichert werden.
  		*/
	TNPreSynapticValue get_last_input();

};

#endif
