/***************************************************************************
                          cneuronica.cpp  -  description
                             -------------------
    begin                : Fri Sep 21 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
 /***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "cneuronica.h"
#include <fstream.h>
#include "tools.h" 	//format_string

#ifdef USE_LIBXML
	#include <libxml/parser.h>
#endif



CNeuronica::CNeuronica(){
}
CNeuronica::~CNeuronica(){
}
//------------------------------------------------------------------
/** erzeugt ausgehend von neuron_name ein neues Neuron*/
CNeuron* CNeuronica::new_neuron(const CString& neuron_name){
	if (neuron_name=="CBackpropNeuron"){
		return new CBackpropNeuron();
	}else
	if (neuron_name=="CInputNeuron"){
		return new CInputNeuron();
	}else
	if (neuron_name=="CStochasticBackpropNeuron"){
		return new CStochasticBackpropNeuron();
	}else
	if (neuron_name=="CKohonenNeuron"){
		return new CKohonenNeuron();
	}else
	return 0;
}
//------------------------------------------------------------------
/** erzeugt ausgehend von layer_name eine neue Schicht*/
CNeuronicLayer* CNeuronica::new_layer(const CString& layer_name){
	if (layer_name=="CNeuronicLayer"){
		return new CNeuronicLayer();
	}else
	if (layer_name=="CKohonenLayer"){
		return new CKohonenLayer();
	}else
	return 0;
}
//------------------------------------------------------------------
/** erzeugt ausgehend von layer_name und neuron_name eine neue Schicht mit anzahl Neuronen der entsprechenden Art*/
CNeuronicLayer* CNeuronica::new_layer(const CString& layer_name, const CString& neuron_name, long noof_neurons){
	CNeuronicLayer* lay;
	lay=new_layer(layer_name);
	if (lay){
	   	for (long i=1;i<=noof_neurons;i++){
				CNeuron* nrn = new_neuron(neuron_name);
				if (!nrn){
					delete lay;
					return 0;
				}
				lay->add_neuron(nrn);	
		  }
	}else lay=0;
	return lay;
}
//------------------------------------------------------------------
/** erzeugt ausgehend von layer_name eine neue Schicht*/
CNeuronicNet* CNeuronica::new_net(const CString& net_name){
	if (net_name=="CNeuronicNet"){
		return new CNeuronicNet();
	}else
	if (net_name=="CNeuronicMultiNet"){
		return new CNeuronicMultiNet();
	}else
	return 0;
}
/** erzeugt ausgehend von synaosen_name eine neue Synapse und verbindet
  * diese mit den Neuronen*/
CSynapse* CNeuronica::new_synapse(const CString& synapsen_name, CNeuron* pre_neuron,CNeuron* post_neuron){
	if (synapsen_name=="CSynapse"){
		return new CSynapse(pre_neuron,post_neuron,0);
	}else
	if (synapsen_name=="CKohonenSynapse"){
		return new CKohonenSynapse(pre_neuron,post_neuron,0);
	}else
	return 0;
}


#ifdef USE_LIBXML
  //------------------------------------------------------------------
  /** erzeugt ausgehend von der Beschreibung in dem XML-tree ein neues Neuron*/
  CNeuron* CNeuronica::new_neuron(xmlNodePtr tree){
  		CNeuron* neuron;
  		neuron=new_neuron((const char*)tree->name);
  		if (neuron){
  			neuron->load_from_xml(tree);
  		}
  		return neuron;
  }
  //------------------------------------------------------------------
  /** erzeugt ausgehend von der Beschreibung in dem XML-tree eine neue Schicht*/
  CNeuronicLayer* CNeuronica::new_layer(xmlNodePtr tree){
  		CNeuronicLayer* layer;
  		layer=new_layer((const char*)tree->name);
  		if (layer){
  			layer->load_from_xml(tree);
  		}
  		return layer;
  }
  //------------------------------------------------------------------
  /** erzeugt ausgehend von der Beschreibung in dem XML-tree eine neues Netz*/
  CNeuronicNet* CNeuronica::new_net(xmlNodePtr tree){
  		CNeuronicNet* net;
  		net=new_net((const char*)tree->name);
  		if (net){
  			net->load_from_xml(tree);
  		}
  		return net;
  }
  //------------------------------------------------------------------
  /** erzeugt ausgehend von der Beschreibung in dem XML-tree eine neue Synapse
    * und verbindet diese mit den Neuronen*/
  CSynapse* CNeuronica::new_synapse(xmlNodePtr tree,CNeuron* pre_neuron,CNeuron* post_neuron){
  		CSynapse* syn;
  		syn=new_synapse((const char*)tree->name,pre_neuron,post_neuron);
  		if (syn){
  			syn->load_from_xml(tree);
  		}
  		return syn;
  }
#endif

//-----------------------------------------------------------------------------
/**Erzeugt ein neues Netz entsprechend der Datei. Ldt das Netz in der Datei filename*/
CNeuronicNet* CNeuronica::load_new_net(const char* filename,  const char* format="DEFAULT"){
  assert(filename);
  assert(format);
   	
#ifdef USE_LIBXML
  	if (!strcmp(format, "NND")||!strcmp(format, "DEFAULT")){
      xmlDocPtr doc;
      xmlNodePtr tree;
      CNeuronicNet* nnet=0;

      doc= xmlParseFile(filename);
      if (doc){
        tree = doc->xmlChildrenNode;
        if (tree){
          tree = tree->xmlChildrenNode;		//NEURONICA
    			while (tree!=0 && nnet==0){
    				//Falls der Eintrag passt, wird das Netz geladen, ansonsten bleibt nnet==0 ..
    				nnet=CNeuronica::new_net(tree);
    				//.. und der nchste Eintrag wird betrachtet
    				tree=tree->next;
    	    }
        }
       	xmlFreeDoc(doc);
      }//else
      return nnet;
    }
#endif

  if (!strcmp(format, "SNNSNET")){
      CNeuronicSnnsNet* snet=new CNeuronicSnnsNet();
      snet->load(filename,format);
      //TODO: Implementieren
      return snet;
  }
  //else
  return 0;
}


//------------------------------------------------------------------
/** ldt aus einer Datei im SNNS-Pat-Format die In- und Outputs fr ein Netz. */
bool CNeuronica::load_patterns(const char* snns_pat_filename, std::vector< CDoubleVector> & inputs, std::vector< CDoubleVector> & outputs){
	inputs.resize(0);
	outputs.resize(0);
	//ffnen der Datei
	istream* in = &cin;
	if (snns_pat_filename!="-"){
		in = new ifstream(snns_pat_filename);
		if (!*in)
			throw 666; //Patternfile not found;
	}

	enum EPatternStates{ PS_START,PS_IN,PS_OUT}; //PS_IN =im Bereich der Input-Pattern, PS_Out im Bereich der Output Patterns
	EPatternStates state=PS_START;
	CString zeile;
	
	//Einlesen der Daten (bis zum bitteren Ende)
	while (getline(*in,zeile)){	
		//Einlesen der Allgemeinen Informationen
		if (state==PS_START){
			if (zeile.find("#")==0){ //Falls die Zeile mit # beginnt umschalten
 				if (zeile.find("nput")==NOT_FOUND) cout<<"warning: comment \"Input\" expected.\n";
			  state=PS_IN;
			  inputs.push_back(CDoubleVector());
			}
			//ansonsten ignorieren
		}else if (state==PS_IN){
			if (zeile.find("#")==0){ //Falls die Zeile mit # beginnt
				if (zeile.find("utput")==NOT_FOUND) cout<<"warning: comment \"Output\" expected.\n";
			  state=PS_OUT;
 			  outputs.push_back(CDoubleVector());				
			}else{
				//Werte einlesen
				//if (!is_numeric(zeile[0])) throw 666;		//Falsches Foramt: Zeile muss mit Zahl beginnen
				CDoubleVector vec(zeile);
				assert(inputs.size()>0);
				inputs[inputs.size()-1].push_back_doublevector(vec);
			}
		}else if (state==PS_OUT){
			if (zeile.find("#")==0){ //Falls die Zeile mit # beginnt
				if (zeile.find("nput")==NOT_FOUND)  cout<<"warning: comment \"Input\" expected.\n";
			  state=PS_IN;
 			  inputs.push_back(CDoubleVector());				
			}else{
				//Werte einlesen
				CDoubleVector vec(zeile);
				assert(outputs.size()>0);
				outputs[outputs.size()-1].push_back_doublevector(vec);
			}
		}
	}
	//Schlieen der Datei
	if (in != &cin)
		delete in;
	
	//berprfen der Ergebnisse
	if (inputs.size()!= outputs.size()){
	 	cerr<<"Error while parsing pattern file.";
		return false;
	}else
		return true;
}

//------------------------------------------------------------------
/** schreibt zusammengehrenden In-und Output in ein SNNS-Pattern-File */
bool CNeuronica::save_pattern(ostream& out, CDoubleVector& input, CDoubleVector& output,long pos=0){
	
		out<<"# Input pattern ";
		if (pos!=0) out<<pos;
		out<<" :\n";
		for (ulong i=0; i<=((input.size()-1)/10);i++){
			input.write_partially(out,i*10,((i*10+10)<input.size())? i*10+10 :input.size()-1);
			out<<"\n";
		}

		out<<"# Output pattern ";
		if (pos!=0) out<<pos;
		out<<" :\n";
		for (ulong i=0; i<=((output.size()-1)/10);i++){
			output.write_partially(out,i*10,((i*10+10)<output.size())? i*10+10 :output.size()-1);
			out<<"\n";
		}
		
		return true;
}

//------------------------------------------------------------------
/** schreibt die Muster in ein SNNS-Pattern-File */
bool CNeuronica::save_patterns(const char* filename,std::vector< CDoubleVector > & inputs, std::vector<CDoubleVector> & outputs){
		assert( inputs.size()==outputs.size());
		assert(inputs.size()>=1);
		
		//Datei anlegen
		ostream* out=&cout;
		if (filename!="-")
			out = new ofstream (filename);		
		if (!out) throw_error( 666, "File could not be opened for writing.","CNeuronica::save_patterns");
		
		(*out) <<"SNNS pattern definition file V3.2\n"
					"generated at <time>\n\n"
			    "\nNo. of patterns : "<<inputs.size()
			 <<"\nNo. of input units : " <<inputs[0].size()
			 <<"\nNo. of output units : "<<outputs[0].size()<<"\n\n";

		for (ulong i=0; i<inputs.size(); i++) save_pattern((*out),inputs[i],outputs[i]);
		
		//Datei schlieen
		if (out != &cout)
			delete out;
		return true;
}
//------------------------------------------------------------------
/** verbindet net_out und net_in zu einem Netz. Die Parameter haben dieselben
Bedeutungen wie bei CNeuronicMultiNet::connect() */
long CNeuronica::merge(CNeuronicNet* net_out, CNeuronicNet* net_in, ELayerCode layer_net_out, ELayerCode layer_net_in, long startneuron){
	assert(net_out);
	assert(net_in);
	CNeuronicLayer* lay1=(long(layer_net_out)>=IL0) ? net_out->input_layer(long(layer_net_out)):net_out->output_layer(layer_net_out);
	CNeuronicLayer* lay2=(long(layer_net_in)>=IL0) ? net_in->input_layer(long(layer_net_in)):net_in->output_layer(layer_net_in);
	assert(lay1);
	assert(lay2);
	long erg;
	//
	//Verbinden:
	if (startneuron==-1) startneuron=0;
	//die Neuronen einzeln verbinden und zwar ab dem Startneuron
	assert(lay1->get_anzahl_neurons()<=(lay2->get_anzahl_neurons()-startneuron));
	assert(startneuron<=lay2->get_anzahl_neurons());
	for (long i=lay1->get_anzahl_neurons()-1;i>=0;i--){
			assert(dynamic_cast< CInputNeuron*>(lay2->neuron(i+startneuron)));	//es sollte sich um ein Inputneuron handeln, an das angeknpft wird
			//ersetze das Inputneuron in lay2 durch Neuron in Lay1
			//dazu zuerst alle OuptputSynapsen des InputNeurons auf dieses Neuron umleiten..
			for (long syn=lay2->neuron(i+startneuron)->get_anzahl_outputs()-1;syn>=0; syn--){	//dieser Code ist etwas gefhrlich
				 lay2->neuron(i+startneuron)->output_synapse(syn)->replace_pre_synaptic_neuron(lay1->neuron(i));
			}
			//.. und dann das InputNeuron lschen
			assert(lay2->neuron(i+startneuron)->get_anzahl_outputs()==0);
			lay2->remove_neuron(i+startneuron,true);
	}
	//die brigen Schichten von net_in nach net_out bernehmen (lay2 nur, wenn diese nicht leer ist)
	//Achtung REIHENFOLGE wichtig!!!
	for (long i=0; i<net_in->get_anzahl_layers();i++){
		if (net_in->layer(i)!=lay2){
			net_out->add_layer( net_in->layer(i));
		}else{
			//ist lay2 leer?
    	if (lay2->get_anzahl_neurons()!=0){
    		 erg=net_out->add_layer(lay2);
    	}else{
    		 erg=-1;
    	}
    }
	}
	for (long i=net_in->get_anzahl_layers()-1;i>=0;i--){
			net_in->remove_layer(i,false);
	}
	if (lay2->get_anzahl_neurons()==0) delete lay2;
	return erg;	
}


#ifdef USE_LIBXML
//------------------------------------------------------------------
/** Konvertiert eine Datei im Neuronica-NND-Format ins SNNS-NET-Format.
		Dabei wird (aus historischen Grnden) nicht das Netz erst aufgebaut, sondern direkt die
		NND-Datei geparst.
		Es werden allerdings nicht alle Netztypen untersttzt (z.B. kein CNeuronicMultiNet)
		*/
bool CNeuronica::nnd2net(const char* nndfile, const char * netfile){
	bool ergebnis =false;
	//Die Datei anlegen
	ostream* out=&cout;
	if (netfile!="-") out = new ofstream(netfile);
	if (!out) throw_error(666, "File couldn't be opened for writing","CNeuronica::nnd2net");

  //Einlesen des nnd-Files
  xmlDocPtr doc;
  xmlNodePtr tree=0;
  doc= xmlParseFile(nndfile);
  if (doc){
    tree = doc->xmlChildrenNode;
	 	if (tree) tree=xmlSearchNode(tree,"CNeuronicNet");
	}
	if (!tree) throw_error(666,"Wrong File Format","CNeuronica::nnd2net");

	//das eigentliche umwandeln
	ergebnis=nnd2net_intern(*out,tree);

	if (out != &cout) delete out;
  xmlFreeDoc(doc);

  return ergebnis;
}
#endif

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Hilfsfunktionen
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#ifdef USE_LIBXML
//------------------------------------------------------------------
/**	generates a SNNS network file (net) from a Neuronica network definition file (nnd)*/
bool CNeuronica::nnd2net_intern(ostream& out,xmlNodePtr tree)
{
    assert (!strcmp((const char*)tree->name, "CNeuronicNet"));
    xmlNodePtr subtree,subsubtree;
    //zuerst die Anzahl der Input-Schichten bestimmen
    long noof_input_layers=1;
    subtree=tree->xmlChildrenNode;
    while (subtree!=0){
    	xmlLoadLong(subtree,"noof_input_layers",noof_input_layers);
    	subtree=subtree->next;
    }
    //die anderen Infos sammeln
    subtree=tree->xmlChildrenNode;
    long layernr=0;			//Die Schicht
    TNeuronsVecVec neurons;	//beinhaltet die Informationen ber die Neuronen
    while (subtree!=0)
    {
    	char* type;
    	type=(char*) xmlGetProp(subtree, (const xmlChar*) "type");
    	if (type && (!strcmp( type, "layer"))){ //falls es sich um eine Schicht handelt ->hurra
          free (type);
          layernr++;
          std::vector< TNeuronInfo> tmplayer;
          neurons.push_back( tmplayer);
          //finde die Neuronen dieser Schicht
          subsubtree=subtree->xmlChildrenNode;
          while (subsubtree!=0){
          	type=( char*)xmlGetProp(subsubtree,(const xmlChar*) "type");
          	//falls es sich um eine Schicht handelt ->weiter eintauchen bis wir zu einem Neuron kommen
          	//FIXME: Wenn einmal ein Neuron aufgetaucht ist, darf keine Schicht mehr kommen
          	if (type && !strcmp(type, "layer")){
          		subsubtree=subtree->xmlChildrenNode;
          	}else if (type &&!strcmp(type, "neuron")){		//es handelt sich um ein Neuron in dieser Schicht
          		//Neuron finden und Infos speichern
          		neurons[layernr-1].push_back(nnd2net_get_neuron_info(subsubtree));
             subsubtree=subsubtree->next;
          	}else{
          		subsubtree=subsubtree->next;
          	}
          	free(type);
          }
          
      }else{
          free (type);
      }

    	subtree=subtree->next;
    }

    //Lesen der Verbindungen
    subtree=xmlSearchNode(tree->xmlChildrenNode,"Connection",false);  //Breitensuche nach Connection
    while (subtree!=0){
    	if ((!strcmp((const char*)subtree->name, "Connection"))){
          xmlNodePtr proptree=subtree->xmlChildrenNode;

          long anz_read=0;
          long pre_layer,pre_pos,post_layer,post_pos;

          while (proptree!=0){
          	if (xmlLoadLong(proptree,"pre_layer",pre_layer)) anz_read++;
          	if (xmlLoadLong(proptree,"pre_pos",pre_pos)) anz_read++;
          	if (xmlLoadLong(proptree,"post_layer",post_layer)) anz_read++;
          	if (xmlLoadLong(proptree,"post_pos",post_pos)) anz_read++;
          	proptree=proptree->next;
          }
          //Von der Synapse interessiert nur die Schwelle
          proptree=xmlSearchNode(subtree->xmlChildrenNode,"CSynapse",false);
          if (!proptree) throw_error( 666,	"ungltiges Datei-Format.", "CNeuronica::nnd2net_intern");
          proptree=proptree->xmlChildrenNode;
          if (!proptree) throw_error( 666,	"ungltiges Datei-Format.", "CNeuronica::nnd2net_intern");
          anz_read=0;
          double weight;
          while (proptree!=0){
          	if (xmlLoadDouble(proptree,"weight",weight)) anz_read++;
          	proptree=proptree->next;
          }
          if (!anz_read>0) throw_error( 666,	"ungltiges Datei-Format.", "CNeuronica::nnd2net_intern");
          //Speichern der Info
          TSynInfo wght_info;
          wght_info.pos=nnd2net_get_neuron_nr(neurons,pre_layer,pre_pos);
          wght_info.weight=weight;
          neurons[post_layer][post_pos].weights.push_back(wght_info);
      }
    	subtree=subtree->next;
    }
    //Schreiben der NEuronen
    nnd2net_write_snns_file(neurons,noof_input_layers,out);
    return true;
};
#endif


//------------------------------------------------------------------
/** Hilfsfunktion zum Schreiben ins SNNS-Format
 		schreibt auf Basis der in neurons gespeicherten Werte das Netz
 		*/
void CNeuronica::nnd2net_write_snns_file(TNeuronsVecVec& neurons,ulong noof_input_layers,ostream& out){
    assert (neurons.size()>0);

    //Bestimmen der Anzahl der Neuronen
    long anz_neurons=0;
    ulong anz_layers=neurons.size();
    for ( ulong i=0; i<neurons.size();i++) for (ulong j=0;j<neurons[i].size(); j++) anz_neurons++;

    //ev. Bestimmen der hauptschlich verwendeten Aktivierungsfunktion   
    CString default_act_func_name="Act_Logistic" ;

    //Schreiben der allgemeinen Sachen
    out << "SNNS network definition file V1.4-3D\n"
      << "generated at <time>\n\n"
      << "network name : Neuronica-SNNS-Net\n"
      << "source files :\n"
      << "no. of units : " <<anz_neurons << "\n"
      << "no. of connections : 0\n"
      << "no. of unit types : 0\n"
      << "no. of site types : 0\n\n\n"
      << "learning function : Std_Backpropagation\n"
      << "update function   : Topological_Order\n\n\n"
      << "unit default section :\n\n"
      << "act      | bias     | st | subnet | layer | act func     | out func\n"
      << "---------|----------|----|--------|-------|--------------|-------------\n"
      << " 0.00000 |  0.00000 | h  |      0 |     1 | "<<default_act_func_name<<" | Out_Identity\n"
      << "---------|----------|----|--------|-------|--------------|-------------\n\n\n";

    // List units
    out << "unit definition section :\n\n"
      << "no. | typeName | unitName | act      | bias     | st | position | act func             | out func | sites\n"
      << "----|----------|----------|----------|----------|----|----------|----------------------|----------|-------\n";

    //Schreiben der Neuronen Werte
    long neuronnr=0;
    for (unsigned long lay=0; lay<neurons.size();lay++){
        assert (neurons[lay].size()>0);
        for (ulong nrn=0; nrn<neurons[lay].size();nrn++){
            neuronnr++;
            //Ausschreiben der Werte fr das Neuron [lay][nrn]
            char layer_type =(lay<noof_input_layers) ? 'i': (lay==(anz_layers-1)) ? 'o' : 'h';
            out << format_string ("%3d |          | unit     | % 3.5f | % 3.5f | %c  | %2d,%2d,%2d |",
              neuronnr, neurons[lay][nrn].activation, neurons[lay][nrn].bias, layer_type,
              (lay<noof_input_layers) ? lay: (lay==anz_layers-1) ? lay+2:lay+1 ,nrn,0);
            //Die Inputschichten im SNNS haben keine Aktivierungsfunktionen
            CString af_name=((layer_type=='i') ? "" :get_snns_act_func_name( neurons[lay][nrn].act_func,default_act_func_name));
            out <<  ((af_name=="") ? "" : " "+af_name+" ")<<"||\n";
        }
    }
    out << "----|----------|----------|----------|----------|----|----------|----------------------|----------|-------\n";

    // List connections
    out << "\n\nconnection definition section :\n\n"
    	<< "target | site | source:weight\n"
    	<< "-------|------|---------------------------------------------------------------------------------------------------------------------\n";
    for (ulong lay=0; lay<neurons.size(); lay++){
     for (ulong nrn=0; nrn<neurons[lay].size(); nrn++){
    	if (neurons[lay][nrn].weights.size()>0) {
    		//long k=neurons[lay].size();
    		out << format_string ("%6d |      |", nnd2net_get_neuron_nr(neurons,lay,nrn)); // Target neuron
    		for (ulong i=0; i<neurons[lay][nrn].weights.size();i++){
    			if (i>0) out << ',';
    			out << format_string ("%3d:% .5f",
    						   neurons[lay][nrn].weights[i].pos, // Source neuron ID
    						   neurons[lay][nrn].weights[i].weight); // Connection weight
    		}
    		out << '\n';
    	}
    }}
    out << "-------|------|---------------------------------------------------------------------------------------------------------------------\n";
}



#ifdef USE_LIBXML
//------------------------------------------------------------------
/** Hilfsfunktion um die Informationen zu einem Neuron auszulesen (wird in nnd2net verwendet)
		*/
TNeuronInfo CNeuronica::nnd2net_get_neuron_info(xmlNodePtr node){
    TNeuronInfo ergebnis;
    //uns interessieren nur die Eigenschaften Aktivierung und Schwelle von CNeuron
    xmlNodePtr neuron_node=xmlSearchNode(node,"CNeuron");
    if (neuron_node==0) throw_error( 666, "Falsches Datei-Format.","CNeuronica::nnd2net_get_neuron_info");
    //Auslesen der Eigenschaften
    xmlNodePtr proptree=neuron_node->xmlChildrenNode;
    //int i;
    //double activation;
    //double bias;
    long anz_read=0;
    while (proptree!=0){
        if (xmlLoadDouble(proptree,"activation",ergebnis.activation)) anz_read++;
        if( xmlLoadDouble(proptree,"schwelle",ergebnis.bias)) anz_read++;
        int act_func;
        if( xmlLoadInt(proptree,"act_activation_function",act_func)){
            ergebnis.act_func=static_cast< EActivationFunction>(act_func);
            anz_read++;
        }
        proptree=proptree->next;
    }
    if (anz_read<3) throw_error( 666, "Falsches Datei-Format.","CNeuronica::nnd2net_get_neuron_info");
    return ergebnis;
}
#endif

//------------------------------------------------------------------
/** Hilfsfunktion zum Bestimmen der Nummer eines Neurons in einer TNeuronsVecVec - Struktur
		*/
long CNeuronica::nnd2net_get_neuron_nr(TNeuronsVecVec& neurons,long layer,long pos){
	assert((layer>=0) && (layer<(long)neurons.size()));
	assert((pos>=0) && (pos<(long)neurons[layer].size()));
	long erg=0;
	for (long i=0;i<layer;i++) erg+=neurons[i].size();
	return erg+pos+1;
}

//------------------------------------------------------------------
  /** Hilfsfunktion, liefert zu einer Aktivierungsfunktion den entsprechenden Namen im SNNS
      */
CString CNeuronica::get_snns_act_func_name(EActivationFunction act_func, const CString& default_name=""){
    CString name;
    switch (act_func){
        case AF_LOGISTIC:       name="Act_Logistic";break;
        case AF_TANGENSH:       name="Act_TanH";break;
        case AF_IDENT:          name="Act_Identity";break;
        case AF_LOGISTIC_SNNS:  name="Act_Logistic";break;
        default:                name="";
    }
    if (name==default_name) name="";
    return name;
}
