/***************************************************************************
                          cneuroniclayer.cpp  -  description
                             -------------------
    begin                : Tue Jun 20 2000
    copyright            : (C) 2000 by Jon Hennig
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cneuroniclayer.h"
#include "cbasicneuron.h"
#include "cdoublevector.h"
#include <iomanip>   //for setw()
#include "cneuronica.h"
#include "tools.h"								//for xml
//-----------------------------------------------------------------------------
CNeuronicLayer::CNeuronicLayer()
{
	faktor=0;
	offset=0;
}
//-----------------------------------------------------------------------------
CNeuronicLayer::~CNeuronicLayer(){
	while (neurons.size()>0){
		delete neurons[0];
		neurons[0]=0;
		neurons.erase(neurons.begin());  	
  }
  assert( get_anzahl_neurons()==0);

}

//-----------------------------------------------------------------------------
/** liefert das Neuron mit der Nummer number aus der Schicht */
CNeuron* CNeuronicLayer::neuron(long number)
{
	assert(number>=0);
	assert(number<(long)neurons.size());
	assert(neurons[number]);
	return neurons[number];
	
}

//-----------------------------------------------------------------------------
/** verbindet alle Neuronen dieser Schicht mit allen Neuronen aus der Schicht other */
void CNeuronicLayer::connect_output(CNeuronicLayer* other, TNSynapticWeight min_weight=-0.2, TNSynapticWeight max_weight=0.2,EConnectionType connection_type=CT_ALL_BEETWEEN_LAYERS )
{
	assert(other);
	switch (connection_type)
	{
		case CT_ALL_BEETWEEN_LAYERS:{
			for(long i=0; i<other->get_anzahl_neurons();i++){
				connect_output(other->neuron(i),min_weight,max_weight);
			}
		}break;
		case CT_ONE_TO_ONE:{
			//das erste Neuron wird mit dem ersten der anderen Schicht verbunden usw.
			if(get_anzahl_neurons()!=other->get_anzahl_neurons()) throw_error( 666,"The number of neurons in the layers to be connected must be the same if you use CT_ONE_TO_ONE.","CNeuronicLayer::connect_output"); //With this connection type both layer must have the same size
			for(long i=0; i<get_anzahl_neurons();i++){
				//FIXME:	Implementierung ist falsch, beachtet max_weight nicht (wird zur Zeit aber nur so bentigt)
				assert(min_weight==max_weight);
				neuron(i)->connect_output(other->neuron(i),min_weight);
			}			
		}break;
		default: throw_error( 666, "Unknown connection type", "CNeuronicLayer::connect_output");
	}
			
}
//-----------------------------------------------------------------------------
/** liefert die Anzahl der Neuronen in der Schicht */
long CNeuronicLayer::get_anzahl_neurons()
{
	return neurons.size();
}
//-----------------------------------------------------------------------------
/** verbindet Outputs aller Neuronen der Schicht mit dem Neruon neuron */
void CNeuronicLayer::connect_output(CNeuron* a_neuron, TNSynapticWeight min_weight,TNSynapticWeight max_weight)
{
	assert(a_neuron);
	assert(min_weight<=max_weight);
	//
	for(long i=0; i<get_anzahl_neurons(); i++)
	{
		assert(neuron(i));	
		double range=max_weight-min_weight;
		TNSynapticWeight weight;
		if (range==0)
		{
			//die beiden Werte sind gleich
			weight=min_weight;
		}else
		{
			weight=((double)rand()/(double)RAND_MAX) //liefert eine Zufallszahl zw. 0 und 1
				*range				 //liefert eine Zahl zw. 0 und range
				+min_weight;			 //liefert eine Zahl zw. min_weight und min_weight+range
		}
		neuron(i)->connect_output(a_neuron,weight);
	
	}
}
//-----------------------------------------------------------------------------
/** liefert die angesammelten Aktivierungen an die nachfolgenden Schichten weiter */
void CNeuronicLayer::feed_forward(TNTimeUnit time)
{
	for (long i=0;i<get_anzahl_neurons();i++)
	{	
		neuron(i)->feed_forward(time);
	}
}


//-----------------------------------------------------------------------------
/** liefert ein feedback auf eine fehleraktivierung hin */
void CNeuronicLayer::feed_back(TNTimeUnit time)
{
	for (long i=get_anzahl_neurons()-1;i>=0;i--)
	{	
		neuron(i)->feed_back(time);
	}
}
//-----------------------------------------------------------------------------
/** fgt das Neuron a_neuron zu der Menge der verwalteten Neuronen hinzu
und liefert dessen Nummer zurck. */
long CNeuronicLayer::add_neuron(CNeuron* a_neuron)
{
	neurons.push_back(a_neuron);
	return get_anzahl_neurons()-1;
}
//-----------------------------------------------------------------------------
/** setzt die Aktivierungen der Neuronen in der Schciht */
void CNeuronicLayer::set_activation(const CDoubleVector& activations, bool use_scaling=false)
{
	assert((long)activations.size()==get_anzahl_neurons());
	//
	if (!use_scaling)
	{ //normal die Werte setzen
		for (long i=0;i<get_anzahl_neurons();i++)
		{
			neuron(i)->set_activation(activations[i]);
		}
	}else
	{ //unter Verwendung der Skalierung die Werte setzen
		for (long i=0;i<get_anzahl_neurons();i++)
		{
			assert(offset);
			assert(faktor);
			neuron(i)->set_activation((activations[i]-(*offset)[i])/(*faktor)[i]);
		}		
	}
}
//-----------------------------------------------------------------------------
/** liefert die Outputs der Neuronen der Schicht in a_outputs zuruueck */
void CNeuronicLayer::get_output(CDoubleVector& a_outputs, bool use_scaling=false)
{
	assert(((long)a_outputs.size()==get_anzahl_neurons())||((long)a_outputs.size()==0));
	//
	a_outputs.resize(get_anzahl_neurons());
	for (long i=0; i<(long) get_anzahl_neurons();i++)
	{
		a_outputs[i]=neuron(i)->get_last_output();
	}
	if (use_scaling)
	{
		//FIXME: throw is better
		assert(faktor);
		assert(offset);
		a_outputs.multiply(*faktor);
		a_outputs.add(*offset);
	}
}
//-----------------------------------------------------------------------------
/** setzt den Fehler fr alle Neuronen der Schicht */
void CNeuronicLayer::set_error(const CDoubleVector& a_errors, bool use_scaling=false)
{
	assert((long)a_errors.size()==get_anzahl_neurons());
	//
	if (!use_scaling)
	{
		for (long i=0;i<get_anzahl_neurons();i++)
		{
			neuron(i)->set_error(a_errors[i]);
		}
	}else
	{	
		assert(offset);
		assert(faktor);
		for (long i=0;i<get_anzahl_neurons();i++)
		{
			neuron(i)->set_error((a_errors[i]/(*faktor)[i])); //der Offset ist irrelevant, da es sich ja um einen Fehler handelt
		}
	}	
}
//-----------------------------------------------------------------------------
/** setzt die Skalierung so, dass der Input so eingerichtet wird, dass er zwischen scaled_min und scaled_max liegt
  * wenn der Output skaliert wird muss der ursprngliche Wert zwischen scaled_min und scaled_max liegen,
  * damit er erfolgreich zurckskaliert werden kann */
void CNeuronicLayer::set_scaling( CDoubleVector& unscaled_min,	//Zeiger auf Vektor der minimalen unskalierten Werte
																  CDoubleVector& unscaled_max,	//Zeiger auf Vektor der maximalen unskalierten Werte
																  double scaled_min,						//der min. unsk. wert entspricht diesem Wert
																  double scaled_max)						//der max. unsk. Wert entspricht diesem Wert
{
	assert((long) unscaled_min.size()==get_anzahl_neurons());
	assert((long) unscaled_max.size()==get_anzahl_neurons());
	assert(scaled_min<scaled_max);
	assert(unscaled_min<unscaled_max);
	
	//s=scaled, us=unscaled
	double s_range;
	if (faktor!=0) delete faktor;
	if (offset!=0) delete offset;
	faktor= new CDoubleVector();
	offset= new CDoubleVector();
	
	s_range=scaled_max-scaled_min;
	
  //jetzt fr jedes element den Faktor ermitteln
  //der ergibt sich aus dem unskalierten durch den skalierten Bereich
		(*faktor)=unscaled_max-unscaled_min; //=us_range
		faktor->divide(s_range);				//faktor = us_range/s_range

	//offset=unskaliertes_min-(skaliertes_min*faktor)
	(*offset)=(*faktor);
	offset->multiply(-scaled_min);
	(*offset)=(*offset)+unscaled_min;
		
}
//-----------------------------------------------------------------------------
/* setzt die Skalierung automatisch, so dass die skalierten Werte so liegen,
dass automatisch der volle Bereich (-reserve) ausgeschpft wird*/
void CNeuronicLayer::set_scaling_auto( CDoubleVector& unscaled_min,	//Zeiger auf Vektor der minimalen unskalierten Werte
																  CDoubleVector& unscaled_max,	//Zeiger auf Vektor der maximalen unskalierten Werte
																  double reserve)								//die Prozentuelle Reserve am unteren und oberen Rand
{
	assert(neuron(0));
	assert(reserve<50 && reserve>0);
	double min=neuron(0)->get_min_output();
	double max=neuron(0)->get_max_output();
	double ist_reserve=(max-min)*((reserve)/100.0);
	set_scaling(unscaled_min,unscaled_max,min+ist_reserve,max-ist_reserve);
}

//-----------------------------------------------------------------------------
/** ruft in allen Neuronen der Schicht die Funktion
allow_learning auf. Falls true bergeben wird., wird
das Lernen fr alle Neuronen der Schicht gestattet,
sonst nicht. */
void CNeuronicLayer::allow_learning(bool allow){
    for (long i=0;i<get_anzahl_neurons();i++){
        neuron(i)->allow_learning(allow);
    }

}
//-----------------------------------------------------------------------------
/** ruft in allen Neuronen der Schicht die Funktion
allow_feedback auf. */
void CNeuronicLayer::allow_feedback(bool allow){
    for (long i=0;i<get_anzahl_neurons();i++){
        neuron(i)->allow_feedback(allow);
    }
}

//-----------------------------------------------------------------------------
#ifdef USE_CCOMMUNICATION
/** ausfhren eines Kommandos*/
bool CNeuronicLayer::process_cmd(const char* a_command, const void* data,TPtrVector* add_data=0,long id=0){
    if (check_command(a_command,"print_output_and_error","Gibt Fehler und Output der Neuronen des Netzes aus.",id)){
        print_output_and_error();
        return true;
    }        
    if (check_command(a_command,"neuron","Gibt den Befehl an das angegebene Neuron weiter. (Syntax: neuron(nr))",id)){
        char* new_command;	//der weiterzureichende Befehl
        long start,stop;		//die Element, an die der Befehl weiterzureichen ist
        if (parse_command(a_command,new_command,start,stop,0,get_anzahl_neurons()-1)){
            //weiterreichen an den entsprechende Neuron
            for (long i=start; i<=stop; i++){
                if (i!=start) infos(format_string("neuron(%d).",i));
                neuron(i)->process_cmd(new_command,data,add_data,id);
            }
            return true;
        }
        else infos("Fehlerhaftes Kommando.\n");
    }
    // die Parentklasse wei hoffentlich was damit anzufangen , und falls nicht weiterleiten an die neuronen
    if (!CCommunication::process_cmd(a_command,data,add_data,id)){
        return process_cmd(cat_strings("neuron.",a_command).c_str(),data,add_data,CCommunication::ID_NO_ERROR_MSG);
    }else return true;
}
#endif
//-----------------------------------------------------------------------------
/** gibt die Input-Gewichte der Schicht als Matrix aus */
void CNeuronicLayer::print_input_matrix(){
    for (long i=0;i<get_anzahl_neurons();i++){
        cout<<setw(3)<<i<<": ";
        neuron(i)->print_input_vector();
        cout<<"\n";
    }
}
//-----------------------------------------------------------------------------
/** gibt den Fehler der einzelnen Neuronen, die Fehleraktivierung, die Aktivierung und den Output aus      */
void CNeuronicLayer::print_output_and_error(){
    const long SW=15;
    cout<<"Neuron: ";
    for (long i=0;i<get_anzahl_neurons();i++) cout << setw(SW)<<i<<",";
    cout<<endl<<"Output: ";
    for (long i=0;i<get_anzahl_neurons();i++)    cout<< setw(SW)<<neuron(i)->get_last_output()<<",";
    cout<<endl<<"Aktivi: ";
    for (long i=0;i<get_anzahl_neurons();i++)    cout<< setw(SW)<<neuron(i)->get_activation()<<",";
    cout<<endl<<"Error : ";
    for (long i=0;i<get_anzahl_neurons();i++)    cout<< setw(SW)<<neuron(i)->get_last_error_output()<<",";
    cout<<endl<<"ErrAkt: ";
    for (long i=0;i<get_anzahl_neurons();i++)    cout<< setw(SW)<<neuron(i)->error_activation<<",";
    cout<<endl;
}
    

//-----------------------------------------------------------------------------
/** setzt die Aktivierungen aller Neuronen der Schicht auf 0 zurck */
void CNeuronicLayer::reset_activation(){
		for (long i=0;i<get_anzahl_neurons();i++)
		{
			neuron(i)->set_activation(0);
		}
}
//-----------------------------------------------------------------------------
/** Setzt die Outputs der Neuronen in der Schicht auf die vorgegebenen
Werte. Die Outputs der Neuronen werden in einem folgenden Feed-Forward-Schritt
nicht (neu) berechnet. */
void CNeuronicLayer::set_output(const CDoubleVector & values, bool use_scaling){
	assert((long)values.size()==get_anzahl_neurons());
	//
	if (!use_scaling)
	{ //normal die Werte setzen
		for (long i=0;i<get_anzahl_neurons();i++)
		{
			neuron(i)->set_output(values[i]);
		}
	}else
	{ //unter Verwendung der Skalierung die Werte setzen
		assert(offset);
		assert(faktor);
		assert(values.size()==offset->size());
		assert(values.size()==faktor->size());
		for (long i=0;i<get_anzahl_neurons();i++)
		{
			neuron(i)->set_output((values[i]-(*offset)[i])/(*faktor)[i]);
		}		
	}
}

#ifdef USE_LIBXML
//-----------------------------------------------------------------------------
/** speichert in das XML-File, zu dem tree gehrt */
xmlNodePtr CNeuronicLayer::save_to_xml(xmlNodePtr tree){
	assert(tree);
	//einen neuen Knoten anlegen (mit dem Klassennamen)
	xmlNodePtr subtree;
	subtree = xmlNewChild(tree, NULL, (const xmlChar *)"CNeuronicLayer",NULL);
	xmlSetProp(subtree, (const xmlChar *)"type",(const xmlChar *) "layer");	//es handelt sich um ein Element vom Typ Layer
	for (long i=0;i<get_anzahl_neurons();i++){
		neuron(i)->save_to_xml(subtree);
	}
	if (faktor) xmlSaveDoubleVector(subtree,"factor",*faktor);
	if (offset) xmlSaveDoubleVector(subtree,"offset",*offset);
  return subtree;
}
//-----------------------------------------------------------------------------
/** ldt aus dem XML-File, zu dem tree gehrt */
void CNeuronicLayer::load_from_xml(xmlNodePtr tree){
	assert(tree);
	//in den Knoten gehen (mit dem Klassennamen)
	if (strcmp((const char*)tree->name, "CNeuronicLayer")) throw_error(666,"Wrong file format. Node name 'CNeuronicLayer' expected.","CNeuronicLayer::load_from_xml");
	xmlNodePtr subtree;
	subtree=tree->xmlChildrenNode;
	CNeuron* neuron;
	CDoubleVector vec;
	while (subtree!=0)
	{
		neuron=CNeuronica::new_neuron(subtree);
		if (neuron){ add_neuron(neuron);} else{
			if (xmlLoadDoubleVector(subtree,"factor",vec)){
				faktor=new CDoubleVector();
				*faktor=vec;
			}	
  		if (xmlLoadDoubleVector(subtree,"offset",vec)){
				offset=new CDoubleVector();
				*offset=vec;
			}
  	}
		subtree=subtree->next;
	}
}
#endif

//-----------------------------------------------------------------------------
/** liefert die letzten Fehler der Neuronen der Schicht in a_errorss zuruueck
falls skaliert gesetzt ist, werden die nach set_scaling skalierten Werte zurckgeliefert */
void CNeuronicLayer::get_error(CDoubleVector & a_errors, bool use_scaling){
	assert(((long)a_errors.size()==get_anzahl_neurons())||((long)a_errors.size()==0));
	//
	a_errors.resize(get_anzahl_neurons());
	for (long i=0; i<(long) get_anzahl_neurons();i++)
	{
		a_errors[i]=neuron(i)->get_last_error();
	}
	if (use_scaling)
	{
		//FIXME: throw is better
		assert(faktor);
		a_errors.multiply(*faktor);
		//offset interessiert bei einem Fehler nicht
	}
}
//-----------------------------------------------------------------------------
/** entfernt das Neuron mit der Nummer nrn_nr */
void CNeuronicLayer::remove_neuron(long nrn_nr,bool delete_neuron){
		CNeuron* nrn=neuron(nrn_nr);
		assert(nrn);
		neurons.erase(neurons.begin()+nrn_nr);  	
		if (delete_neuron) delete nrn;
}
