/***************************************************************************
                          CNeuronicMultiNet.cpp  -  description
                             -------------------
    begin                : Sun Jan 7 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cneuronicmultinet.h"

#include "cneuroniclayer.h"
#include "cneuronica.h"
#include "cpatterns.h"

#ifdef USE_LIBXML
	#include <libxml/parser.h>
#endif


CNeuronicMultiNet::CNeuronicMultiNet(){
}
CNeuronicMultiNet::~CNeuronicMultiNet(){
}
/** liefert einen Zeiger auf das Subnet nr zurck */
CNeuronicNet* CNeuronicMultiNet::subnet(long nr)
{
	assert( (nr>=0)&&(nr<(long)subnets.size()) );
	assert( subnets[nr] );
	
	return subnets[nr];
}

//----------------------------------------------------------------------------------------
/** fgt dem Netz eine Schicht Neuronen hinzu
und liefert als Ergebnis die Nummer der Schicht im Net */
long CNeuronicMultiNet::add_layer(CNeuronicLayer* a_layer)
{
	throw_error(666, "Aufruf von add_layer ist fr CNeuronicMultiNet nicht erlaubt.");
	return -1;
}
//----------------------------------------------------------------------------------------
/** fgt dem Netz eine Teilnetz hinzu
und liefert als Ergebnis die Nummer des Teilnetzes im Netz */
long CNeuronicMultiNet::add_subnet(CNeuronicNet* a_net){
    assert(a_net);
    subnets.push_back(a_net);
    return subnets.size()-1;
}

//----------------------------------------------------------------------------------------
/** fhrt einen feed_forward schrittt ber das gesamte Netz aus */
void CNeuronicMultiNet::feed_forward(TNTimeUnit time=0){
    for (unsigned long i=0;i<subnets.size();i++){
        subnet(i)->feed_forward(time);
    }
}
//----------------------------------------------------------------------------------------
/** fhrt einen Feed-Back-Schritt ber das gesamte Netz aus */
void CNeuronicMultiNet::feed_back(TNTimeUnit time=0)
{
	for (long i=subnets.size()-1;i>=0;i--)
	{
		subnet(i)->feed_back(time);
	}
  after_feed_back(time);

}
//----------------------------------------------------------------------------------------
/** liefert die Zahl der Subnetze des Netzes */
long CNeuronicMultiNet::get_anzahl_subnets(){
	return subnets.size();
}
//----------------------------------------------------------------------------------------
/** verbindet die Neuronen des Netzes */
void CNeuronicMultiNet::connect(EConnectionType connection_type=CT_DEFAULT, TNSynapticWeight min_weight=-0.2, TNSynapticWeight max_weight=0.2)
{
	switch (connection_type)
	{
		case CT_ALL_BEETWEEN_LAYERS:
		{  throw_error( 666, "CT_ALL_BEETWEEN_LAYERS ist keine gltige Verbindungsart bei connect() in CNeuronicMultiNet.");
		}break;
		
		case CT_DEFAULT:
		{
			//Achtung! Diese Verbindungsart wird nur in seltenen Fllen das sein, was man wirklich will
			//nmlich dann, wenn alle Netze der Reihe nach hintereinandergeschaltet sind
			for (long i=0; i<get_anzahl_subnets()-1;i++){
				connect(i,i+1,OL0,IL0);
			}
		}break;
		
		default: throw_error( 666, "unbekannte Verbindungsart bei connect()");
		
	}
}
//----------------------------------------------------------------------------------------
/** setzt das gesamte Netz zurck (lscht alle Subnetze) */
void CNeuronicMultiNet::reset(bool force_delete)
{
	for (long i=0; i<get_anzahl_subnets();i++){
		subnet(i)->reset(force_delete);
		delete subnet(i);
	}
	subnets.resize(0);
	connection_infos.resize(0);
}


//-----------------------------------------------------------------------------
#ifdef USE_CCOMMUNICATION
/** ausfhren eines Kommandos*/
bool CNeuronicMultiNet::process_cmd(const char* a_command, const void* data,TPtrVector* add_data=0,long id=0){
    //----------
    if (check_command(a_command,"subnet","Gibt den Befehl an das Subnet weiter. (Syntax: subnet(nr))",id)||
      check_command(a_command,"sn","siehe subnet",id))
    {
        char* new_command;	//der weiterzureichende Befehl
        long start,stop;		//die Element, an die der Befehl weiterzureichen ist
        if (parse_command(a_command,new_command,start,stop,0,get_anzahl_subnets()-1))
        {	//TRICK um Laden der Art subnet(2).load mit der Funktion load_subnet zu ermglichen
            if (check_command(new_command,"load","Ldt das Netz, erzeugt dabei die bentigte Klasse",id)){
                for (long i=start; i<=stop; i++){
                    if (i!=start) infos(format_string("subnet(%d).",i));
                    const char* str;
                    if (!data){
                        str="net.nnd";
                    }else str=(const char *)data;
                    if (load_subnet(i,str)) infos("Netz wurde geladen.(multinet)\n");
                        else infos("Netz konnte nicht geladen werden.\n");
                }
                return true;
            }
            if (check_command(new_command,"load_snns","Ldt ein Netz im SNNS-Format (erzeugt die entspr. Klasse)",id)){
                for (long i=start; i<=stop; i++){
                    if (i!=start) infos(format_string("subnet(%d).",i));
                    const char* str;
                    if (!data){
                      str="net.net";
                    }else str=(const char *)data;
                    if (load_subnet(i,str,"SNNSNET")) infos("SNNS-Netz wurde geladen.(multinet)\n");
                      else infos("Netz konnte nicht geladen werden.\n");
                }
                return true;
            }
            //Standard: weiterreichen an das entsprechende Subnet
            for (long i=start; i<=stop; i++){
                if (i!=start) infos(format_string("subnet(%d).",i));
                subnet(i)->process_cmd(new_command,data,add_data,id);
            }
            return true;
        }
        else infos("- Fehlerhaftes Kommando.\n");
    }else
    //Die Funktion Layer hat in dieser Klasse zur Zeit noch keine bestimmte funktion
    //daher jeweils fr alle Subnetze das Kommando layer aufrufen
    if (check_command(a_command,"layer","",id)|| check_command(a_command,"ly","",id)){
        return process_cmd(cat_strings("subnet.",a_command).c_str(),data,add_data,id);
    }
    //----------
    // die Parentklasse wei hoffentlich was damit anzufangen , und falls nicht weiterleiten an die Subnetze leiten
    if (!CNeuronicNet::process_cmd(a_command,data,add_data,id)){
        return process_cmd(cat_strings("subnet.",a_command).c_str(),data,add_data,CCommunication::ID_NO_ERROR_MSG);
    }else return true;
}
#endif

#ifdef USE_LIBXML
//-----------------------------------------------------------------------------
/** speichert in das XML-File, zu dem tree gehrt */
xmlNodePtr CNeuronicMultiNet::save_to_xml(xmlNodePtr tree){
	assert(tree);
	//einen neuen Knoten anlegen (mit dem Klassennamen)
	xmlNodePtr subtree,subtree2;
	subtree2 = xmlNewChild(tree, NULL, (const xmlChar *)"CNeuronicMultiNet",NULL);
	//Subnets abspeichern
	for (long i=0;i<get_anzahl_subnets();i++){
		subnet(i)->save_to_xml(subtree2);
	}
	//Verbindungen abspeichern und zwar die gespeicherten in connection_infos
	for (ulong inf=0; inf<connection_infos.size();inf++){
		subtree = xmlNewChild(subtree2, NULL, (const xmlChar *)"LayerConnection",NULL);	
		xmlSaveLong(subtree,"pre_net",connection_infos[inf].net1);
		xmlSaveLong(subtree,"pre_layer",connection_infos[inf].lay1);
		xmlSaveLong(subtree,"post_net",connection_infos[inf].net2);
		xmlSaveLong(subtree,"post_layer",connection_infos[inf].lay2);
		xmlSaveLong(subtree,"startneuron",connection_infos[inf].startneuron);
	}
  return subtree;
}
//-----------------------------------------------------------------------------
/** ldt aus dem XML-File, zu dem tree gehrt */
void CNeuronicMultiNet::load_from_xml(xmlNodePtr tree){
	assert(tree);
	//in den Knoten gehen (mit dem Klassennamen)
	reset();
	if (strcmp((const char*)tree->name, "CNeuronicMultiNet")) throw_error(666,"load_from_xml: Node name CNeuronicMultiNet expected");	//Unexpected node name
	xmlNodePtr subtree,subsubtree;
	subtree=tree->xmlChildrenNode;
	CNeuronicNet* nnet;
//	CSynapse* syn;
	CDoubleVector vec;
	while (subtree!=0)
	{
		nnet=CNeuronica::new_net(subtree);
		if (nnet){ add_subnet(nnet);}
		else	//die Verbindungen laden				//FIXME: Achtung es mssen erst alle Netze geladen sein, bevor die Connections geladen werden knnen
  	if (!strcmp((const char*)subtree->name,"LayerConnection")){
  		long net1,net2,lay1,lay2,sn;
 			//die Verbindungsdaten auslesen
  		subsubtree=subtree->children;
  		while (subsubtree!=0){
  			xmlLoadLong(subsubtree,"pre_net",net1);
  			xmlLoadLong(subsubtree,"pre_layer",lay1);
  			xmlLoadLong(subsubtree,"post_net",net2);
  			xmlLoadLong(subsubtree,"post_layer",lay2); 			
  			xmlLoadLong(subsubtree,"startneuron",sn); 			
  			subsubtree=subsubtree->next;
  		}
  		//FIXME: Test ob auch alle Parameter eingegangen
  		//Die Funktion connect mit den entsprechenden Parametern aufrufen
  		connect(net1,net2,static_cast<ELayerCode>(lay1),static_cast<ELayerCode>(lay2),sn);
  	}
		subtree=subtree->next;
	}
}
#endif

//-----------------------------------------------------------------------------
/** ldt aus dem XML-File */
bool CNeuronicMultiNet::load(const char* filename , const char* format="DEFAULT" ){
 	assert(filename);
 	assert(format);
	#ifdef USE_LIBXML
  	if (!strcmp(format, "NND")||!strcmp(format, "DEFAULT")){
      xmlDocPtr doc;
      xmlNodePtr tree;

      doc= xmlParseFile(filename);
      if (doc){
        tree = doc->xmlChildrenNode;
        if (tree){
          tree = tree->xmlChildrenNode;		//NEURONICA
          if (tree){
          	//nach CNeuronicNet suchen und dann das Laden beginnen
    	      while (strcmp("CNeuronicMultiNet",(const char*) tree->name)) tree=tree->next;
    	      try{
        	  	load_from_xml(tree);
        	  }catch(int n){
    	        cout << "Error number is " << n << endl;
    	        xmlFreeDoc(doc);
    	        return false;
        	  }
          	xmlFreeDoc(doc);
    	      return true;
    	    }
        }
      }//else
      return false;
    }
  #endif
  if (!strcmp(format, "SNNSNET")){
  	//TODO: Implementieren - ist es mglich?
  	return false;
  }
  //else
  return false;
}

//-----------------------------------------------------------------------------
/** liefert einen Zeiger auf die Input-Schicht-Nr */
CNeuronicLayer * CNeuronicMultiNet::input_layer(long nr){
	//FIXME: ev. Variable einfhren, die festlegt, welches die Inputschichten sind
	assert(subnet(nr));
	return subnet(nr)->input_layer(0);
}
//-----------------------------------------------------------------------------
/** liefert einen Zeiger auf die Output-Schicht */
CNeuronicLayer * CNeuronicMultiNet::output_layer(ELayerCode nr=OL0){
	//FIXME: ev. Variable einfhren, die festlegt, welches die Outputschichten sind
  long laynr=(nr<0)? get_anzahl_subnets()+nr:nr;
	assert(laynr< get_anzahl_subnets() && laynr>=0);
	return subnet(laynr)->output_layer();
}
//-----------------------------------------------------------------------------
/** Verbindet die Schicht  layer_net_out des Subnetzes net_out mit der Schicht layer_net_in
des Subnetzes net_in
Dabei steht OL0..OLn fr die Outputlayer und IL0...ILn fr die Nr. der Inputlayer.  */
void CNeuronicMultiNet::connect(long net_out,long net_in,ELayerCode layer_net_out,ELayerCode layer_net_in,long startneuron,bool store_connection_info){
    assert(subnet(net_out));
    assert(subnet(net_in));
    //die Schicht entsprechend dem layercode heraussuchen (>=IL0 bedeuted, es ist eine Input-Schicht, sonst eine Output-Schicht)
    CNeuronicLayer* lay1=(long(layer_net_out)>=IL0) ? subnet(net_out)->input_layer(long(layer_net_out)):subnet(net_out)->output_layer(layer_net_out);
    CNeuronicLayer* lay2=(long(layer_net_in)>=IL0) ? subnet(net_in)->input_layer(long(layer_net_in)):subnet(net_in)->output_layer(layer_net_in);
    assert(lay1);
    assert(lay2);
    //
    CLayerConnectionInfo inf;
    inf.net1=net_out;
    inf.net2=net_in;
    inf.lay1=layer_net_out;
    inf.lay2=layer_net_in;
    inf.startneuron=startneuron;
    if (store_connection_info) connection_infos.push_back(inf);
    //Verbinden:
    if (startneuron==-1){ //Standardfall, nur ganze Schichten verbinden
        lay1->connect_output(lay2,0,0,CT_ONE_TO_ONE);
    }else{	//die Neuronen einzeln verbinden und zwar ab dem Startneuron
        assert(lay1->get_anzahl_neurons()<=(lay2->get_anzahl_neurons()-startneuron));
        assert(startneuron<=lay2->get_anzahl_neurons());
        for (long i=0; i<lay1->get_anzahl_neurons();i++){
            assert(dynamic_cast< CInputNeuron*>(lay2->neuron(i+startneuron)));	//es sollte sich um ein Inputneuron handeln, an das angeknpft wird
            lay1->neuron(i)->connect_output(lay2->neuron(i+startneuron),0);
        }
    }
}




//-----------------------------------------------------------------------------
/** Ldt das Netz mit der Nummer nr aus der Datei filename.
Ein eventuell schon vorhandenes Netz wird gelscht, die Verbindungen aber
wie bisher wieder hergestellt. */
bool CNeuronicMultiNet::load_subnet(long nr, const char* filename,  const char* format="DEFAULT"){
	assert(subnet(nr));
 	assert(filename);
 	assert(format);
	CNeuronicNet* net;
	net=CNeuronica::load_new_net(filename,format);
	if (net){
		//Dateien im Format SNNSNET enthalten keine Skalierung. Daher wird die bestehende Skalierung beibehalten.
		bool preserve_scaling=false;
		if (!strcmp(format, "SNNSNET")) preserve_scaling=true;		
		replace_subnet(nr,net,preserve_scaling);
	}
	return (net);
}
//-----------------------------------------------------------------------------
/** lscht das vorhandene Subnet mit der Nummer nr und ersetzt es durch
ein das Netz net.
Die Verbindungen zu anderen Netzen (die mit connect() erzeugt wurden)
bleiben dabei erhalten. (D.h. sie werden zunchst gelscht und fr das 
bergebene Netz neu angelegt.) */
void CNeuronicMultiNet::replace_subnet(long nr,CNeuronicNet* net,bool preserve_scaling=false){
    assert(subnet(nr));
    assert(net);
    CNeuronicNet* oldnet=subnet(nr);
    //optional bernehmen der alten Skalierungen
    if (preserve_scaling){
        //die Input-Schichten retten
        for (long i=0; i<oldnet->get_noof_input_layers();i++){
            if (oldnet->input_layer(i)->faktor){
                assert(oldnet->input_layer(i)->offset);
                if (net->input_layer(i)->faktor) delete net->input_layer(i)->faktor;
                if (net->input_layer(i)->offset) delete net->input_layer(i)->offset;
                net->input_layer(i)->faktor=oldnet->input_layer(i)->faktor;
                net->input_layer(i)->offset=oldnet->input_layer(i)->offset;
                //weglschen des Verweises auf faktor und offset im alten Netz, damit darauf kein delete erfolgt
                oldnet->input_layer(i)->offset=0;
                oldnet->input_layer(i)->faktor=0;
            }
        }
        //die Output-Schicht
        if (oldnet->output_layer()->faktor){
            assert(oldnet->output_layer()->offset);
            if (net->output_layer()->faktor) delete net->output_layer()->faktor;
            if (net->output_layer()->offset) delete net->output_layer()->offset;
            net->output_layer()->faktor=oldnet->output_layer()->faktor;
            net->output_layer()->offset=oldnet->output_layer()->offset;
            //weglschen des Verweises auf faktor und offset im alten Netz, damit darauf kein delete erfolgt
            oldnet->output_layer()->offset=0;
            oldnet->output_layer()->faktor=0;
        }
    }
    //lschen (Verbindungen werden automatisch entfernt) ...
    delete oldnet;
    //...und durch neues Netz ersetzen
    subnets[nr]=net;
    //Nun die Verbindungen wiederherstellen
    reconnect(nr);
}
//-----------------------------------------------------------------------------
/** Fhrt alle bisher durchgefhrten Verknpfungen, die das Netz nr betreffen erneut aus.
(wird von replace_subnet() bentigt) */
long CNeuronicMultiNet::reconnect(long nr){
    long anz=0;
    //connection_infos muss gelscht werden, da connect die Verbindungen erneut eintrgt
    //also zuerst Kopie herstellen
    //std::vector<CLayerConnectionInfo> cp_con_infos;
    //cp_con_infos=connection_infos;
    //connection_infos.resize(0);
    //der eigentliche Wiederverbindungsvorgang
    for (ulong i=0; i<connection_infos.size(); i++){
        CLayerConnectionInfo inf=connection_infos[i];
        if (inf.net1==nr || inf.net2==nr){
            anz++;
            //Das false gibt an, das die Verbindungsinformationen nicht nochmal gespeichert
            //werden sollen, da sie ja bereits vorhanden sind.
            connect(inf.net1,inf.net2,inf.lay1,inf.lay2,inf.startneuron,false);
        }
    }
    return anz;
}
//-----------------------------------------------------------------------------
/** ruft in allen Schichten des Netzes die Funktion
allow_learning auf. Falls true bergeben wird., wird
das Lernen fr alle Schichten des Netzes gestattet,
sonst nicht. */
void CNeuronicMultiNet::allow_learning(bool allow){
    for (long nets=0;nets<get_anzahl_subnets();nets++){
        subnet(nets)->allow_learning(allow);
    }
}
//-----------------------------------------------------------------------------
/** stellt die Lernrate aller Neuronen des Netzes auf den bergebenen Wert ein */
void CNeuronicMultiNet::set_learning_rate(double lr){
    //TODO: direkt aufrufen, um Geschw. zu verbessern
    bool old_value=CCommunication::global_showinfo;
    CCommunication::global_showinfo=false;
    process_command("subnet(*).layer(*).neuron(*).learning_rate",lr);
    CCommunication::global_showinfo=old_value;    
}
