/* **************************************************************************
                          cneuronicmultinet.h  -  description
                             -------------------
    begin                : Sun Jan 7 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CNEURONICMULTINET_H
#define CNEURONICMULTINET_H


#include "cneuronicnet.h"


/** NUR_INTERN: Struktur zum speichern von Informationen zur Verknpfung der Schichten zwischen den Subnetzen*/
struct CLayerConnectionInfo{
	long net1;
	long net2;
	ELayerCode lay1;
	ELayerCode lay2;	
	long startneuron;
};


/** BETA: implementiert ein (knstliches) Neuronales Netz wobei auch Teilnetze mglich sind.

    Das Netz kann entweder aus mehreren Schichten (layer) (siehe @ref CNeuronicNet) oder
    wiederum aus mehreren Netzen bestehen (subnet). Beides zusammen ist (z.Zt.) nicht mglich.
    Die Sub-Netze knnen selbst wieder Multi-Netze enthalten.

    Beispiel:
    <pre>
      CNeuronicNet* net;
      CNeuronicMultiNet* mnet=new CNeuronicMultiNet();
      // 3 Xor-Netze laden
      mnet->add_subnet(CNeuronica::load_new_net("xor.nnd"));	//Netz0
      mnet->add_subnet(CNeuronica::load_new_net("xor.nnd"));	//Netz1
      mnet->add_subnet(CNeuronica::load_new_net("xor.nnd"));	//Netz2
      // Zusammenschaltung der Netze (zu XOR(XOR(X1,X2),XOR(X3,X4))
      mnet->connect(0,2,OL0,IL0,0);	// verknpfe Outputlayer von Netz0 mit Inputlayer von Netz2 (ab Neuron 0)
      mnet->connect(1,2,OL0,IL0,1);	// verknpfe Outputlayer von Netz1 mit Inputlayer von Netz2 (ab Neuron 1)
      //Testen:
      CDoubleVector erg,tst; tst.resize(2); tst[0]=0; tst[1]=0;
      mnet->set_input(tst,true,0);				//Netz0 fttern
      mnet->set_input(tst,true,1);				//Netz1 fttern
      mnet->feed_forward();       				//Ergebnis berechnen ...
      mnet->get_output(erg,true); 				//...und auslesen
      cout << "      ERGEBNIS: " << erg[0] << endl;
    </pre>

    <b> TODO: </b>
    <li> Testen
    <li> Dokumentation vervollstndigen
    <li> load und save vollstndig implementieren
    <li> Funktionalitt vervollstndigen
    </li>

    <b> BUGS: </b>
    <li> load_subnet funktioniert scheinbar nicht korrekt.</li>
    </li>

    @author Jon Hennig
  */
class CNeuronicMultiNet : public CNeuronicNet

{
public:
	/** erzeugt ein leeres Netz
			*/
	CNeuronicMultiNet();
	virtual ~CNeuronicMultiNet();

	//noch? nicht implementiert
  /** liefert einen Zeiger auf die Schicht nr zurck */
  //CNeuronicLayer* layer(long nr);

  /** liefert einen Zeiger auf das Subnet nr zurck
  		*/
  CNeuronicNet* subnet(long nr);

  /** liefert einen Zeiger auf die Input-Schicht nr. Meist gibt es nur eine Input-Schicht und der
  		Aufruf kann ohne Parameter erfolgen.
  		*/
  virtual CNeuronicLayer * input_layer(long nr=0);

  /** liefert einen Zeiger auf die Output-Schicht nr.
      verwendet ELayerCode um auf die Outputschichten zuzugreifen
      Parameter OL0 liefert die Output-Layer des letzten Subnetzes.
      Parameter OL1 liefert die Output-Layer des vorletzten Subnetzes usw.
      nr=0 liefert die output_layer des subnetztes 0,
      nr=1 die output_layer des subnetzes 1 usw.
      */
  virtual CNeuronicLayer * output_layer(ELayerCode nr=OL0);


  /** liefert die Zahl der Schichten des Netzes */
  //long get_anzahl_layers();

  /** liefert die Zahl der Subnetze des Netzes
  		*/
  long get_anzahl_subnets();

  /** fhrt einen Feed-Back-Schritt ber das gesamte Netz aus.
  		Dazu wird feed_back fr jedes sub-Netz (gem ihrer Reihenfolge) aufgerufen.
  		@see CNeuronicNet::feed_back
  		*/
  void feed_back(TNTimeUnit time=0);

  /** fhrt einen Feed-Forward-Schritt ber das gesamte Netz aus.
  		Dazu wird feed_forward fr jedes sub-Netz (gem ihrer Reihenfolge) aufgerufen.
  		@see CNeuronicNet::feed_forward
  		*/
  void feed_forward(TNTimeUnit time=0);

  /** Aufruf dieser Funktion ist in dieser Klasse nicht erlaubt
  		*/
  long add_layer(CNeuronicLayer* a_layer);

  /** fgt dem Netz eine Teilnetz hinzu.
      @return liefert als Ergebnis die Nummer des Teilnetzes im Netz.
      */
  long add_subnet(CNeuronicNet* a_net);

  /** Aufruf dieser Funktion ist in dieser Klasse nicht erlaubt
  		*/
  template <class T> long add_layer(T* mother,long anzahl) { throw(666);} //add_layer ist in dieser Art Netz nicht erlaubt

  /** verbindet die Sub-Netze untereinander und zwar standardmig so, dass die Ausgnge von Sub-Netz(i) mit den Eingngen von Sub-Netz(i+1) verschaltet werden.
  		Will man eine andere Art der Verbindung, muss man diese manuell einrichten.
  		@see connect
  		*/
  void connect(EConnectionType connection_type=CT_DEFAULT, TNSynapticWeight min_weight=-0.2, TNSynapticWeight max_weight=0.2);

  /** setzt das gesamte Netz zurck (lscht alle Schichten)
      @param force_delete falls gesetzt (Standard) werden alle Zeiger mit delete weggelscht
    	*/
  void reset(bool force_delete=true);

	#ifdef USE_CCOMMUNICATION
	  /** Ausfhren eines Kommandos. Folgende stehen zur Verfgung: <b>
	  		<li> 	"subnet(i)" ... Gibt den Befehl an das Sub-Netz i weiter ("subnet(*)" - an alle Sub-Netze)
	  		</li>
	  		@see CCommunication::process_cmd
			  */
	  virtual bool process_cmd(const char* command, const void* data,TPtrVector* add_data=0,long id=0);
	#endif	

  #ifdef USE_LIBXML
	  /** ldt aus dem XML-File, zu dem tree gehrt
	  		*/
  	void load_from_xml(xmlNodePtr tree);
	  /** speichert ins XML-File, an der Stelle tree
	  		*/
  	xmlNodePtr save_to_xml(xmlNodePtr tree);
  #endif

  /** ldt das Netz aus einer Datei
  		@see CNeuronicNet::load
  		*/
  bool load(const char* filename , const char* format="DEFAULT" );

  /** Verbindet die Schicht layer_net_out des Subnetzes net_out mit der Schicht layer_net_in
			des Subnetzes net_in
			Dabei steht OL fr die Outputlayer und IL0...ILn fr die Nr. der Inputlayer.
			Die Verbindung wird hergestellt und die Verbindungsinformationen werden gespeichert.
			Im Standard-Fall wird immer die letzte Output-Schicht von Net1 mit der ersten Inputsicht  von net2 verbunden.
			Wenn Startneuron nicht angegeben ist (-1), mssen beide Schichten die selbe Zahl von Neuronen haben.
			Ist ein Startneuron angegeben, so wird das erste Neuron der layer_net_out mit dem neuron(start_neuron)der
			layer_net_in verbunden, das zweite mit neuron(startneuron+1) usw. In diesem Fall muessen genug Neuronen
			in layer_net_in vorhanden sein, um alle Neuronen aus layer_net_out zu verbinden.
			@param store_connection_info muss immer true sein, damit die Informationen ber die Verbindung zwischen den
																	 Netzen gespeichert werden kann. Wird nur intern (bei reconnect) auf false gesetzt
			*/
  void connect(long net_out,long net_in, ELayerCode layer_net_out=OL0,ELayerCode layer_net_in=IL0,long startneuron=-1,bool store_connection_info=true);

  /** Ldt das Netz mit der Nummer nr aus der Datei filename.
			Ein eventuell schon vorhandenes Netz wird gelscht, die Verbindungen aber
			wie bisher wieder hergestellt. D.h. das Netz wird ersetzt.
			*/
  bool load_subnet(long nr, const char* filename, const char* format="DEFAULT");

  /** lscht das vorhandene Subnet mit der Nummer nr und ersetzt es durch
			das Netz net.
			Die Verbindungen zu anderen Netzen (die mit connect() erzeugt wurden)
			bleiben dabei erhalten. (D.h. sie werden zunchst gelscht und fr das
			bergebene Netz neu angelegt.)
			@param preserve_scaling Falls dieser Parameter true ist, wird die Skalierung
			der In- und Output-Schichten beibehalten. (Default: false)
			*/
  void replace_subnet(long nr,CNeuronicNet* net,bool preserve_scaling=false);

  /** ruft in allen Schichten des Netzes die Funktion
			allow_learning auf. Falls true bergeben wird., wird
			das Lernen fr alle Schichten des Netzes gestattet,
			sonst nicht.
			*/
  virtual void allow_learning(bool allow);

  /** stellt die Lernrate aller Neuronen des Netzes auf den bergebenen Wert ein
  		*/
  virtual void set_learning_rate(double lr);

protected: // Protected functions
  /** Fhrt alle bisher durchgefhrten Verknpfungen, die die Schicht nr betreffen erneut aus.
      (wird von replace_subnet() bentigt).
      @return Anzahl der Ersetzungen.
      */
  long reconnect(long nr);


protected: // Protected attributes
  /** die Schichten, aus denen das Netz besteht
  		*/
  std::vector<CNeuronicNet*> subnets;
  /** Verbindungsinformationen
  		*/
  std::vector<CLayerConnectionInfo> connection_infos;

};

#endif
