/***************************************************************************
                          cneuronicnet.cpp  -  description
                             -------------------
    begin                : Sun Jan 7 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
 /***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "cneuronicnet.h"

#include "cneuroniclayer.h"
#include "cneuronica.h"
#include "cpatterns.h"

#ifdef USE_LIBXML
	#include <libxml/parser.h>
	#include <libxml/tree.h>
#endif


CNeuronicNet::CNeuronicNet(){
	noof_input_layers=0;
  pattern_logging=false;
}
CNeuronicNet::~CNeuronicNet(){
	reset();
}
//----------------------------------------------------------------------------------------
/** liefert einen Zeiger auf die Schicht nr zurck */
CNeuronicLayer* CNeuronicNet::layer(long nr){
	assert( (nr>=0)&&(nr<(long)layers.size()) );
	assert( layers[nr] );
	
	return layers[nr];
}

//----------------------------------------------------------------------------------------
/** fgt dem Netz eine Schicht Neuronen hinzu
und liefert als Ergebnis die Nummer der Schicht im Net */
long CNeuronicNet::add_layer(CNeuronicLayer* a_layer)
{
	assert(a_layer);
	layers.push_back(a_layer);
	if (noof_input_layers==0) noof_input_layers=1;	//Eine Schicht sollte normalerweise eine Input-Schicht sein
	return layers.size()-1;
}

//----------------------------------------------------------------------------------------
/** fhrt einen feed_forward schrittt ber das gesamte Netz aus */
void CNeuronicNet::feed_forward(TNTimeUnit time=0)
{
    for (unsigned long i=0;i<layers.size();i++)
    {
        layer(i)->feed_forward(time);
    }
}
//----------------------------------------------------------------------------------------
/** fhrt einen Feed-Back-Schritt ber das gesamte Netz aus */
void CNeuronicNet::feed_back(TNTimeUnit time=0){
    for (long i=layers.size()-1;i>=0;i--){
        layer(i)->feed_back(time);
    }
    after_feed_back(time);
}
//----------------------------------------------------------------------------------------
/** setzt den Input der ersten Schicht des Netzes*/
void CNeuronicNet::set_input(const CDoubleVector& input, bool use_scaling=false,long nr_input_layer=0){
	//Diese Implementierung gab Probleme mit MultiNet, da output eines Inputneurones so nicht gesetzt wird
	//wenn eine Inputschicht eines zweiten Neztes mit der Inputschicht des ersten Netzes verknpft ist,
	//fr das erste Netz aber nicht feed_forward aufgerufen wird, so erhlt das 2. Netz als Input immer 0 :
	//input_layer(nr_input_layer)->set_activation(input,use_scaling);
	
	//direkt das Output setzen
	//diese Variante verworfen, da so gekapselte Neuronen nicht abgefragt werden
	//wenn vor der Weiterverarbeitung muss fr die Schicht feed_forward aufgerufen wird,
	//werden die outputs wieder berschrieben?? (od. output_set?)
	//FIXME: falls input_layer keine Inputlayer ist, sollte es eine Warnung geben
	input_layer(nr_input_layer)->set_output(input,use_scaling);
}

//----------------------------------------------------------------------------------------
/** liefert den aktuellen Output der letzten Schicht des Netzes */
void CNeuronicNet::get_output(CDoubleVector& a_outputs, bool use_scaling=false,ELayerCode nr){
	output_layer(nr)->get_output(a_outputs,use_scaling);	
}
//----------------------------------------------------------------------------------------
/** liefert die Zahl der Schichten des Netzes */
long CNeuronicNet::get_anzahl_layers(){
	return layers.size();
}
//----------------------------------------------------------------------------------------
/** setzt den Fehler f. (ergibt sich aus soll-ist) die letzte Schicht des Netzes */
void CNeuronicNet::set_error(const CDoubleVector& a_errors, bool use_scaling=false){
	output_layer()->set_error(a_errors,use_scaling);	
}
//----------------------------------------------------------------------------------------
/** verbindet die Neuronen des Netzes */
void CNeuronicNet::connect(EConnectionType connection_type=CT_DEFAULT, TNSynapticWeight min_weight=-0.2, TNSynapticWeight max_weight=0.2)
{
	switch (connection_type)
	{
		case CT_ALL_BEETWEEN_LAYERS:
		case CT_DEFAULT:
		{
			//es sollten mind. 2 Schichten da sein
			if (get_anzahl_layers()<2) throw_error(666, "A minimum of 2 layers is required.", "CNeuronicNet::connect");
			long first_noinp_layer=get_noof_input_layers();
			//zuerst alle Input-Schichten mit der ersten nicht-Input-Schicht verbinden ...
			for (long i=0; i<get_noof_input_layers();i++){
				input_layer(i)->connect_output(layer(first_noinp_layer),min_weight,max_weight);				
			}
			//...dann die restlichen Schichten Schicht fr Schicht verbinden
			for (long i=first_noinp_layer; i<get_anzahl_layers()-1;i++)
			{
				layer(i)->connect_output(layer(i+1),min_weight,max_weight);
			}
		}break;
		
		default: throw 666;
		
	}
}
//----------------------------------------------------------------------------------------
/** setzt das gesamte Netz zurck (lscht alle Schichten) */
void CNeuronicNet::reset(bool force_delete)
{
	while (layers.size()>0){
		delete layers[0];
		layers[0]=0;
		layers.erase(layers.begin());  	
		//cerr<< layers.size()<<endl;
  }
  noof_input_layers=0;
  assert( get_anzahl_layers()==0);
}

//----------------------------------------------------------------------------------------
/** liefert zu dem Netzinput in in den entsprechenden Output nach out */
void CNeuronicNet::calc_output(CDoubleVector& in, CDoubleVector& out,bool use_scaling=false){
    //dem Netz die Inputs repsentieren
    set_input(in,use_scaling);
    //die Werte vorwrts propagieren
    feed_forward();
    //das Netz-Ergebnis auslesen
    get_output(out,use_scaling);	
}

//-----------------------------------------------------------------------------
#ifdef USE_CCOMMUNICATION
/** ausfhren eines Kommandos*/
bool CNeuronicNet::process_cmd(const char* a_command, const void* data,TPtrVector* add_data=0,long id=0){
    if (check_command(a_command,"print_weights","Gibt die Gewichte des Netzes aus.",id)){
        cout<<"Gebe die Gewichte des Netzes aus:  \n\n";
        print_input_weights();
        return true;
    }
    if (check_command(a_command,"print_output_and_error","Gibt Fehler und Output der Neuronen des Netzes aus.",id)){
        print_output_and_error();
        return true;
    }    
    if (check_command(a_command,"save","speichert das Netz ab (im Default-Format).",id)){
        const char* new_data;
        if (data) new_data=reinterpret_cast<const char*>(data);
          else new_data="net.nnd";
        assert(new_data);
        if (save(new_data)){ infos(new_data); infos(": Netz wurde gespeichert. (mutlinet)\n");}
          else infos("Netz konnte nicht gespeichert werden.\n");
        return true;
    }
    if (check_command(a_command,"save_snns","speichert das Netz als SNNS-Net-Datei",id)){
        const char* new_data;
        if (data) new_data=reinterpret_cast<const char*>(data);
          else new_data="net.net";
        assert(new_data);
        if (save(new_data,"SNNSNET")){ infos(new_data); infos(": SNNS-Netz wurde gespeichert.\n");}
          else infos("Netz konnte nicht gespeichert werden.\n");
    return true;
    }
    if (check_command(a_command,"save_nnd","speichert das Netz als NND-Datei",id)){
        const char* new_data;
        if (data) new_data=reinterpret_cast<const char*>(data);
          else new_data="net.nnd";
        assert(new_data);
        if (save(new_data),"NND"){ infos(new_data); infos(": NND-Netz wurde gespeichert.\n");}
          else infos("Netz konnte nicht gespeichert werden.\n");
        return true;
    }
    if (check_command(a_command,"load","ldt das Netz aus dem XML-File.",id)){
        const char* new_data;
        if (data)	new_data=reinterpret_cast<const char*>(data);
          else new_data="net.nnd";
        assert(new_data);
        if (load(new_data)) infos("Netz wurde geladen.\n");
          else infos("Netz konnte nicht geladen werden.\n");
        return true;
    }
    if (check_command(a_command,"layer","Gibt den Befehl an die angegebene Schicht weiter. (Syntax: layer(nr))",id)){
        char* new_command;	//der weiterzureichende Befehl
        long start,stop;		//die Element, an die der Befehl weiterzureichen ist
        if (parse_command(a_command,new_command,start,stop,0,get_anzahl_layers()-1)){
            //weiterreichen an die entsprechende Layer
            for (long i=start; i<=stop; i++){
                if (i!=start) infos(format_string("layer(%d).",i));
                layer(i)->process_cmd(new_command,data,add_data,id);
            }
        }else infos("- Fehlerhaftes Kommando.\n");
        return true;
    }
    // die Parentklasse wei hoffentlich was damit anzufangen , und falls nicht weiterleiten an die Layers leiten
    if (!CCommunication::process_cmd(a_command,data,add_data,id)){
        return process_cmd(cat_strings("layer.",a_command).c_str(),data,add_data,CCommunication::ID_NO_ERROR_MSG);
    }else return true;// return CCommunication::process_cmd(a_command,data,add_data,id);
}
#endif
//-----------------------------------------------------------------------------
/** liefert die Input-Matrizen der einzelnen Netz-Schichten */
void CNeuronicNet::print_input_weights(){
    for (long i=1; i< get_anzahl_layers();i++){
        cout<<"--------------------------------------------------\n";
        cout<<"SCHICHT "<<i<<"\n";
        cout<<"------------\n";
        layer(i)->print_input_matrix();
        cout<<"\n";
    }

}
//-----------------------------------------------------------------------------
/** gibt den Fehler der einzelnen Neuronen, die Fehleraktivierung, die Aktivierung und den Output aus*/
void CNeuronicNet::print_output_and_error(){
    for (long i=1; i< get_anzahl_layers();i++){
        cout<<"--------------------------------------------------\n";
        cout<<"SCHICHT "<<i<<"\n";
        cout<<"------------\n";
        layer(i)->print_output_and_error();
        cout<<"\n";
    }

}


#ifdef USE_LIBXML
//-----------------------------------------------------------------------------
/** speichert in das XML-File, zu dem tree gehrt */
xmlNodePtr CNeuronicNet::save_to_xml(xmlNodePtr tree){
	assert(tree);
	//einen neuen Knoten anlegen (mit dem Klassennamen)
	xmlNodePtr subtree,subtree2;
	subtree2 = xmlNewChild(tree, NULL, (const xmlChar *)"CNeuronicNet",NULL);
	//Schichten abspeichern
	for (long i=0;i<get_anzahl_layers();i++){
		layer(i)->save_to_xml(subtree2);
	}
	//Netzvariablen abspeichern
	xmlSaveLong(subtree2,"noof_input_layers",noof_input_layers);
	//Verbindungen abspeichern
	//zunchst die Positionsinformationen der Neuronen anhand der Pointer ermitteln
	generate_pos_info();
	//Alle Neuronen abklappern und die entsprechenden Synapsen abspeichern
	for ( long lay=0; lay<get_anzahl_layers(); lay++){
		for (long pos=0; pos<layer(lay)->get_anzahl_neurons();pos++){
			for (long i=0; i<layer(lay)->neuron(pos)->get_anzahl_outputs();i++){
				//die Verbindungsinformationen abspeichern
				CSynapse* syn=layer(lay)->neuron(pos)->output_synapse(i);
				assert(syn->pre_synaptic_neuron); assert(syn->post_synaptic_neuron);
				//Checken ob das andere Neuron auch zum Netz gehrt und nur dann abspeichern
		    std::map< CNeuron*, CNeuronPosInfo >::const_iterator it = pos_info.find(syn->post_synaptic_neuron);
    		if (it != pos_info.end()) {  //Gefunden
  				//Nur die Netzinternen Verbindungen abspeichern
    			CNeuronPosInfo inf=pos_info[syn->post_synaptic_neuron];
  				subtree = xmlNewChild(subtree2, NULL, (const xmlChar *)"Connection",NULL);
    			xmlSaveLong(subtree,"pre_layer",lay);
    			xmlSaveLong(subtree,"pre_pos",pos);
    			xmlSaveLong(subtree,"post_layer",inf.layer);
    			xmlSaveLong(subtree,"post_pos",inf.pos);
  				//die Informationen der Synapse abspeichern
  				syn->save_to_xml(subtree);
  			}//else Verbindung nach auerhalb wird nicht mit gespeichert
  		}
		}
	}
  return subtree;
}
//-----------------------------------------------------------------------------
/** ldt aus dem XML-File, zu dem tree gehrt */
void CNeuronicNet::load_from_xml(xmlNodePtr tree){
	assert(tree);
	//in den Knoten gehen (mit dem Klassennamen)
	reset();
	if (strcmp((const char*)tree->name, "CNeuronicNet")) throw(666);
	xmlNodePtr subtree,subsubtree;
	subtree=tree->xmlChildrenNode;
	CNeuronicLayer* nlayer;
	CSynapse* syn;
	CDoubleVector vec;
	while (subtree!=0)
	{
		nlayer=CNeuronica::new_layer(subtree);
		if (nlayer){ add_layer(nlayer);}
		else	//die Verbindungen laden				//FIXME: Achtung es mssen erst alle Schichten geladen sein, bevor die Connections geladen werden knnen
  	if (!strcmp((const char*)subtree->name,"Connection")){
  		long lay_pre,lay_post,pos_pre,pos_post;
 			//zuerst alle Verbindungsdaten auslesen
  		subsubtree=subtree->children;
  		while (subsubtree!=0){
  			xmlLoadLong(subsubtree,"pre_layer",lay_pre);
  			xmlLoadLong(subsubtree,"pre_pos",pos_pre);
  			xmlLoadLong(subsubtree,"post_layer",lay_post);
  			xmlLoadLong(subsubtree,"post_pos",pos_post); 			
  			subsubtree=subsubtree->next;
  		}
  		//FIXME: Test ob auch alle Parameter eingegangen
  		//Die Neuronen auffinden
  		if (!layer(lay_pre) || !layer(lay_post)) throw(666); //Die Schichten ex. nicht
  		CNeuron* pre_neuron=layer(lay_pre)->neuron(pos_pre);
  		CNeuron* post_neuron=layer(lay_post)->neuron(pos_post);
  		if (!pre_neuron || !post_neuron) throw(666);	//Das Neuron ex. nicht
  		//Den SubSubtree nochmals durchsuchen, um die Synapsendaten zu finden
  		subsubtree=subtree->children;
  		syn=0;
  		while (subsubtree!=0 && syn==0){
	 			syn=CNeuronica::new_synapse(subsubtree,pre_neuron,post_neuron);
  			subsubtree=subsubtree->next;
	 		}
  	}
  	//Laden der Netzeigenschaften
  	xmlLoadLong(subtree,"noof_input_layers",noof_input_layers);
		subtree=subtree->next;
	}
}
#endif
//-----------------------------------------------------------------------------
/** ldt ein Netz aus einer Datei im XML-basierten NND-Format (Neuronica Network Definition)
		oder aus dem SNNS NET-Format
*/
bool CNeuronicNet::load(const char* filename, const char* format="DEFAULT" ){
    assert(filename);
    assert(format);
#ifdef USE_LIBXML
    //LIBXML>>>
        if (!strcmp(format, "NND")||!strcmp(format, "DEFAULT")){
            xmlDocPtr doc;
            xmlNodePtr tree;

            doc= xmlParseFile(filename);
            if (doc){
                tree = doc->xmlChildrenNode;
                if (tree){
                    tree = tree->xmlChildrenNode;		//NEURONICA
                    if (tree){
                        //nach CNeuronicNet suchen und dann das Laden beginnen
                        while (strcmp("CNeuronicNet",(const char*) tree->name)) tree=tree->next;
                        try{
                            load_from_xml(tree);
                        }catch(int n){
                            cout << "Error number is " << n << endl;
                            xmlFreeDoc(doc);
                            return false;
                        }
                        xmlFreeDoc(doc);
                        return true;
                    }
                }
            }//else
            return false;
        }
    //<<<LIBXML
#endif
    if (!strcmp(format, "SNNSNET")){
        //TODO: Implementieren
        return false;
    }
    //else
    return false;
}
//-----------------------------------------------------------------------------
/** speichert das Netz */
bool CNeuronicNet::save(const char * filename, const char* format="DEFAULT"){
    assert(filename);
    assert(format);
    int ergebnis;
#ifdef USE_LIBXML
    //LIBXML>>>
        if (!strcmp(format, "NND")||!strcmp(format, "DEFAULT")){
            xmlDocPtr doc;
            //xmlNodePtr tree;
            doc = xmlNewDoc((const xmlChar *) "1.0");
            doc->children = xmlNewDocNode(doc, NULL, (const xmlChar *)"NEURONICA", NULL);
            save_to_xml(doc->children);
            ergebnis=xmlSaveFormatFile(filename,doc,true);
            xmlFreeDoc(doc);
            if (ergebnis>=0) return true;
            //else
            return false;
        }
    //<<<LIBXML
#endif
    if (!strcmp(format, "SNNSNET")){
        // Zuerst in eine temporre NND-Datei speichern
        const char* tmpfilenam=tempnam(NULL,"neuronica.nnd");
        save(tmpfilenam,"NND");
        // umwandeln ins NET-Format
        ergebnis=CNeuronica::nnd2net(tmpfilenam,filename);
        //temp. Datei lschen
        remove(tmpfilenam);
        return (ergebnis==true);
    }
    //else
    return false;
}

//-----------------------------------------------------------------------------
/** generiert Informationen zur Position des Neurons im Netz anhand
seiner Adresse */
void CNeuronicNet::generate_pos_info(){
	//Die alte pos_info lschen
	pos_info.erase(pos_info.begin(),pos_info.end());
	//und die neue generieren
	//Ein Neuron sollte nicht in mehreren Schichten sein
	for ( long lay=0; lay<get_anzahl_layers(); lay++){
		for (long pos=0; pos<layer(lay)->get_anzahl_neurons();pos++){
			CNeuronPosInfo inf;
			inf.layer=lay;
			inf.pos=pos;
			pos_info[layer(lay)->neuron(pos)]=inf;
		}
	}
}
//-----------------------------------------------------------------------------
/** fhrt einen Lernschritt mit jedem der Patterns in CPatterns durch */
void CNeuronicNet::learn_all(CPatterns& patterns, bool use_scaling=false){
    assert(patterns.inputs.size()==patterns.outputs.size());
    //
    CDoubleVector ergebnis;
    for (ulong i=0; i<patterns.inputs.size();i++)
    {
        //dem Netz die Inputs repsentieren
        set_input(patterns.inputs[i], use_scaling);
        //die Werte vorwrts propagieren
        feed_forward();
        //das Netz-Ergebnis auslesen
        get_output(ergebnis,false);
        //den Fehler ermitteln und setzen
        set_error(patterns.outputs[i]-ergebnis,use_scaling);
        //den Fehler zurck propagieren
        feed_back();
    }
}
//-----------------------------------------------------------------------------
/** liefert den Error der letzten Schicht des Netzes */
void CNeuronicNet::get_error(CDoubleVector & a_errors, bool use_scaling=false){
	output_layer()->get_error(a_errors,use_scaling);
}
//-----------------------------------------------------------------------------
/** ermittelt den summierten quadratischen Fehler des Neurons vor dem letzten Backup-Schritt */
double CNeuronicNet::get_sse(){
	CDoubleVector err;
	get_error(err);
	double sse=0;
	for (ulong i=0; i<err.size(); i++){
		sse+=(err[i]*err[i]);
	}
	return sse;
	
}
//-----------------------------------------------------------------------------
/** liefert einen Zeiger auf die Input-Schicht-Nr */
CNeuronicLayer * CNeuronicNet::input_layer(long nr){
	assert(nr<=noof_input_layers);	
	assert(nr>=0);
	return layer(nr);
}
//-----------------------------------------------------------------------------
/** liefert einen Zeiger auf die Output-Schicht */
CNeuronicLayer * CNeuronicNet::output_layer(ELayerCode nr){
	//FIXME: ev. Variable einfhren, die festlegt, welches die Outputschichten sind
  long laynr=(nr<0)? get_anzahl_layers()+nr:nr;
	assert(laynr< get_anzahl_layers() && laynr>=0);
	return layer(laynr);
}
//-----------------------------------------------------------------------------
/** ruft in allen Schichten des Netzes die Funktion
allow_learning auf. Falls true bergeben wird., wird
das Lernen fr alle Schichten des Netzes gestattet,
sonst nicht. */
void CNeuronicNet::allow_learning(bool allow){
  for (long i=0;i<get_anzahl_layers();i++)
	{
		layer(i)->allow_learning(allow);
	}
}
//-----------------------------------------------------------------------------
/** entfernt die Schicht mit der Nummer aus der Verwaltung des NEtzes
Schicht wird gelscht, wenn delete_layer true ist */
void CNeuronicNet::remove_layer(long lay_nr,  bool delete_layer=false){
    CNeuronicLayer* lay=layer(lay_nr);
    assert(lay);
    layers.erase(layers.begin()+lay_nr);
    if (delete_layer) delete lay;	
}
//-----------------------------------------------------------------------------
/** entfernt die Schicht  aus der Verwaltung des NEtzes
Schicht wird gelscht, wenn delete_layer true ist */
void CNeuronicNet::remove_layer(CNeuronicLayer* lay, bool delete_layer){
    for (long i=get_anzahl_layers()-1; i>=0;i--){
        if (layer(i)==lay) remove_layer(i,delete_layer);
    }
}
//-----------------------------------------------------------------------------
/** stellt die Lernrate aller Neuronen des Netzes auf den bergebenen Wert ein */
void CNeuronicNet::set_learning_rate(double lr){
    //TODO: direkt aufrufen, um Geschw. zu verbessern
    //die Ausgaben unterdrcken
    bool old_value=CCommunication::global_showinfo;
    CCommunication::global_showinfo=false;
    process_command("layer(*).neuron(*).learning_rate",lr);
    CCommunication::global_showinfo=old_value;
}
//-----------------------------------------------------------------------------
/** liefert die Anzahl der Input-Schichten fr ein Netz. Die Input-Schichten sind (zur Zeit) immer die ersten im Netzwerk. */
long CNeuronicNet::get_noof_input_layers(){
    return noof_input_layers;
}
//-----------------------------------------------------------------------------
/** erklrt die ersten noof_layers Schichten des Netzes zu Input-Schichten */
void CNeuronicNet::set_noof_input_layers(long noof_layers){
    assert (noof_layers <=get_anzahl_layers());
    noof_input_layers=noof_layers;
}
//-----------------------------------------------------------------------------
/** setzt min_error fr alle Neuronen der Output-Schicht.
min_error ist der kleinste Fehlerwert, der noch als Fehler betrachtet wird (SNNS: dmax) */
void CNeuronicNet::set_min_error(double min_err){
    //FIXME: besser implementieren
    output_layer()->process_command("neuron(*).min_error",min_err);
}
//-----------------------------------------------------------------------------
/** Trainiert das Netz mit allen Patterns. Jeder Pattern wird noof_steps mal trainiert.
    */
void CNeuronicNet::train(CPatterns & patterns, long noof_steps, bool use_scaling){
    for (long i=0; i<noof_steps;i++){
        learn_all(patterns,use_scaling);
    }
}
//-----------------------------------------------------------------------------
/** Trainiert das Netz mit den Inputs und den dazugehrigen Outputs. Jedes Input-Output-Paar wird
    noof_steps mal trainiert. (wie train mit Patterns)
    */
void CNeuronicNet::train(const std::vector< CDoubleVector>& inputs, const std::vector< CDoubleVector>& outputs, long noof_steps, bool use_scaling){
    assert (inputs.size()==outputs.size());
    CPatterns pat;
    pat.inputs=inputs;
    pat.outputs=outputs;
    train(pat,noof_steps,use_scaling);
}

//-----------------------------------------------------------------------------
/** Wenn true als Parameter bergeben wird, so wird bei jedem Aufruf von feed_back
logged_patterns mit den aktuellen Werten von Input und Output gefllt.  Wird false
bergeben, so wird nichts hinzugefgt.
@param delete_old Wenn dieser Parameter true ist, werden alle schon gespeicherten
Patterns in saved_patterns wieder gelscht (Standard: false) */
void CNeuronicNet::use_pattern_logging(bool logging,bool delete_old=false){
     pattern_logging=logging;
     if (delete_old){
         logged_patterns.delete_all();
         cout<<"Reset pattern logging.";
     }

}

//-----------------------------------------------------------------------------
/** muss in jeder abgeleiteten Klasse vor dem Feedback-Schritt aufgerufen werden */
void CNeuronicNet::after_feed_back(TNTimeUnit time){
    if (pattern_logging){
        CDoubleVector dv_in,dv_out,dv_err;
        //den anliegenden Input ermitteln
        //die Werte aller InputSchichten werden zu einem Vektor zusammengefasst
        //wenn mehr als eine Inputschicht ex.      
        get_output(dv_in,false,IL0);
        for (int i=1;i<noof_input_layers;i++){
            CDoubleVector dv_in2;
            get_output(dv_in2,false,ELayerCode(IL0+i));
            dv_in.push_back_doublevector(dv_in2);
        }
        //den anliegenden Soll-Output ermitteln: dieser wird nicht direkt
        //eingestellt sondern nur der Fehler. Der Soll output ergibt sich als Summe
        get_output(dv_out,false,OL0);
        get_error(dv_err,false);
        assert(dv_out.size()==dv_err.size());
        dv_out+=dv_err;
        logged_patterns.add_pattern(dv_in,dv_out);
    }
}
