/* **************************************************************************
                          CNeuronicSnnsNet.cpp  -  description
                             -------------------
    begin                : Sun Jan 7 2001
 ***************************************************************************/

 /* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "cneuronicsnnsnet.h"

#include "cneuroniclayer.h"
#include "cneuronica.h"
#include "cpatterns.h"

#ifdef USE_SNNS

//SNNS Kernel functions
extern "C" {
#include "snnskernel/glob_typ.h"
#include "snnskernel/krui_typ.h"
#include "snnskernel/kr_const.h"    /*  Constant Declarators for SNNS-Kernel  */
#include "snnskernel/kr_def.h"      /*  Default Values  */
#include "snnskernel/kr_ui.h"      	/*  PatternSet  */
#include "snns_sup.h"								/*  Hilfsdatei, damit das ganze auch mit neueren flex-Bibliotheken kompiliert*/
}

#ifdef USE_LIBXML
	#include <libxml/parser.h>
#endif



CNeuronicSnnsNet::CNeuronicSnnsNet(){
	//patternset anlegen, wird spter bentigt
	//int patset_no;
	learning_allowed=true;
	chk_err(krui_allocNewPatternSet(&patset_no));
	//FALSCH: assert(patset_no==0); //sollte eig. immer 0 sein, falls nicht muss man Var. in Klasse anlegen
}
CNeuronicSnnsNet::~CNeuronicSnnsNet(){
	//Speicher freigeben
	chk_err(krui_deletePatSet(patset_no));
}

//----------------------------------------------------------------------------------------
/** fgt dem Netz eine Schicht Neuronen hinzu
und liefert als Ergebnis die Nummer der Schicht im Net */
long CNeuronicSnnsNet::add_layer(CNeuronicLayer* a_layer){
	return CNeuronicNet::add_layer(a_layer);
}

//----------------------------------------------------------------------------------------
/** fhrt einen feed_forward schrittt ber das gesamte Netz aus */
void CNeuronicSnnsNet::feed_forward(TNTimeUnit time=0)
{
    //FIXME
    //Die Aktivierungen der Neuronen auslesen und als Aktivierung in das SNNS-Netz stecken
    set_snns_input_activations();
    //ein testweiser Aufruf von chk_err(krui_modifyPattern()) bringt keine Besserung der Probleme mit Quickprop
    //den Update-Schritt durchfhren
    //FIXME : par nicht jedesmal neu berechnen
    float par[10];   assert(params.size()<10);
    for (ulong i=0; i<params.size();i++) par[i]=(float) params[i];
    chk_err(krui_updateNet(par,params.size()));
    //Werte aus dem Netz in die Dummy-Neuronen bernehmen
    get_snns_outputs();
}
//----------------------------------------------------------------------------------------
/** Hilfsfunktion, die testet, ob ein Fehler aufgetreten ist und falls ja, diesen ausgibt
  * und eine Exception auslst*/
int CNeuronicSnnsNet::chk_err(int err){
  if (err !=KRERR_NO_ERROR){
  	cerr<<"SNNS-Error #"<<-err<<": "<<krui_error(err)<<endl;
  	throw 666;
  }
  return err;
}
//----------------------------------------------------------------------------------------
/** Hilfsfunktion zur bername der Daten in das SNNS-Netz
	* liest die Aktivierungen in den Dummy-Neuronen aus und schreibt sie in das SNNS-Netz*/
void CNeuronicSnnsNet::set_snns_input_activations(){
    for (long j=0; j<get_noof_input_layers();j++){
        assert(input_layer(j));
        input_layer(j)->feed_forward(); //berechnet die Werte fr die Input-Neuronen(falls ntig)
        for (long i=0; i<input_layer(j)->get_anzahl_neurons();i++){
            //Aktivierung fr jedes SNNS-Neuron setzen
            //Alte (falsche) Variante: 		chk_err(krui_setUnitActivation(input_layer()->neuron(i)->tag,input_layer()->neuron(i)->get_activation()));
            int unit=(int) input_layer(j)->neuron(i)->tag;
            FlintTypeParam act=(FlintTypeParam)input_layer(j)->neuron(i)->get_last_output();
            //chk_err(krui_setUnitActivation(input_layer(j)->neuron(i)->tag,input_layer(j)->neuron(i)->get_last_output()));
            //Jon DEBUG:
            //printf("DEBUG: SOLL:unit: %d, activation: %f \n",unit,act);
            //cout<<unit<<":"<<act<<"\t";
            chk_err(krui_setUnitActivation(unit,act));
            //Debug:
            //cout <<"innrn #"<<input_layer()->neuron(i)->tag<<",\t act: "<<input_layer()->neuron(i)->get_activation()<<endl;
        }
    }
    //DEBUG:
    //cout<<endl;
}
//----------------------------------------------------------------------------------------
/** Hilfsfunktion, prft ob die Daten korrekt im Netz stehen (wird nur ber assert() aufgerufen*/
bool CNeuronicSnnsNet::assert_snns_input_activations(){
    for (long j=0; j<get_noof_input_layers();j++){
        assert(input_layer(j));
        for (long i=0; i<input_layer(j)->get_anzahl_neurons();i++){
            //Aktivierung fr jedes SNNS-Input-Neuron testen
            int unit=(int) input_layer(j)->neuron(i)->tag;
            float act_snns=krui_getUnitActivation(unit);
            float act_nrn=input_layer(j)->neuron(i)->get_last_output();
            if (act_snns!=act_nrn) return false;
        }
    }
    return true;
}
//----------------------------------------------------------------------------------------
/** Hilfsfunktion zur bername der Daten aus dem SNNS-Netz
  * liest die Outputs des SNNS-Netzes in die Dummy-Neuronen ein*/
void CNeuronicSnnsNet::get_snns_outputs(){
    //den Output auslesen und in die Outputneuronen schreiben
    for (long i=0; i<output_layer()->get_anzahl_neurons();i++){
        output_layer()->neuron(i)->set_output(krui_getUnitOutput(output_layer()->neuron(i)->tag));
        //Debug
        //cout <<"outnrn #"<<output_layer()->neuron(i)->tag<<",\t outp: "<<output_layer()->neuron(i)->get_last_output()<<endl;
    }
}

//----------------------------------------------------------------------------------------
/** Hilfsfunktion
  * ermittelt anhand der Fehler, welche Outputs eigentlich htten dastehen sollen und setzt
  * diese im SNNS-Netz = Vorraussetzung frs Lernen*/
void CNeuronicSnnsNet::set_snns_soll_activations(){
    //die Aktivierungen der SNNS-Output-units so setzen, dass diese sen Soll-Wert ergeben (fr Pattern)
    for (long i=0; i<output_layer()->get_anzahl_neurons();i++){
        output_layer()->neuron(i)->calc_error_output(); //nur bentigt um die fehleraktivierung zurck zu setzen
        chk_err(krui_setUnitActivation(output_layer()->neuron(i)->tag,output_layer()->neuron(i)->get_last_output()+output_layer()->neuron(i)->get_last_error()));
    }
}

//----------------------------------------------------------------------------------------
/** fhrt einen Feed-Back-Schritt ber das gesamte Netz aus */
void CNeuronicSnnsNet::feed_back(TNTimeUnit time=0){
    //before_feed_back(time);
    //lernen funktioniert nur mit Pattern dieser sollte im Normalfall immer eingestellt sein
    //sicherheitshabler nochmal abchecken
    assert(assert_snns_input_activations()); //die Werte sollten korrekt im SNNSNetz stehen (feed_forward muss immer vor feed_back aufgerufen werden)
    //Die vorgegebenen Output-Werte (Soll-Werte) einstellen
    set_snns_soll_activations();
    chk_err(krui_modifyPattern());	//Pattern auf die aktuell eingestellten Aktivierungen setzen
    //der eigentliche Lernschritt
    //FIXME: hier mssen die Lernparameter fr den SNNS derzeit noch per Hand eingestellt werden:
    // Die Parameter sind je nach Netz unterschiedlich und mssen der SNNS-Dokumentation entnommen werden (Seite 67 ff)
    /* hier die Parameter fr die wichtigsten Netze:
    ----------------
    Std Backpropagation 1: ff learning rate (0-1),
    										2: dmax, the maximum error that is tolerated. use 0 or a small value.
    BackpropMomentum 		1: ff learning rate (0-1),
    										2: momentum term (0-0.99),
    										3: flat spot elimination (ignore)
    										4: dmax max ignored error
    Rprop 							1: starting values of Delta ij (0-0.2)
    										2: Delta max maximum update value (30..50 works well..)
    										3: exponent of weight decay 4->0.0001
    Quickprop    				1: learning parameter, specifies the step width of the gradient descent
    										2: maximum growth parameter, realizes a kind of dynamic momentum term. A typical value is 2.0.
                        3: weight decay term to shrink the weights. A typical value is 0.0001
    										4: dmax max ignored error

    */
    //FIXME : par nicht jedesmal neu berechnen
    float par[10];   assert(params.size()<10);
    for (ulong i=0; i<params.size();i++) par[i]=(float) params[i];
    float *erg;
    int erg_cnt;
    //Aufruf krui_learnSinglePatternFF(Pattern_nummer_muss_1_sein, Parameter, Anzahl_Parameter, Zeiger_auf_feld_von_ausgabe_parametern, Zeiger_auf_anzahl_ausgabeparameter)
    //if (learning_allowed) chk_err(krui_learnSinglePatternFF(1,par,params.size(),&erg,&erg_cnt));
    if (learning_allowed) chk_err(krui_learnSinglePattern(1,par,params.size(),&erg,&erg_cnt));
        else cout<<"Learning not allowed.";
    //TODO : die ursprnglichen Outputs wieder herstellen? Sollten eig. noch vorhanden sein
    after_feed_back(time);
}

//----------------------------------------------------------------------------------------
/** setzt das gesamte Netz zurck */
void CNeuronicSnnsNet::reset(bool force_delete)
{
    //das SNNS-Netz lschen
    krui_deleteNet();
    chk_err(krui_deletePatSet(patset_no));
    CNeuronicNet::reset(force_delete);
    learning_allowed=true;
    chk_err(krui_allocNewPatternSet(&patset_no));
}


//-----------------------------------------------------------------------------
#ifdef USE_CCOMMUNICATION
/** ausfhren eines Kommandos*/
bool CNeuronicSnnsNet::process_cmd(const char* a_command, const void* data,TPtrVector* add_data=0,long id=0)
{
	if (check_command(a_command,"learning_function","bestimmt die verwendete Lernfunktion fr das SNNS-Netz",id) ||
	    check_command(a_command,"lf","siehe learning_function",id))
	{
		infos("aktuell eingestellte Lernfunktion: ");
		infos(get_learning_function());
		infos("\n");
		if (data){ //sonst nur Anzeige der Funktion
  	  const char* wert=reinterpret_cast<const char*>(data);
  	  assert(wert);
  	  set_learning_function(wert);
  		infos("Die neue Lernfunktion ist nun :");
  		infos(get_learning_function());
  		infos("\n");
  	}
		return true;
	}
	// die Parentklasse weis hoffentlich was damit anzufangen
	return CNeuronicNet::process_cmd(a_command,data,add_data,id);
}
#endif


//-----------------------------------------------------------------------------
/** hier wird das SNNS-File geladen und das Netz vorbereitet*/
bool CNeuronicSnnsNet::load(const char* filename, const char* format  ){
    assert(filename);
    assert(format);
    if (!strcmp(format, "NND")){
        const char* tmpfilenam=tempnam(NULL,"neuronica.net");
        // umwandeln ins NET-Format
        if (CNeuronica::nnd2net(filename,tmpfilenam)){
            //laden aus dem .NET-Format
            bool ergebnis=load(tmpfilenam,"SNNSNET");
            //temp. Datei lschen
            remove(tmpfilenam);
            return ergebnis;
        }
        return false;
    }

    if (!strcmp(format, "SNNSNET")||!strcmp(format, "DEFAULT")){
        //Alles altes entfernen:
        reset();
        assert(layers.size()==0);

        krui_err err_nr;
        char* netname;
        //laden des Netzes
        err_nr=krui_loadNet(const_cast<char*>(filename), &netname );
        if (err_nr!=0) throw_error(666,format_string("Error Nr. %d loading file '%s'",err_nr,filename).c_str(),"CNeuronicSnnsNet::load");

        //Bestimmen der Anzahl der Input und Output-Neuronen und entsprechende Dummy-Neuronen anlegen
        //wenn es sich um ein Netz handelt, dass mit Neuronica gespeichert wurde (Name: Neuronica-SNNS-Net)
        //dann sind mehrere Input-Schichten mglich: abhngig von der Position: x gibt die Schicht an
        //sonst nur eine Schicht
        assert(layers.size()==0);
        long inps, outps;
        if (!strcmp(netname, "Neuronica-SNNS-Net")){	//es handelt sich um ein mit Neuronica gespeichertes Netz. ev.mehere Schichten
            CLongSet inp_layers;
            //FIXME ist das korrekt / oder ist getFirstUnit usw. ntig?
            //erster Durchlauf alle Schichten bestimmen
            for (long i=1;i<=krui_getNoOfUnits();i++){
                //Position bestimmen ...
                PosType pos;
                krui_getUnitPosition(i,&pos);
                //DEBUG:
                //cout << "neuron #:"<<i<<",\t Typ:"<<krui_getUnitTType(i)<<",\t Layer:"<< krui_getUnitLayerNo(i)<<"\t Pos (x,y,z):"<<pos.x<<","<<pos.y<<","<<pos.z <<endl;
                if (krui_getUnitTType(i)==INPUT){
                    //x-Wert als Schicht eintragen
                    inp_layers.insert(pos.x);
                }
            }
            //die Anzahl der Sichten bestimmen und anlegen+ Map anlegen zur spteren Zuordnung
            typedef std::map<long, long, less<long> > LongMap;
            LongMap lay_map;
            long act_layer=0;
            for (CLongSet::iterator it=inp_layers.begin();it!=inp_layers.end();it++){
                lay_map[(*it)]=act_layer++;
                CNeuronicLayer* lay=new CNeuronicLayer();
                add_layer(lay);
            }
            set_noof_input_layers(inp_layers.size());
            outps=krui_getNoOfOutputUnits();
            add_layer(CNeuron::TYP,outps);
            //FIXME ist das korrekt / oder ist getFirstUnit usw. ntig?
            //zweiter Durchlauf: Neuronen eintragen
            outps=-1;
            inps=-1;
            for (long i=1;i<=krui_getNoOfUnits();i++){
                if (krui_getUnitTType(i)==INPUT){
                    //Position bestimmen und x-Wert als Schicht eintragen
                    PosType pos;
                    krui_getUnitPosition(i,&pos);
                    LongMap::const_iterator it = lay_map.find(pos.x);
                    if (it==lay_map.end()) throw_error(666,"internal error","CNeuronicSnnsNet::load");
                    CInputNeuron* nrn=new CInputNeuron();
                    nrn->tag=i;	//Bezug auf das Neuron vermerken
                    input_layer((*it).second)->add_neuron(nrn);
                    ++inps;	//nur zum Testen
                }else if (krui_getUnitTType(i)==OUTPUT) output_layer()->neuron(++outps)->tag=i;
            }
            assert(inps==krui_getNoOfInputUnits()-1);
            assert(outps==krui_getNoOfOutputUnits()-1);
        }else{  //falls es sich nicht um ein von Neuronica erzeugtes Netz handelt: nur eine Input-Schicht
            inps=krui_getNoOfInputUnits();
            outps=krui_getNoOfOutputUnits();
            //FIXME
            add_layer(CInputNeuron::TYP,inps);
            add_layer(CNeuron::TYP,outps);

            //bestimmen der einzelnen schichten der Neuronen
            //FIXME ist das korrekt / oder ist getFirstUnit usw. ntig?
            inps=-1;
            outps=-1;
            for (long i=1;i<=krui_getNoOfUnits();i++){
                if (krui_getUnitTType(i)==INPUT) input_layer()->neuron(++inps)->tag=i;
                else if (krui_getUnitTType(i)==OUTPUT) output_layer()->neuron(++outps)->tag=i;
            }
            assert(inps==krui_getNoOfInputUnits()-1);
            assert(outps==krui_getNoOfOutputUnits()-1);
        }

        //Pattern anlegen, wird spter noch bentigt:  (feedback)
        krui_deleteAllPatterns();
        chk_err(krui_newPattern());
        assert(krui_getNoOfPatterns()==1);
        //Es werden nur normale Pattern verwendet (keine Sub-Pattern), aber das muss man dem SNNS explizit sagen
        int insize[1];
        int instep[1];
        int outsize[1];
        int outstep[1];
        insize[0]=++inps;
        instep[0]=0;
        outsize[0]=++outps;
        outstep[0]=0;
        int erg;
        chk_err(krui_DefTrainSubPat(insize,outsize,instep,outstep,&erg));
        assert(erg==1); //muss genau einen Pattern ergeben

        //da ursprnglich krui_learnSinglePatternFF zum lernen verwendet wurde, wurde hier die FF-Lernfunktion mit eingestellt. Das ist nicht mehr ntig
        //if (!strcmp(krui_getFFLearnFunc(), "Std_Backpropagation")) set_learning_function(krui_getLearnFunc());

        //Einstellen der Standard-Lernparameter
        //Achtung das Einstellen der Parameter funktioniert nicht fr alle Netze
        //eventuelle Standardvorgaben werden berschrieben
        set_snns_default_params(get_learning_function());
        return true;
    }
    return false;
}
//-----------------------------------------------------------------------------
/** speichert das Netz */
bool CNeuronicSnnsNet::save(const char * filename, const char* format){
 	assert(filename);
 	assert(format);
	if (!strcmp(format, "SNNSNET")||!strcmp(format, "DEFAULT")){
    	krui_err err_nr;
    	char*  netname="Neuronica-SNNS-Net";
    	//laden des Netzes
    	err_nr=krui_saveNet(const_cast<char*>(filename), netname );
    	if (err_nr!=0){
    		cerr<<"Error Nr. "<<err_nr<<" saving file '"<<filename<<"'"<<endl;
    	 	throw 666;
    	}
			return true;
  }
  if (!strcmp(format, "NND")){
      //TODO: Implementieren
      //Idee: Erzeuge ein CNeuronicNet, nutze die SNNS-Funktionen, um Infos ber das SNNS-Netz
      //heraus zu finden und ein gleichartiges NND-Netz zu erzeugen. Benutze danach die save(NND)-Funktion von CNeuronicNet

      CNeuronicNet tmpnet;
      //Benutze hnliches Prinzip wie beim Laden
      //mehrere Input-Schichten, aber nur eine Hidden und eine Output-Schicht mglich
      //FIXME: der x-Wert bestimmt die Schichten (korrekter wre eine Zuordnung entweder zu einer groen hiddenlayer oder nach den Connections)

                        
      return false;
  }
  //else
  return false;
}


//-----------------------------------------------------------------------------
void CNeuronicSnnsNet::allow_learning(bool allow){
	learning_allowed=allow;
}

//-----------------------------------------------------------------------------
/** stellt die Lernrate des Netzes auf den bergebenen Wert ein. */
void CNeuronicSnnsNet::set_learning_rate(double lr){
	// die Lernrate ex. nicht fr alle Lernverfahren, bzw. ist nur fr bestimmte Verfahren implementiert
	char* lf_name=get_learning_function();
	if (!strcmp(lf_name, "Std_Backpropagation")
	   || !strcmp(lf_name, "BackpropChunk")
	   || !strcmp(lf_name, "BPTT")
	   || !strcmp(lf_name, "BackpropWeightDecay")
	   || !strcmp(lf_name, "BackpropMomentum")
	   || !strcmp(lf_name, "Hebbian")
	   || !strcmp(lf_name, "Quickprop")
	   || !strcmp(lf_name, "RadialBasisLearning")
	   || !strcmp(lf_name, "BackPercolation")
	   ){
	  assert( params.size()>=1);
		params[0]=lr;
	}else
		cerr<<"Setzen der Lernrate fr die Funktion '"<<lf_name<<"' nicht untersttzt.";	
}

//-----------------------------------------------------------------------------
/** liefert den Namen der eingestellten Lern-Funktion*/
char* CNeuronicSnnsNet::get_learning_function(){
	return krui_getLearnFunc();
}
//-----------------------------------------------------------------------------
/** stellt die Lern-Funktion ein*/
void CNeuronicSnnsNet::set_learning_function(const char * lf_name){
    char lf_name_neu[200];
    strcpy(lf_name_neu,lf_name); //wegen char*->const char* (/ev. langt auch ein cast)
    chk_err(krui_setLearnFunc(lf_name_neu));
    set_snns_default_params(lf_name_neu);
    assert(!strcmp(lf_name_neu, get_learning_function()));
}
//-----------------------------------------------------------------------------
/** Hilfsfunktion, stellt die Default-Werte fr das bergebene Lernverfahren ein.*/
void CNeuronicSnnsNet::set_snns_default_params(const char* lf_name){
    if (!strcmp(lf_name, "Std_Backpropagation")){
        params.resize(2);
        params[0]=0.9;		//ff learning rate (0.1-1)
        params[1]=0;    //dmax, the maximum error that is tolerated. use 0 or a small value
    }else if (!strcmp(lf_name, "BackpropWeightDecay")){
        params.resize(4);
        params[0]=0.9;		//ff_learning_rate (0.1-1)
        params[1]=0.005;    //weight decay term, specifies how much of the old weight value is subtracted after learning. Try values between 0.005 and 0.3.
        params[2]=0;    //the minimum weight that is tolerated for a link. All links with a smaller weight will be pruned.
        params[3]=0;    //dmax, the maximum error that is tolerated. use 0 or a small value
    }else if (!strcmp(lf_name, "Quickprop")){
        params.resize(4);
        params[0]=0.2;		//ff_learning_rate (0.1-0.3)
        params[1]=2.0;    //maximum growth parameter, specifies the maximum amount of weight change (relative to 1) which is added to the current change. Typical values 1.75 ... 2.25.
        params[2]=0.0001; //weight decay term to shrink the weights.Typical values are 0.0001. Quickprop is rather sensitive to this parameter. It should not be set too large
        params[3]=0; 		  //dmax, the maximum error that is tolerated. use 0 or a small value*/
    }else if (!strcmp(lf_name, "BackpropMomentum")){
        params.resize(4);
        params[0]=0.9;		//ff_learning_rate (0.1-0.9)
        params[1]=0.5;    //momentum term, specifies the amount of the old weight change (relative to 1) which is added to the current change.Typical values are 0..1.0
        params[2]=0.1;    //flat spot elimination value, a constant value which is added to the derivative of the activation function to enable the network to pass flat spots of the error surface.Typical values of c are 0 ...0.25, most often 0.1 is used
        params[3]=0;    //dmax, the maximum error that is tolerated. use 0 or a small value
    }else if (!strcmp(lf_name, "RadialBasisLearning")){
        params.resize(4);
        params[0]=0.01;		//the learning rate used for the modification \Delta ~tj of center vectors. Common value 0.01
        params[1]=0.0;    //learning rate used for the modification of the parameters p of the base function.Usualy set to 0.0
        params[2]=0.01;   //learning rate which influences the training of all link weights that are leading to the output layer as well as the bias of all output neurons. A common value is 0.01
        params[3]=0;      //dmax, the maximum error that is tolerated. use 0 or a small value
    }else if (!strcmp(lf_name, "Rprop")){
        params.resize(3);
        params[0]=0.1;	 	//starting values for all delta_ij. Default value is 0.1
        params[1]=50.0;   //the upper limit for the update values delta_ij.The default value is 50.0
        params[2]=5;		  //the weight-decay determines the relationship between the output error and to reduction in the size of the weights.Important: denotes the exponent 4 corresponds to  1 : 10000.
    }else if (!strcmp(lf_name, "BackPercolation")){
        params.resize(3);
        params[0]=1.0;		//the learning rate, Typical value 1. Bigger values (up to 10) may also be used here.
        params[1]=1.0;    //If the error value , drops below this threshold value, the adaption according to the Backpercolation algorithm begins. Default Value??
        params[2]=0;      //dmax, the maximum error that is tolerated. use 0 or a small value
    }else if (!strcmp(lf_name, "Monte-Carlo")){
        params.resize(2);
        params[0]=-1.0;		//lower limit of weights and biases. Typical values are -10...-1
        params[1]=1.0;    //upper limit of weights and biases. Typical values are 1..10
    }else if (!strcmp(lf_name, "Sim_Ann_SS") || !strcmp(lf_name, "Sim_Ann_WTA") || !strcmp(lf_name, "Sim_Ann_WWTA")){
        params.resize(4);
        params[0]=-3.0;		//lower limit of weights and biases. Typical values are -10...-1
        params[1]=3.0;    //upper limit of weights and biases. Typical values are 1..10
        params[2]=3;		  //specifies the Simulated Annealing start temperature. Typical values are 1.0 ... 10.0.
        params[3]=0.9999; //degradation term of the temperature: Tnew = Told*deg. Typical values are 0.99 ... 0.99999.
    }else {
        params.resize(0);
        cerr<<"Setzen der Lernparameter fr '"<<lf_name<<"' nicht implementiert. Manuell einstellen!"<<endl;
    }
}




#endif //USE_SNNS