/* **************************************************************************
                          cneuronicsnnsnet.h  -  description
                             -------------------
    begin                : Sun Jan 7 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ************************************************************************** */
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ************************************************************************** */


#ifndef CNEURONICSNNSNET_H
#define CNEURONICSNNSNET_H

//Probleme beim Kompilieren von SNNS
//#ifndef __cplusplus 	typedef int     bool; #endif in glob_typ.h
//symtab.c #include <values.h> aufnehmen (for MAXFLOAT)
//./snnskernel_/libkernel.a ./snnskernel_/libfunc.a in Makefile aufnehmen :
//neuronica_LDADD   = ./snnskernel/libkernel.a ./snnskernel/libfunc.a  ./snnskernel/yywrap_stub.a



#include "cneuronicnet.h"
#include "cneuroniclayer.h"


#ifdef USE_SNNS
/** BETA: implementiert ein Neuronales Netz auf Basis des SNNS.

    Zur Implementation wird das Kernel Function Interface des SNNS verwendet. Die Quellcodes
    dieses Kernels werden in das Projekt mit einkompiliert.
    Das Laden der SNNS-Net-Datei wird ber das Kernel-Interface realisiert. Gleichzeitig
    wird eine Schicht von Input-Neuronen und eine Schicht von Output-Neuronen angelegt.
    Der Inhalt dieser Neuronen wir mit denen in der SNNS-Datei abgeglichen und bei allen
    weiteren Aktionen immer aktualisiert.
    Da die meisten Funktionen, die das Netz betreffen nur auf die In- und Output-Schicht zugreifen,
    wirkt das Netz nach auen wie jede andere CNeuronicNet-Klasse.

    Nicht alle Funktionen der Basisklasse @ref CNeuronicNet sind im Moment auch in dieser Klasse implementiert.
    Funktionen, die zur Verfgung stehen sind u.a. @ref load, @ref feed_forward und @ref feed_back, sowie die Funktionen
    zum Setzen der Inputs und Auslesen der Outputs.

    Damit ein Lernverfahren des SNNS benutzt werden kann muss der jeweilige Algorithmus mit den
    in dieser Klasse verwendeten Restriktionen kompatibel sein.
    Voll untersttzt und erfolgreich getestet sind:
    <li> Std_Backpropagation        Standard-Backprop-Verfahren ("vanilla backpropagation")
    <li> BackpropMomentum           Backpropagation mit momentum term and flat spot elimination
    <li> BackPercolation            lt. Doku sollte Act_TANH_XDiv2 als Akt.Funktion und Random_Weights_Perc als Gewichts-Init verwendet werden (funktioniert aber auch mit TANH, normalem init)
    </li>

    Untersttzt, aber Probleme beim Test. (muss noch untersucht werden):
    <li> BackpropWeightDecay        Backpropagation mit Weight Decay.
    <li> Quickprop
    <li> Monte-Carlo
    <li> Rprop                      Resilent Backpropogation
    <li> Sim_Ann_SS                 Simulated Annealing
    <li> Sim_Ann_WTA                Simulated Annealing
    <li> Sim_Ann_WWTA               Simulated Annealing
    </li>

    ber den SNNS stehen insgesamt folgende Lernverfahren zur Verfgung, die Lernparameter mssen
    bei nicht untersttzten Verfahren per Hand eingegeben werden.
    ART1                  Counterpropagation                Quickprop
    ART2                  Dynamic_LVQ                       RadialBasisLearning
    ARTMAP                Hebbian                           RBF-DDA
    BackPercolation       JE_BP                             RM_delta
    BackpropBatch         JE_BP_Momentum                    Rprop
    BackpropChunk         JE_Quickprop                      Sim_Ann_SS
    BackpropMomentum      JE_Rprop                          Sim_Ann_WTA
    BackpropWeightDecay   Kohonen                           Sim_Ann_WWTA
    BPTT                  Monte-Carlo                       Std_Backpropagation
    BBPTT                 PruningFeedForward                TimeDelayBackprop
    CC                    QPTT                              TACOMA


    Beispiel fr die Anwendung:
    <pre>
      //Netz laden
      CNeuronicSnnsNet* snet=new CNeuronicSnnsNet();
      snet->load("xor.net");
      //Laden der Patterns
      std::vector<CDoubleVector> inputs;
      std::vector<CDoubleVector> outputs;
      CNeuronica::load_patterns("xor.pat",inputs,outputs);
      //Training
      for (long j=0; j < 5000; j++)
      {
        for (ulong i=0; i < inputs.size(); i++ )
        {
          snet->set_input(inputs[i],false);     //dem Netz die Inputs repsentieren
          snet->feed_forward();                 //die Werte vorwrts propagieren
          snet->get_output(ergebnis,false);     //das Netz-Ergebnis auslesen
          snet->set_error(outputs[i]-ergebnis,false);   //den Fehler ermitteln und setzen
          snet->feed_back();                    //den Fehler zurck propagieren
          if ((j % 500) ==0) cout << i+1 << ". Fall (" << inputs[i] << ") :" << ergebnis << endl;
        }
      }
    </pre>


    <b> nicht untersttzte Funktionen der Basisklasse:</b>
    <li> @ref layer - nicht geplant
    <li> @ref connect - nicht geplant
    <li> @ref get_anzahl_layers - geplant, ACHTUNG auch wird intern verwendet (z.B. output_layer())
    <li> @ref load_from_xml, @save_to_xml - nicht geplant
    </li>

    <b> untersttzte Funktionen der Basisklasse:</b>
    <li> @ref input_layer - nur fr Default-Parameter
    <li> @ref output_layer - nur fr Default-Parameter
    <li> @ref get_output
    <li> @ref set_error
    <li> @ref set_input
    <li> @ref reset
    <li> @ref process_command
    <li> @ref allow_learning
    <li> @ref load, @ref save
    <li> @ref feed_forward, @ref feed_back
    </li>

    <b> TODO: </b>
    <li> Alle Funktionen der Basisklasse implementieren
    <li> Testen
    <li> Zur Zeit ist alles auf Feed-Forward-Netze zugeschnitten. Ev.andere Netze ausprobieren und untersttzen
    <li> auch das direkte erstellen von Netzen ermglichen (zur Zeit ist nur das Laden mglich)
    <li> Laden/Speichern von Netzen im NND-Format
    <li> Untersttzung fr Cascade-Correlation / TACOMA
    <li> Init-Funktionen untersttzen
    <li> Untersttzung zur Auswahl der Aktivierungsfunktion
    <li> Untersttzung von Batch-Lernverfahren. Zur Zeit wird immer nur ein Pattern prsentiert
    </li>

    <b> BUGS: </b>
    <li> es ist nur eine Instanz des SNNS-Netzes mglich. Fehler werfen, falls versucht wird eine zweite Instanz zu ffnen
    <li> eine nderung der Lernrate mit jedem Schritt ist nicht implementiert (ber Neuronica)
    <li> eventuell in der Netz-Datei voreingestellten Lern-Parameter werden ignoriert, stattdessen werden beim Laden Standardwerte eingestellt 				 (fr untersttzte Lernverfahren). Fr nicht untersttzte Lernverfahren mssen die Parameter (@ref params) manuell vorgegeben werden.
    </li>

    @author Jon Hennig
  */
class CNeuronicSnnsNet : public CNeuronicNet{
public:
    /** erzeugt ein leeres Netz
    		*/
    CNeuronicSnnsNet();
    virtual ~CNeuronicSnnsNet();

    //ist implementierbar, wenn Bedarf besteht - nicht geplant
    //CNeuronicLayer* layer(long nr);

    /* liefert die Zahl der Schichten des Netzes */  //geplant
    //long get_anzahl_layers();

    /** fhrt einen Feed-Back-Schritt ber das gesamte Netz aus.
    		*/
    void feed_back(TNTimeUnit time=0);
    /** fhrt einen Feed-Forward Schritt ber das gesamte Netz aus.
    		*/
    void feed_forward(TNTimeUnit time=0);

    /** lscht das gesamte Netz
    		*/
    void reset(bool force_delete=true);

    #ifdef USE_CCOMMUNICATION
      /** Ausfhren eines Kommandos
      	*/
      virtual bool process_cmd(const char* command, const void* data,TPtrVector* add_data=0,long id=0);
    #endif

    /** ldt das Netz aus einer Datei im SNNS-NET-Format. Das Laden aus dem NND-Format wird noch nicht untersttzt
    		@param format "SNNSNET" fr das SNNS-NET-Format
    		@return true, falls erfolgreich
    		*/
    virtual bool load(const char* filename, const char* format ="DEFAULT");

    /** speichert in eine Datei im SNNS-NET-Format. Das Speichern im NND-Format wird noch nicht untersttzt.
    		Achtung, bei der Speicherung im SNNS-Net-Format wird eine eventuelle Skalierung nicht mit abgespeichert.
    		Allgemein knnen nur die Parameter abgespeichert werden, die im SNNS-NET-Format vorgesehen sind.
        Standardmig wird immer nur die erste Schicht als Input-Schicht deklariert. Wenn das Netz aber aus
        Neuronica heraus gespeichert wurde erhlt es den Namen Neuronica-SNNS-Net und es knnen mehrere Schichten
        als Input-Schichten deklariert werden. Eine Schicht wird dann als Input-Schicht betrachtet, wenn zwischen den
        vorherigen Input-Schichten und der aktuellen Schicht kein Freiraum ist. (Bsp: 1 2 4 5 6 8, 1 und 2 sind Input-
        Schichten)
    		@param format "NND" steht fr das NND-Format "SNNSNET" fr das SNNS-NET-Format
    		@return true, falls erfolgreich
    		*/
     virtual bool save(const char * filename, const char* format="DEFAULT");

    /** wenn true bergeben wird, so wird das Lernen erlaubt, sonst nicht
    		*/
    virtual void allow_learning(bool allow);

    /** stellt die Lernrate des Netzes auf den bergebenen Wert ein.
    		*/
    virtual void set_learning_rate(double lr);

    /** liefert den Namen der eingestellten Lern-Funktion
    		@see set_learning_function
    		*/
    char* get_learning_function();

    /** stellt die Lern-Funktion ein. Die Lern-Parameter (auch die Lernrate) werden auf die Standard-Werte zurckgesetzt.
        @param lf_name gibt die SNNS-Lern-Funktion an und ist ein Wert aus:
    	ART1									Counterpropagation								Quickprop
    	ART2 									Dynamic_LVQ 											RadialBasisLearning
    	ARTMAP 								Hebbian 													RBF-DDA
    	BackPercolation 			JE_BP 														RM_delta
    	BackpropBatch 				JE_BP_Momentum 										Rprop
    	BackpropChunk					JE_Quickprop 											Sim_Ann_SS
    	BackpropMomentum 			JE_Rprop 													Sim_Ann_WTA
    	BackpropWeightDecay 	Kohonen 													Sim_Ann_WWTA
    	BPTT 									Monte-Carlo 											Std_Backpropagation
    	BBPTT 								PruningFeedForward 								TimeDelayBackprop
    	CC 										QPTT 															TACOMA
    	*/
    void set_learning_function(const char * lf_name);

    
protected: // Protected functions
	/** Hilfsfunktion zur bername der Daten in das SNNS-Netz
		* liest die Aktivierungen in den Dummy-Neuronen aus und schreibt sie in das SNNS-Netz
		*/
	void set_snns_input_activations();

  /** Hilfsfunktion, prft ob die Daten korrekt im Netz stehen (wird nur ber assert() aufgerufen)
      */
  bool assert_snns_input_activations();
 
	
	/** Hilfsfunktion zur bername der Daten aus dem SNNS-Netz
		* liest die Outputs des SNNS-Netzes in die Dummy-Neuronen ein
		*/
	void get_snns_outputs();
	
	/** Hilfsfunktion, ermittelt anhand der Fehler, welche Outputs eigentlich htten dastehen sollen und setzt
  	* diese im SNNS-Netz als Aktivierungen= Vorraussetzung um Pattern frs Lernen zu erzeugen
  	*/
	void set_snns_soll_activations();
	
	/** Hilfsfunktion, die testet, ob ein Fehler aufgetreten ist und falls ja, diesen ausgibt
  	* und eine Exception auslst
  	*/
	int chk_err(int err);
	
	/** Hilfsfunktion, stellt die Gre des Parameter-Arrays und die Default-Lern-Parameter fr das bergebene Lernverfahren ein.
			*/
	void set_snns_default_params(const char* lf_name);

	
private: //Private functions

  /** Funktion privat machen, da sie nur noch intern verwendet werden darf
  		*/
  template <class T> long add_layer(T* mother,long anzahl){
  	return CNeuronicNet::add_layer(mother,anzahl);
  }

  /** darf so hier nur intern verwendet werden
  		*/
  long add_layer(CNeuronicLayer* a_layer);


protected: // Protected attributes
  /** true, falls lernen erlaubt, muss hier abgespeichert werden, da im SNNS nur global mgl.
      */
  bool	learning_allowed;
  /** speichert welches Patternset allokiert wurde
      */
  int patset_no;


};

#endif //#ifdef SNNS
#endif //#ifndef CNEURONICSNNSNET_H
