/***************************************************************************
                          cpatterns.cpp  -  description
                             -------------------
    begin                : Fri Dec 7 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "cpatterns.h"
#include <fstream.h>

CPatterns::CPatterns(){
}
//------------------------------------------------------------------
CPatterns::~CPatterns(){
}
//------------------------------------------------------------------
/**	lscht alle Patterns*/
void CPatterns::delete_all(){
	inputs.resize(0);
	outputs.resize(0);
}

//------------------------------------------------------------------
/** ldt aus einer Datei im SNNS-Pat-Format die In- und Outputs fr ein Netz. */
//FIXME: Die korrekte Spezifikation fr die SNNS-Pattern-Files nachschlagen und implementieren
bool CPatterns::load_patterns(const char* snns_pat_filename){
    delete_all();

    //ffnen der Datei
    istream* in = &cin;
    if (snns_pat_filename!="-"){
        in = new ifstream(snns_pat_filename);
        if (!*in) throw 666; //Patternfile not found;
    }

    enum EPatternStates{ PS_START,PS_IN,PS_OUT}; //PS_IN =im Bereich der Input-Pattern, PS_Out im Bereich der Output Patterns
    EPatternStates state=PS_START;
    CString zeile;

    //Einlesen der Daten (bis zum bitteren Ende)
    while (getline(*in,zeile)){
    //Einlesen der Allgemeinen Informationen
    if (state==PS_START){
        if (zeile.find("#")==0){ //Falls die Zeile mit # beginnt umschalten
            if (zeile.find("nput")==NOT_FOUND) cout<<"warning: comment \"Input\" expected.\n";
            state=PS_IN;
            inputs.push_back(CDoubleVector());
        }
        //ansonsten ignorieren
    }else if (state==PS_IN){
        if (zeile.find("#")==0){ //Falls die Zeile mit # beginnt
            if (zeile.find("utput")==NOT_FOUND) cout<<"warning: comment \"Output\" expected.\n";
            state=PS_OUT;
             outputs.push_back(CDoubleVector());
        }else{
            //if (!is_numeric(zeile[0])) throw 666;		//Falsches Foramt: Zeile muss mit Zahl beginnen
            CDoubleVector vec(zeile);
            assert(inputs.size()>0);
            inputs[inputs.size()-1].push_back_doublevector(vec);
        }
    }else if (state==PS_OUT){
        if (zeile.find("#")==0){ //Falls die Zeile mit # beginnt
            if (zeile.find("nput")==NOT_FOUND)  cout<<"warning: comment \"Input\" expected.\n";
            state=PS_IN;
            inputs.push_back(CDoubleVector());
            //}else if (zeile.find("Output pattern")!=NOT_FOUND) throw 666; //falches Dateiformat
        }else{
            //Werte einlesen
            CDoubleVector vec(zeile);
            assert(outputs.size()>0);
            outputs[outputs.size()-1].push_back_doublevector(vec);
        }
    }
    }
    /**/
    //Schlieen der Datei
    if (in != &cin) delete in;

    //berprfen der Ergebnisse
    if (inputs.size()!= outputs.size()){
        cerr<<"Error while parsing pattern file.";
        return false;
    }else return true;
}

//------------------------------------------------------------------
/** schreibt die Muster in ein SNNS-Pattern-File */
bool CPatterns::save_patterns(const char* filename){
    assert( inputs.size()==outputs.size());
    if (inputs.size()<1) throw_error( 666,"Patterns cannot be saved. No patterns available.","CPatterns::save_pattern");

    //Datei anlegen
    ostream* out=&cout;
    if (filename!="-")
        out = new ofstream (filename);
    if (!out) throw_error( 666,"File could not be opened for writing.","CPatterns::save_pattern");

    (*out) <<"SNNS pattern definition file V3.2\n"
         "generated at <time>\n\n"
         "\nNo. of patterns : "<<inputs.size()
         <<"\nNo. of input units : " <<inputs[0].size()
         <<"\nNo. of output units : "<<outputs[0].size()<<"\n\n";

    for (ulong i=0; i<inputs.size(); i++) save_one_pattern((*out),inputs[i],outputs[i],i+1);

    //Datei schlieen
    if (out != &cout)
        delete out;
    return true;
}

//------------------------------------------------------------------
/** Hilfsfunktion, schreibt zusammengehrenden In-und Output in ein SNNS-Pattern-File */
bool CPatterns::save_one_pattern(ostream& out, CDoubleVector& input, CDoubleVector& output,long pos=0){
    assert(input.size()>0);
    assert(output.size()>0);
    out<<"# Input pattern ";
    if (pos!=0) out<<pos;
    out<<" :\n";
    for (ulong i=0; i<=((input.size()-1)/10);i++){
        input.write_partially(out,i*10,((i*10+10)<input.size())? i*10+10 :input.size()-1);
        out<<"\n";
    }

    out<<"# Output pattern ";
    if (pos!=0) out<<pos;
    out<<" :\n";
    for (ulong i=0; i<=((output.size()-1)/10);i++){
        output.write_partially(out,i*10,((i*10+10)<output.size())? i*10+10 :output.size()-1);
        out<<"\n";
    }

    return true;
}

//------------------------------------------------------------------
/** Fgt einen Pattern zur Aufnahme hinzu. Es erfolgt kein Check, auf die korrekten 
Gren der einzelnen Vektoren */
void CPatterns::add_pattern( CDoubleVector& input, CDoubleVector& output){
    inputs.push_back(input);
    outputs.push_back(output);
}
//------------------------------------------------------------------
/** erzeugt eine neue zufllige Reihenfolge der Muster O(n) */
void CPatterns::shuffle(){
    assert(inputs.size()==outputs.size());
    //einen Vektor mit Zahlen 0 bis inputs.size() erstellen
    std::vector<int> ivec(inputs.size());
    iota (ivec.begin(), ivec.end(), 0);   //zuweisen 0, 1, 2, ....
    //Zufllig durcheinanderwirbeln
    //cout<<"Debug: shuffling index vector..."<<endl;
    random_shuffle(ivec.begin(), ivec.end());
    //Auf Basis dieser Permutation neue Vekoren erstellen:
    //cout<<"Debug:generating new sorting order based on index vector..."<<endl;
    std::vector< CDoubleVector> tmp_in(inputs.size());
    std::vector< CDoubleVector> tmp_out(inputs.size());
    for( ulong i=0; i<inputs.size(); i++){
        tmp_in[i]=inputs[ivec[i]];
        tmp_out[i]=outputs[ivec[i]];
    }
    assert(inputs.size()==tmp_in.size());
    inputs=tmp_in;
    outputs=tmp_out;
}

