/* **************************************************************************
                          cpatterns.h  -  description
                             -------------------
    begin                : Fri Dec 7 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CPATTERNS_H
#define CPATTERNS_H

#include "global.h"
#include "tools.h"			//wegen CString, #defines

#include "cdoublevector.h"

/** Die Muster zum trainieren fr ein Netz.

		Mit dieser Klasse knnen die Vektoren, die einem Netz beim Training prsentiert werden, verwaltet
		werden. Besteht aus den Inputs fr das Netz (CDoubleVector) und den zugehrigen vorgegebenen
		Outputs (CDoubleVector). Das Laden/Speichern aus einem SNNS-Pattern-File ist mglich.
	
		
	  <b> TODO: </b>
	  <li> bessere Integration mit CNEuronicNet (s. @ref CNeuronicNet::learn_all)
	  <li> Aufteilung in Traings-, Validation- und Test-Set
	  </li>
	
	  <b> BUGS: </b>
		<li> nicht alle SNNS-Pattern Files gem Spezifikation werden korrekt eingelesen 
		</li>
		
		@author Jon Hennig
  */
class CPatterns {
public: 
	CPatterns();
	~CPatterns();
	
	/** lscht alle angelegten Patterns
			*/
	void delete_all();
	
	/** ldt aus einer Datei im SNNS-Pat-Format die Patterns (In- und Outputs fr ein Netz).
			*/
	bool load_patterns(const char* snns_pat_filename);
	
	/** schreibt die Muster in ein SNNS-Pattern-File
      */
  bool save_patterns(const char* filename);

  /** Fgt einen Pattern zur Aufnahme hinzu. Es erfolgt kein Check, auf die korrekten 
      Gren der einzelnen Vektoren
      */
  void add_pattern( CDoubleVector& input, CDoubleVector& output);

  /** erzeugt eine neue Zufllige Reihenfolge der Muster O(n). Vor dem Aufruf
      dieser Funktion muss der Zufallszahlengenerator mit srand bzw. srand48
      initialisiert werden (abh. von STL-Define: __STL_NO_DRAND48)
      */
  void shuffle();

protected:
	/** Hilfsfunktion, schreibt zusammengehrenden In-und Output in ein SNNS-Pattern-File
			*/
	bool save_one_pattern(ostream& out, CDoubleVector& input, CDoubleVector& output,long pos=0);

public:
    std::vector< CDoubleVector> inputs;
    std::vector< CDoubleVector> outputs;
};

#endif
