/***************************************************************************
                          cstochasticbackpropneuron.cpp  -  description
                             -------------------
    begin                : Fri Jun 8 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cstochasticbackpropneuron.h"

#include "csynapse.h"
#include "cdoublevector.h" //for Debug

//-----------------------------------------------------------------------------
CStochasticBackpropNeuron::CStochasticBackpropNeuron(){
    temperature=1;
    temperature.set_goal(0);
    temperature.set_change_factor(0.99997);
}
//-----------------------------------------------------------------------------
CStochasticBackpropNeuron::~CStochasticBackpropNeuron(){}
//-----------------------------------------------------------------------------
CStochasticBackpropNeuron* CStochasticBackpropNeuron::TYP=0;

#ifdef STOCHASTIC_MODE_GAUSS
//-----------------------------------------------------------------------------
TNPreSynapticValue CStochasticBackpropNeuron::calc_output(){
    double erg;
    //Berechnen des normalen Backprop-Outputs
    erg=CBackpropNeuron::calc_output();
    //auf Basis des normalen Outputs den Output fr das Neuron berechnen,
    //dabei die Standard-Grenzen beachten (z.B. (-1,1) fr tanh
    output=get_random_gauss(erg,temperature,get_min_output(),get_max_output());
    temperature.step();
    return output;
}

#endif


#ifdef USE_CCOMMUNICATION
//-----------------------------------------------------------------------------
/** ausfhren eines Kommandos*/
bool CStochasticBackpropNeuron::process_cmd(const char* a_command, const void* data,TPtrVector* add_data=0,long id=0){
    if (check_command(a_command,"temperature","bestimmt die Temperatur (d.h. Varianz fr die Gauss-Fkt.) T=0->deterministisch",id) ||
        check_command(a_command,"temp","siehe temperature",id))
    {
        if (data){
            double wert=get_doub(data,add_data);
            temperature=wert;
        }
        infos(CString("Temperatur ist")+(data?" nun :":":"));
        infos(temperature);
        infos("\n");
        return true;
    }
    if (check_command(a_command,"temperature_goal","bestimmt den Wert, auf den die Temperatur konvergieren soll.",id) ||
        check_command(a_command,"temp_goal","siehe temperature_goal",id))
    {
        double wert=get_doub(data,add_data);
        temperature.set_goal(wert);
        infos("Konvergenzziel fr die Temperatur ist nun :");
        infos(wert);
        infos("\n");
        return true;
    }
    if (check_command(a_command,"temperature_factor","Der Faktor mit dem sich die Temperatur dem Konvergenzziel nhert. (blicherweise >0.99 und <1)",id) ||
        check_command(a_command,"temp_factor","siehe temperature_factor",id))
    {
        double wert=get_doub(data,add_data);
        temperature.set_change_factor(wert);
        infos("Der Faktor fr die Temperatur ist nun :");
        infos(wert);
        infos("\n");
        return true;
    }
    // die Parentklasse weis hoffentlich was damit anzufangen
    return CBackpropNeuron::process_cmd(a_command,data,add_data,id);
}
#endif

#ifdef USE_LIBXML
//-----------------------------------------------------------------------------
/** speichert in das XML-File, zu dem tree gehrt */
xmlNodePtr CStochasticBackpropNeuron::save_to_xml(xmlNodePtr tree){
    assert(tree);
    //einen neuen Knoten anlegen (mit dem Klassennamen)
    xmlNodePtr subtree;
    subtree = xmlNewChild(tree, NULL, (const xmlChar *)"CStochasticBackpropNeuron",NULL);
    xmlSetProp(subtree, (const xmlChar *)"type",(const xmlChar *) "neuron");	//es handelt sich um ein Element vom Typ Neuron
    //Zuerst die Basisklasse abspeichern
    CBackpropNeuron::save_to_xml(subtree);
    xmlSaveCProp(subtree,"temperature",temperature);
    return subtree;
}
//-----------------------------------------------------------------------------
/** ldt aus dem XML-File, zu dem tree gehrt */
void CStochasticBackpropNeuron::load_from_xml(xmlNodePtr tree){
    assert(tree);
    //in den Knoten gehen (mit dem Klassennamen)
    if (strcmp((const char*)tree->name, "CStochasticBackpropNeuron")) throw(666);
    xmlNodePtr subtree;
    subtree=tree->xmlChildrenNode;
    while (subtree!=0){
        if (!strcmp((const char*)subtree->name,"CBackpropNeuron")) CBackpropNeuron::load_from_xml(subtree);
        xmlLoadCProp(subtree,"temperature",temperature);
        subtree=subtree->next;
    }
}
#endif






//zur Zeit nicht mehr benutzt:
#ifdef STOCHASTIC_MODE_WEIGHTS
    //mit einer Wahrscheinlickeit von 1:1/(temperature+1) wird etwas zuflliges ausgefhrt
    // Temperatur 1 heit mit 50 % Wahrscheinlichkeit falsches ausfhren
/** Updated die Neuronen entsprechend der Fehleraktivierung */
void CStochasticBackpropNeuron::learn_step(){
    switch (stochastic_mode)


    //Fehler bestimmen
    double eo=get_last_error_output();
    //ndere die Gewichte zu den Vorgngerneuronen
    for (long i=0;i<get_anzahl_inputs();i++){
        assert(input_neuron(i));
        double delta_weight=learning_rate * eo * input_neuron(i)->get_last_output();
        //Falls der Zufall es so will, genau in die andere Richtung ndern:
        if( ! wuerfeln(1.0/(temperature+1.0))){
            delta_weight=-delta_weight;
        }
        input_synapse(i)->add_weight(delta_weight);
    }
    //ndere die Schwelle
    double delta_schwelle=learning_rate*eo*1;
    cout<<" D"<<long(delta_schwelle*1000)<<endl;
    schwelle+=delta_schwelle;
    learning_rate.step();
    temperature.step();
}

/** liefert mit Wahrscheinlichkeit p true */
bool CStochasticBackpropNeuron::wuerfeln(double p){
    assert(p>=0 && p<=1);
    long rand_nr=rand(); //liefert eine Zufallszahl zwischen 0 und RAND_MAX
    if (rand_nr>((double)RAND_MAX*p)+1)
        return false;
    else
        return true;
}


#endif