/* **************************************************************************
                          cstochasticbackpropneuron.h  -  description
                             -------------------
    begin                : Fri Jun 8 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/



#ifndef CSTOCHASTICBACKPROPNEURON_H
#define CSTOCHASTICBACKPROPNEURON_H

//#define STOCHASTIC_MODE_WEIGHTS   //ndert die Gewichte stochastisch (alte Variante)
#define STOCHASTIC_MODE_GAUSS     //bestimmt den Output nach der Gauss'schen Fehlerverteilung



#include "cbackpropneuron.h"

/** BETA: Ein Backprop-Neuron, das seine Gewichte nach
		einer stochastischen Lernregel anpasst. D.h. die Gewichte
		werden abhngig von einer fiktiven Temperatur anders,
		als eigentlich durch das Lernverfahren vorgegeben gendert.
		
		Noch in der Entwicklung. Nur mit uerster Vorsicht verwenden.

	  <b> TODO: </b>
	  <li> Testen
	  </li>
	
	  <b> BUGS: </b>
		<li> Implementierung entspricht nicht den Vorgaben aus der entsprechenden Literatur.
		</li>
		
    @author Jon Hennig
*/

class CStochasticBackpropNeuron : public CBackpropNeuron  {
public:
//  enum EStochasticMode{
//      SM_RANDOM_WEIGHT=0,         //die Gewichte werden je nach eingestellter Temp. genau entgegengesetzt angepasst
//      SM_RANDOM_GAUSS_OUTPUT=1    //das Output des Neurons wird durch die Normalverteilung bestimmt, die Temperatur entspricht der Varianz
//  }

	CStochasticBackpropNeuron();
	~CStochasticBackpropNeuron();

#ifdef STOCHASTIC_MODE_WEIGHTS
  /** Updated die Verbindungen entsprechend der Fehleraktivierung
  		*/
  virtual void learn_step();
 protected:
  /** Hilfsfunktion, liefert mit Wahrscheinlichkeit p true
  		*/
  bool wuerfeln(double p);
 public:
#endif


#ifdef USE_CCOMMUNICATION
  /** Ausfhren eines Kommandos. Folgende stehen zur Verfgung: <b>
      <li> "temperature" ... bestimmt die Temperatur T=0->deterministisch T=1 -> 50% deterministisch
      <li> "temperature_goal" ... bestimmt den Wert, auf den die Temperatur konvergieren soll."
      <li> "temperature_factor" ... Der Faktor mit dem sich die Temperatur dem Konvergenzziel nhert. (blicherweise >0.99 und <1)
      </li>
      @see CCommunication::process_cmd
      */
  bool process_cmd(const char* a_command, const void* data,TPtrVector* additional_data=0,long id=0);
#endif

#ifdef USE_LIBXML
  /** speichert in das XML-File, zu dem tree gehrt
      */
  virtual xmlNodePtr save_to_xml(xmlNodePtr tree);

  /** ldt aus dem XML-File, zu dem tree gehrt
      */
  virtual void load_from_xml(xmlNodePtr tree);
#endif

  /** die Temperatur bestimmt die Stochastik
      */
  CChangingProperty temperature;

  /** diese Variable wird nur bentigt, um den Typ dieses Neurons in add_neurons von CNeuronicLayer
      bestimmen zu knnen.
      */
  static CStochasticBackpropNeuron* TYP;
  

protected:
#ifdef STOCHASTIC_MODE_GAUSS 
  /** berechnet und setzt den Output des Neurons basierend auf der Aktivierung, und der Gauss'schen Normalverteilung.
      Die Temperatur bestimmt die Varianz. Mit der Temperatur gegen 0 geht der Output gegen den eines Standard-
      Backprop-Neurons.
      */
  virtual TNPreSynapticValue calc_output();
#endif
};

#endif
