/***************************************************************************
                          csynapse.cpp  -  description
                             -------------------
    begin                : Thu Feb 15 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
 /***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/



#include "csynapse.h"
#include "global.h"
#include "tools.h"			//XML


#ifdef USE_EXTENDED_SYNAPSE
	long CSynapse::anz_delta_weight_avg=5;
#endif

CSynapse::CSynapse(CNeuron* pre_neuron, CNeuron* post_neuron, TNSynapticWeight a_weight)
{
	assert(pre_neuron); assert(post_neuron);
	
	pre_synaptic_neuron=pre_neuron;
	post_synaptic_neuron=post_neuron;
	weight=a_weight;
	#ifdef USE_EXTENDED_SYNAPSE
		delta_weight_avg=0;
	#endif
	//Nun den beiden Neuronen mitteilen, das zwischen Ihnen eine neue Synapse ex.
	pre_synaptic_neuron->add_output(this);
	post_synaptic_neuron->add_input(this);
}

CSynapse::~CSynapse(){
}

/** liefert eine Erregung an die entsprechende Synapse */
void CSynapse::process_input(TNPreSynapticValue a_input)
{
	post_synaptic_neuron->add_excitation(get_post_synaptic_value(a_input));
}
/** liefert zu Signal, dass die Synapse erhlt den entsprechenden postsynaptischen Wert.
Diese Funktion ist virtuell und kann fr verschiedene Synapsenarten unterschiedlich sein. */
TNPostSynapticValue CSynapse::get_post_synaptic_value(TNPreSynapticValue a_input)
{
	return weight*a_input;
}
/** ndert das Gewicht der Synapse auf weight) */
void CSynapse::set_weight(TNSynapticWeight new_weight)
{
	weight=new_weight;
}
/** liefert das Gewicht der Synapse */
TNSynapticWeight CSynapse::get_weight()
{
	return weight;
}


/** erhht das Gewicht um to_add, falls to_add kleiner 0 wird das Gewicht entsprechend vermindert. */
void CSynapse::add_weight(TNSynapticWeight to_add)
{	weight+=to_add;
	#ifdef USE_EXTENDED_SYNAPSE
		delta_weight_avg+=to_add/(double)anz_delta_weight_avg;
	#endif
}


#ifdef USE_EXTENDED_SYNAPSE
/** Diese Funktion liefert (ungefhr) die Summe der letzten nderungen an den Synapsen zurck
	* Das Ganze basiert auf dem gleitenden Durchschnitt der nderungen (delta_weight_avg)*/
TNSynapticWeight CSynapse::get_sum_last_weight_changes()
{
	return 	delta_weight_avg*anz_delta_weight_avg;
}
#endif


#ifdef USE_LIBXML
//-----------------------------------------------------------------------------
/** speichert in das XML-File, zu dem tree gehrt */
xmlNodePtr CSynapse::save_to_xml(xmlNodePtr tree){
	assert(tree);
	//einen neuen Knoten anlegen (mit dem Klassennamen)
	xmlNodePtr subtree;
	subtree = xmlNewChild(tree, NULL, (const xmlChar *)"CSynapse",NULL);
	xmlSaveDouble(subtree,"weight",weight);
	#ifdef USE_EXTENDED_SYNAPSE
		xmlSaveLong(subtree,"anz_delta_weight_avg",anz_delta_weight_avg);
		xmlSaveDouble(subtree,"delta_weight_avg",delta_weight_avg);
	#endif	
  return subtree;
}
//-----------------------------------------------------------------------------
/** ldt aus dem XML-File, zu dem tree gehrt */
void CSynapse::load_from_xml(xmlNodePtr tree){
	assert(tree);
	//in den Knoten gehen (mit dem Klassennamen)
	if (strcmp((const char*)tree->name, "CSynapse")) throw(666);
	xmlNodePtr subtree;
	subtree=tree->xmlChildrenNode;
	while (subtree!=0)
	{
		xmlLoadDouble(subtree,"weight",weight);
		#ifdef USE_EXTENDED_SYNAPSE
			xmlLoadLong(subtree,"anz_delta_weight_avg",anz_delta_weight_avg);
			xmlLoadDouble(subtree,"delta_weight_avg",delta_weight_avg);
		#endif	
		subtree=subtree->next;
	}
}
#endif

//-----------------------------------------------------------------------------
/** ndert das Gewicht der Synapse auf einen zuflligen Wert zwischen  min und max */
void CSynapse::set_random_weight(TNSynapticWeight min_weight=-1,TNSynapticWeight max_weight=1){
  assert(min_weight<=max_weight);
	double range=max_weight-min_weight;
	TNSynapticWeight weight;
	if (range==0)
	{ 	//die beiden Werte sind gleich
			weight=min_weight;
	}else
	{
			weight=((double)rand()/(double)RAND_MAX) //liefert eine Zufallszahl zw. 0 und 1
				*range				 //liefert eine Zahl zw. 0 und range
				+min_weight;			 //liefert eine Zahl zw. min_weight und min_weight+range
	}
	set_weight(weight);
}
//-----------------------------------------------------------------------------
/** ersetzt das presynaptische Neuron durch new_neuron */
void CSynapse::replace_pre_synaptic_neuron(CNeuron* new_neuron){
	//als erstes die Synapse vom ursprnglichen Neuron trennen
	pre_synaptic_neuron->remove_output(this);
	//dann das neue bernehmen
	pre_synaptic_neuron=new_neuron;
	//und die Synapse bei new_neuron aufnehmen
	new_neuron->add_output(this);	
}
