/* **************************************************************************
                          csynapse.h  -  description
                             -------------------
    begin                : Thu Feb 15 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CSYNAPSE_H
#define CSYNAPSE_H



#define USE_EXTENDED_SYNAPSE		//Falls dieses Ifdef def. ist wird die erweiterte Synapse verwendet
																//dh der gleitende Durchschnitt der letzten x Eingaben wird gebildet
																//Vererbung wird nicht verwendet, da sonst einige zeitkritische Fkt.
																//virtuell sein mssten

class CNeuron;

#include "global.h"
#include "tools.h"					//#defines
#include "cbasicneuron.h" 	//fr die Definifionen von TNSynapticWeight etc.

#ifdef USE_LIBXML
#endif



/** Implementation einer Synapse.

		Die Synapse besitzt zwei Zeiger. Einen auf das Vorgnger- und einen auf das Nachfolger-Neuron.
		Um eine Synapse zwischen 2 Neuronen zu erzeugen muss diese zustzlich bei beiden Neuronen eingetragen
		werden.
		Die Synapse besitzt ein Gewicht, welches die Strke der Verbindung zwischen den beiden Neuronen
		symbolisiert. Dieses kann durch @ref add_weight verndert werden.
		Die Funktion @ref process_input ist die zentrale Funktion fr die Synapse. Die auf
		die Synapse eintreffenden Signale werden dieser Funktion bergeben und diese erzeugt daraus
		das Signal, das an das postsynaptische Neuron weiter gegeben wird.
		
		Auf virtuelle Funktionen wurde soweit wie mglich verzichtet, um die Geschwindigkeit nicht
		zu sehr zu verschlechtern (Daher auch ein paar #ifdef's). Viele Funktionen hier werden _extrem_ hufig aufgerufen.

    @author Jon Hennig
  */

class CSynapse {
public: 
//	CSynapse();
	virtual ~CSynapse();
	
	/** erstellt eine Synapse zwischen 2 Neuronen, und trgt diese in den Neuronen entsprechend ein.
			*/
  CSynapse(CNeuron* pre_neuron, CNeuron* post_neuron, TNSynapticWeight a_weight);

  /** liefert eine Erregung (a_input) an die Synapse und gibt den daraus resultierenden postsynaptischen Wert
 			(generiert durch get_post_synaptic_value(a_input)) an das nachfolgende Neuron weiter
 			*/
  void process_input(TNPreSynapticValue a_input);

  /** ndert das Gewicht der Synapse auf weight)
  		*/
  void set_weight(TNSynapticWeight new_weight);

  /** liefert das Gewicht der Synapse
  		*/
  TNSynapticWeight get_weight();

  /** erhht das Gewicht um to_add, falls to_add kleiner 0 wird das Gewicht entsprechend vermindert.
  		*/
  void add_weight(TNSynapticWeight to_add);

#ifdef USE_LIBXML
		/** speichert in das XML-File, zu dem tree gehrt
				*/
		xmlNodePtr save_to_xml(xmlNodePtr tree);
		/** ldt aus dem XML-File, zu dem tree gehrt
				*/
		void load_from_xml(xmlNodePtr tree);
#endif

  /** ndert das Gewicht der Synapse auf einen zuflligen Wert zwischen  min und max
  		*/
  void set_random_weight(TNSynapticWeight min_weight=-1,TNSynapticWeight max_weight=1);

 	/** ersetzt das presynaptische Neuron durch new_neuron
 			*/
  void replace_pre_synaptic_neuron(CNeuron* new_neuron);

  /** liefert zu Signal (a_input) den entsprechenden postsynaptischen Wert.
			Diese Funktion ist virtuell und kann fr verschiedene Synapsenarten unterschiedlich sein.
			standardmig wird das Gewicht mit dem Input multipliziert.(wird von @ref process_input aufgerufen)
			*/
  virtual TNPostSynapticValue get_post_synaptic_value(TNPreSynapticValue a_input);

 protected:
 	/** Das Gewicht der Synapse
 			*/
	TNSynapticWeight	weight;
	
 public: //FIXME: should be protected
	CNeuron*		pre_synaptic_neuron;
	CNeuron*		post_synaptic_neuron;
	
#ifdef USE_EXTENDED_SYNAPSE
  	  /** diese Variable bildet den gleitenden Durchschnitt ber die letzten x Gewichtsnderungen
  	  		*/
  		TNSynapticWeight delta_weight_avg;
     public:
  		/** Statische Variable, die angibt, ber wieviele Elemente der gleitende Durchschnitt gebildet werden soll
  				*/
  		static long anz_delta_weight_avg;
  		/** Diese Funktion liefert (ungefhr) die Summe der letzten nderungen an den Synapsen zurck
  			  Das Ganze basiert auf dem gleitenden Durchschnitt der nderungen (delta_weight_avg)
  				*/
  		TNSynapticWeight get_sum_last_weight_changes();
#endif

};

#endif
