/* **************************************************************************
                          tools.h  -  description
                             -------------------
    begin                : Mon Sep 3 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TOOLS_H

#define TOOLS_H

#include "global.h" 				//for #defines
#ifdef USE_LIBXML
#include <libxml/tree.h>
#endif
#include <values.h>         //for MAXDOUBLE...

//#include <string>
#include <std/bastring.h>
#include "cdoublevector.h"

class CChangingProperty;

//Allgemeine Funktionen
//================================================
/** liefert einen Zufallswert zwischen min_value und max_value
    */
extern double get_random_value(double min_value, double max_value);

/** liefert eine Zufallszahl gem Gauss'scher Normalverteilung. Voraussetzung dafr ist
    eine Funktion get_random_value, die gleichverteilte Zufallszahlen liefert.
    Die Implementation beruht auf Donald E. Knuth's "The Art of Computer Programming". (siehe auch G. Marsaglia)
    @param mean       Mittelwert  - Standard: 0
    @param variance   Standardabweichung  - Standard: 1
    @param min_value  falls min_value oder max_value angegeben, so wird sichergestellt, dass das Ergebnis
                      immer im Bereich zwischen min und max liegt. Wenn nicht, wird ein gleichverteilter! Zufalls-
                      wert aus dem Bereich [min_value,max_value] ausgegeben. Fr min_value=-MAXDOUBLE und max_value=MAXDOUBLE
                      ergibt sich die (numerisch ber.) Gauss'sche Normalverteilung. Standard: -MAXDOUBLE
    @param max_value  siehe min_value, Standard: MAXDOUBLE
    */
extern double get_random_gauss(double mean=0, double variance=1,double min_value=-MAXDOUBLE, double max_value=MAXDOUBLE);


  /** throw mit einer Ausgabe der Fehlerursache in cerr
  		knnte man noch erweitern, so dass die Ausgabe unterdrckt, oder umleitbar ist (per globaler Einstellung)
  		@param err_number Fehlernummer
  		@param err_text wird als Fehlermeldung auf cerr ausgegeben
  		@param function_name Name der Funktion, in der der Fehler aufgetreten ist
  		*/
extern void throw_error(int err_number, const char* err_text, const char* function_name=0);
	


//Funktionen fr die XML-Library
//================================================
#ifdef USE_LIBXML
	/** speichert einen Double-Wert in dem XML-File ab
		*/
  extern void xmlSaveDouble(xmlNodePtr tree,const char * name,double value);
	
	/** speichert einen Long-Wert in dem XML-File ab
		*/
  extern void xmlSaveLong(xmlNodePtr tree,const char * name,long value);
	
	/** speichert einen bool-Wert in dem XML-File ab
		*/
  extern void xmlSaveBool(xmlNodePtr tree,const char * name,bool value);
	
	/** speichert einen CChangingProperty-Wert in dem XML-File ab
		*/
  extern void xmlSaveCProp(xmlNodePtr tree,const char * name,const CChangingProperty& value);
	
	/** speichert einen CDoubleVector-Wert in dem XML-File ab
	*/
  extern void xmlSaveDoubleVector(xmlNodePtr tree,const char * name,const CDoubleVector& value);
	
	/** ldt einen Double-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist
		* Der Wert value wird gendert
		*/
	extern bool xmlLoadDouble(xmlNodePtr tree,const char * name,double& value);
	
	/** ldt einen Long-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist
		* Der Wert value wird gendert
		*/
	extern bool xmlLoadLong(xmlNodePtr tree,const char * name,long& value);
	
	/** ldt einen Bool-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist
		* Der Wert value wird gendert
		*/
	extern bool xmlLoadBool(xmlNodePtr tree,const char * name,bool& value);
	
	/** ldt einen CChangingPorperty-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist
		* Der Wert value wird gendert
		*/
	extern bool xmlLoadCProp(xmlNodePtr tree,const char * name,CChangingProperty& value);
	
	/** ldt einen CCDoubleVector-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist
		* Der Vector wird geladen
		*/
	extern bool xmlLoadDoubleVector(xmlNodePtr tree,const char * name,CDoubleVector& vec);
	
	//weitere abgeleitete Funktionen
	extern bool xmlLoadInt(xmlNodePtr tree,const char * name,int& value);
	
	/** macht eine Tiefensuche (falls depth_first true, ansonsten Breitensuche),
		* nach dem Knoten mit den namen name und liefert
		* das erste Auftreten zurck, 0 bei Misserfolg
		*/
	extern xmlNodePtr xmlSearchNode(xmlNodePtr tree, const char* name, bool depth_first=true);
#endif


//Funktionen zur Stringverarbeitung
//================================================
//Die eigene Klasse CString wollte nicht so recht, wie ich wollte
// Daher nur die Typdef. und die entsprechenden Fkt.
//typedef string CString;
typedef basic_string<char> CString;

//Abkrzung
const CString::size_type NOT_FOUND=CString::npos;


/**	wie istream::getline(), aber fr CString
		*/
istream& getline(istream& str, CString& string);

/**	konvertiert von einem Double zu einem String
		*/
extern CString double_to_string(double value);
/**	konvertiert von einem Long zu einem String
		*/
extern CString long_to_string(long value);
/** konvertiert von einem bool zu einem String
		*/
extern CString bool_to_string(bool value);
/** konvertiert von einem String zu einem Double
    */
extern double string_to_double(const CString& value);
/** konvertiert von einem String zu einem Float
		*/
extern float string_to_float(const CString& value);
/** konvertiert von einem String zu einem Long
		*/
extern long string_to_long(const CString& value);
/** gibt wie printf aus, allerdings als CString
		*/
extern CString format_string(const char* format_str, ...);

/** Fgt zwei Strings zusammen
    */
extern CString cat_strings(const char* str1, const char* str2);


/** TOOLS:	
    ist gedacht fr Float-Parameter, die sich im Laufe der Zeit ndern (zb fr eine Lernrate).
	  Man kann einen aktuellen Wert setzen oder auslesen, und einen Zielwert vorgeben, gegen den der Wert
  	konvergieren soll (goal) Dies geschieht durch Multiplikation mit einstellbaren festen Faktor.
	
	  Kann statt des Datentyps float verwendet werden. ber die step() wird eine Iteration durchgefhrt.
	  change_factor() bestimmt die Rate der nderung. (z.B. 0.999 entspricht rel. langsamer Annherung)
	
	  <b> TODO: <b>
	  <li> Implementierung als Template um auch die Basis-Datentypen double, long zu erlauben
	  <li> Erweiterung, so dass auch andere Funktionen (als die Multiplikation mit einem festen Faktor) zum Konvergieren verwendet werden knnen
	  </li>
	  	
	  <b> BUG: <b
	  <li> nur fallende Werte werden untersttzt
	  </li>

    @author Jon Hennig
*/
class CChangingProperty
{
	#ifdef USE_LIBXML
		/* Lesen der Property aus einem XML-File
		   */
		friend bool xmlLoadCProp(xmlNodePtr tree,const char * name,CChangingProperty& value);	
		
		/* Schreiben der Property in ein XML-File
		   */
		friend void xmlSaveCProp(xmlNodePtr tree,const char * name,const CChangingProperty& value);
	#endif
public: // Public methods
   CChangingProperty();

   /* initialisieren mit Wert
      */
   CChangingProperty(float a_value);

  /** setzt den aktuellen Wert
		  */
  void set(float a_value);

  /** liefert den aktuellen Wert
		  */
  float get();

  /** setzt den Zielwert (Grenzwert)
  		*/
  void set_goal(float a_value);

  /** stellt den Faktor ein, mit dem sich die Eigenschaft
			vom aktuellen Wert zum Zielwert hin entfernt (z.B. 0.999)
			*/
  void set_change_factor(float a_value);

  /** macht einen Schritt Richtung Zielwert
  		*/
  void step();

  /** wie @ref get
  		*/
  operator const float() {return get();};

  /** wie @ref set
  		*/
  float operator = (const float a_value) {set(a_value); return a_value;};

protected: // Protected attributes

  /** der intern gespeicherte Wert.
  		*/
  float value;

  /** das Ziel, auf den der Wert zusteuert.
  		*/
  float goal;

  /** der Faktor mit dem die Eigenschaft gendert wird.
			*/
  float change_factor;
};



/**	TOOLS: NUR_INTERN: erlaubt es einfach Durchschnittswerte zu ermitteln
  	@author Jon Hennig
  */
class CAvg{
public:
    /** Initialisierung
        @param noof_values gibt an, wieviele Werte zur Durchschnittsbildung (max) herangezogen werden
        */
    CAvg(long noof_values);

    /** fgt den Wert zu Durchschnittsberechnung hinzu
        */
    void add(double value);

    /** ermittelt den Durchschnitt fr die eingestellte Anzahl der Werte
        @param noof_values gibt an, ber wieviele Werte der Durchschnitt berechnet werden soll
      */
    double get_avg(long noof_values=-1);

    /** setzt die Anzahl der Werte, die fr die Berechnung des Durchschnitts heran gezogen werden */
    void set_noof_values();
protected:
    long noof_values;   //Anzahl der Werte, die zur Berechnung des Durchschnitts heran gezogen werden
    double sum;         //die Summe ber noof_values Werte
    long count;         //wieviele Werte wurden bisher berechnet
    long pos;           //an welcher Stelle steht der Zeiger im Array
    CDoubleVector values; //die einzelnen Werte
};






#endif